//~~ cmdline.c ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
/*
 *  Command line interface to quartz.
 *
 * @author:    Matt Sun   <mattsun@cs.ucsb.edu>
 * @version:   $Date: 2005/10/25 20:22:19 $   $Revision: 1.21 $
 * @requires:   
 */

//~~[ Identity ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#define  __CMDLINE_C__
static char* srcIdentity = 
    "$Id: cmdline.c,v 1.21 2005/10/25 20:22:19 mattsun Exp $";


//~~[ Includes ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#include  <signal.h>
#include  <stdio.h>
#include  <readline/readline.h>
#include  <readline/history.h>

#include  "cmdline.h"
#include  "communication.h"
#include  "device/device.h"
#include  "error.h"
#include  "group.h"
#include  "init_quartz.h"
#include  "libsha1.h"
#include  "misc.h"


//~~[ Typedefs/Structures/Enums ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

typedef int  (*CMD_FUNC)( int argc, char* argv[] );

typedef struct _command_z
{
    char*     text;
    int       minSize;
    CMD_FUNC  handler;
    char*     desc;
}
  COMMAND;


//~~[ Constants ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

static const char* prompt = " Quartz> ";


//~~[ Variables ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//~~[ Prototypes ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    //----------------------------------------------------------------------
    /*
     *  Wrapper around cmd_Quit for signal handling.
     *
     * @param   signal Trapped signal.
     */
	    static void
    sig_Quit(
	int signal
    );

    //----------------------------------------------------------------------
    /*
     *  Breaks a string up on whitespace into an array of strings.
     *
     * @param   input  The string to parse.
     * @param   argc   Input: the size of argv.  Output: the curr size of argv.
     * @param   argv   The array of strings.
     * @return  returns
     */
	    static void
    parsecmd(
	char* input,
	int* argc,	// in/out
	char* argv[]
    );


    //----------------------------------------------------------------------
    /*
     *  Display the command array to the specified FILE*.
     *
     * @param   out   Where to display.
     * @param   argc  Number of arguments.
     * @param   argv  Arguments
     */
	    static void
    dumpCmd(
        FILE* out,
	int argc,
	char* argv[]
    );


    //----------------------------------------------------------------------
    /*
     *  Find a command in the global command list.
     *
     * @param   argc  Number of arguments.
     * @param   argv  Arguments.
     * @return  A pointer to the command structure, or NULL if not found.
     */
	    static COMMAND*
    findCommand(
	int argc,
	char* argv[]
    );


    //----------------------------------------------------------------------
    /*
     *  Command from/for the command line.
     *
     * @param   argc  Number of arguments.
     * @param   argv  Arguments.
     * @return  Exit code.
     */
    static int  cmd_Add		( int argc, char* argv[] );
    static int  cmd_Commit	( int argc, char* argv[] );
    static int  cmd_Hash	( int argc, char* argv[] );
    static int  cmd_Help	( int argc, char* argv[] );
    static int  cmd_Init	( int argc, char* argv[] );
    static int  cmd_Quit	( int argc, char* argv[] );
    static int  cmd_Remove	( int argc, char* argv[] );
    static int  cmd_Show	( int argc, char* argv[] );
    static int  cmd_Groups	( int argc, char* argv[] );
    static int  cmd_User	( int argc, char* argv[] );
    static int  cmd_SignOn	( int argc, char* argv[] );
    static int  cmd_Update	( int argc, char* argv[] );
    static int  cmd_IM          ( int argc, char* argv[] );
    static int  cmd_Stats       ( int argc, char* argv[] );
    static int  cmd_Sync        ( int argc, char* argv[] );
    static int  cmd_Setup       ( int argc, char* argv[] );


    //----------------------------------------------------------------------
    /*
     *  Format a number of bytes into a nicely formatting string representing
     *  the number of kilobytes.
     *
     *  For example, 43 bytes --> 0.04
     *                2 bytes --> ~0
     *             1234 bytes --> 1.23
     *            12345 bytes --> 12.3
     *           123567 bytes --> 124
     *
     * @param   bytes  Number of bytes.
     * @return  A nicely formatted string.
     */
	    static char*
    sizeKb(
	int bytes
    );


    //----------------------------------------------------------------------
    /*
     *  Remove leading and trailing whitespace.  This modifies the buffer
     *  being pointed to, so don't pass in a static value.
     *
     * @param   buffer  A character string.
     * @return  A pointer to the first non-whitespace with a '\0' placed in
     *              the first non-whitespace at the end.
     */
	    static char*
    trim(
	char* buffer
    );

//~~[ Commands ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

static COMMAND command[] = 
    {
	{ "add",      3,  cmd_Add,    "Add a file or group." },
//	{ "commit",   4,  cmd_Commit, "Save a new version." },
	{ "exit",     2,  cmd_Quit,   NULL },
//	{ "hash",     4,  cmd_Hash,   "Create a MD5 hash." },
	{ "help",     4,  cmd_Help,   "Get some help on a topic." },
//	{ "init",     4,  cmd_Init,   "Initialize some component." },
//	{ "login",    5,  cmd_SignOn, NULL },
//	{ "logoff",   6,  cmd_SignOn, NULL },
	{ "quit",     1,  cmd_Quit,   "Quit the program." },
	{ "remove",   3,  cmd_Remove, "Remove a file or group." },
	{ "rm",       2,  cmd_Remove, "Remove a file or group." },
	{ "ls",       3,  cmd_Groups, "List group(s) information." },
//	{ "groups",   5,  cmd_Groups, "List group(s) information." },
	{ "show",     3,  cmd_Show,   "Display a value." },
//	{ "signoff",  7,  cmd_SignOn, "Disconenct from the quartz network." },
//	{ "signon",   6,  cmd_SignOn, "Connect to the Quartz network." },
//	{ "update",   4,  cmd_Update, "Get the most recent version." },
//	{ "IM",       2,  cmd_IM,     "IM a message to your root node." },
	{ "stats",    4,  cmd_Stats,  "Get statistics from a node." },
//	{ "store",    4,  cmd_Commit, "Store/Save a new version." },
	{ "sync",     2,  cmd_Sync,   "Synchronize your group(s)/file(s)." },
	{ "user",     2,  cmd_User,   "Display all user information." },
//	{ "setup",    4,  cmd_Setup,  "Setup the user for Quartz." },

	{ NULL,       0,  NULL,       NULL  }
    };


//~~[ Implementation ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	int
cmdline(
    const char* prompt
){
    int    result;
    char*  input;
    char*  argv[ 32 ];
    int    argc = 32;
    COMMAND* cmd = NULL;

    if( NULL == prompt )
    {
	prompt = "";
    }

    signal(SIGSEGV, sig_Quit);
    signal(SIGBUS,  sig_Quit);
    signal(SIGINT,  sig_Quit);
    
    while( 1 )
    {
      pthread_mutex_lock(&mutex_cmdline);
      
      input = readline( prompt );
      
      if( '\0' != *input )
      {
	add_history( input );
	argc = 32;
	parsecmd( input, &argc, argv );
	COMMAND* cmd = findCommand( argc, argv );
	if( NULL != cmd )
	{
	  CMD_FUNC func = cmd->handler;
	  result = (*(cmd->handler))( argc, argv );
	}
	else if( 0 < argc )
	{
	  // printf( "  **  Unknown command: \"%s\"\n", argv[0] );
	  pthread_mutex_unlock(&mutex_cmdline);
	}
      }
      else
	pthread_mutex_unlock(&mutex_cmdline);
    }

    return( result );
}


//== Static Functions ==========================================================

	static void
parsecmd(
    char* input,
    int* argc,
    char* argv[]
){
    static char*  buffer = NULL;
    int maxArgs = *argc - 1;
    int currArg = 0;


    if( NULL != input )
    {
	if( NULL != buffer )
	{
	    free( buffer );
	}
	buffer = strdup( input );
        int cmd = 0;
	char* ptr = buffer;
	argv[ 0 ] = NULL;

        while( ('\0' != *ptr) && (currArg < maxArgs) )
	{
	    // skip over leading white space.
	    while( isspace( *ptr ) && ('\0' != *ptr) )
		++ptr;
	    if( '\0' == *ptr )
	    {
	        break;
	    }
	    argv[ currArg++ ] = ptr;
	    argv[ currArg ] = NULL;

	    // Skip over non whitespace.
	    while( ! isspace( *ptr ) && ('\0' != *ptr) )
	        ++ptr;
	    if( '\0' == *ptr )
	    {
	        break;
	    }
	    *(ptr++) = '\0';
	}

	*argc = currArg;
    }

    return;
}


//------------------------------------------------------------------------------

	static void
dumpCmd(
    FILE* out,
    int argc,
    char* argv[]
){
    FILE* o = (NULL != out) ? out : stdout;
    register int i;

    if( 0 < argc )
	printf( "\n" );
    for( i=0; i < argc; ++i )
    {
        fprintf( o, "  argv[ %d ] = \"%s\"\n", i, argv[i] );
    }
}


//------------------------------------------------------------------------------

	static COMMAND*
findCommand(
    int argc,
    char* argv[]
){
    COMMAND*  cmd = NULL;

    if( (0 < argc) && (NULL != argv[0]) )
    {
	char* arg0 = argv[0];
	int c = -1;

	while( NULL != command[++c].text )
	{
	    if( 0 == strncasecmp( arg0, command[c].text, command[c].minSize ) )
	    {
	        cmd = &(command[c]);
		break;
	    }
	}
    }

    return( cmd );
}

//------------------------------------------------------------------------------

	static void
sig_Quit(
    int signal
){
    putchar( '\n' );
    char* cmd[1] = { "quit" };
    cmd_Quit( 1, cmd );
}


//------------------------------------------------------------------------------

	static char*
sizeKb(
    int bytes
){
    static char buffer[ 6 ];

    if( bytes < 5 )
    {
        strncpy( buffer, "~0", sizeof(buffer) );
    }
    else if( bytes < 10000 )
    {
        snprintf( buffer, sizeof(buffer), "%4.2f", ((float)bytes)/1000.0 );
    }
    else if( bytes < 100000 )
    {
        snprintf( buffer, sizeof(buffer), "%4.1f", ((float)bytes)/1000.0 );
    }
    else
    {
        snprintf( buffer, sizeof(buffer), "%d", bytes/1000 );
    }

    return( buffer );
}


//------------------------------------------------------------------------------

	static char*
trim(
    char* buffer
){
    char*  ptr = buffer;

    if( NULL != ptr )
    {
        while( (*ptr != '\0') && isspace( *ptr ) )
	    ++ptr;

	if( '\0' != *ptr )
	{
	    char* last = ptr + strlen( ptr );
	    last--;

	    while( (*last != '\0') && isspace( *last ) )
		*(last--) = '\0';
	}
    }

    return( ptr );
}


//------------------------------------------------------------------------------
//--  Commands  ----------------------------------------------------------------
//------------------------------------------------------------------------------

	static int
cmd_Quit(
    int argc,
    char* argv[]
){
    int exitCode = 0;

    if( (1 < argc) && (0 == strcasecmp( argv[1], "help" )) )
    {
        printf( "\n\
--[ Exit/Quit ]--------------------------------------------------------------\n\
\n\
    Exit the program.\n\
\n\
    Usage\n\
        quit   or\n\
	exit   \n\
\n\
-----------------------------------------------------------------------------\n\
\n\
" );

      pthread_mutex_unlock(&mutex_cmdline);
    }
    else
    {
      printf("\n***********************************\n");
      printf(" Shutting down Quartz...");

      user_incr_signoffs();
      update_config_files();

      // only send your usage if you are not administrator
      if (strcmp(deviceInfo->userName, "administrator") != 0)
	handler_stats_req((Key) key_makehash("128.111.40.220")); // beast.cs.ucsb.edu

      printf(" Good bye!\n");
      printf("***********************************\n\n");
      
      pthread_mutex_unlock(&mutex_cmdline);

      exit( exitCode );
    }

    return( exitCode );
}


//------------------------------------------------------------------------------

	static int
cmd_Help(
    int argc,
    char* argv[]
){
    int exitCode = 0;
    COMMAND* cmd = NULL;

    // No args, just list commands.
    if( 1 == argc )
    {
	printf( "\n    Available Commands:\n" );
	printf( "   ----------  -----------------------------------------\n" );

	for( cmd = &(command[0]); NULL != cmd->text; cmd++ )
	{
	    if( NULL != cmd->desc )
	    {
	        printf( "    %-10.10s  %s\n", cmd->text, cmd->desc );
	    }
	}
	putchar( '\n' );
    }

    // Help help
    else if( 0 == strcasecmp( argv[1], "help" ) )
    {
        printf( "\n\
--[ Help ]-------------------------------------------------------------------\n\
\n\
    Get some help on a topic.\n\
\n\
    Usage\n\
        help        - List all available commands.\n\
        help <cmd>  - Get help on a particular command.\n\
\n\
-----------------------------------------------------------------------------\n\
\n"
);
      pthread_mutex_unlock(&mutex_cmdline);

    }

    // Find the command.
    else if( NULL == (cmd = findCommand( argc-1, &(argv[1]) )) )
    {
        printf( "Unknown command: \"%s\"\n", argv[1] );
    }

    // Call the command with "help" as the first argument.
    else
    {
        char* help[3];
	help[0] = argv[1];
	help[1] = "help";
	help[2] = NULL;

	exitCode = (*(cmd->handler))( 2, help );
    }

      pthread_mutex_unlock(&mutex_cmdline);

    return( exitCode );
}


//------------------------------------------------------------------------------

	static int
cmd_Show(
    int argc,
    char* argv[]
){
    int exitCode = 0, group = -1, g = 0;

    if( (1 == argc) || (0 == strcasecmp( argv[1], "help" )) )
    {
        printf( "\n\
--[ Show ]-------------------------------------------------------------------\n\
\n\
    Display a value.\n\
\n\
    Usage\n\
        show value [args]\n\
	    Where \"args\" is one of:\n\
	      group [name] - Display group [name] information.\n\
	      user         - Display user configuration.\n\
	      all          - Display all synchronization groups.\n\
\n\
    Examples\n\
        show groups\n\
            ... list all groups ...\n\
\n\
        show all\n\
            ... list all groups and their files ...\n\
\n\
        show group [name]\n\
            ... lists group [name] and its files ...\n\
\n\
-----------------------------------------------------------------------------\n\
\n\
" );
    }
    else if( 0 == strncasecmp( argv[1], "storage", 7 ) )
    {
      char user_curr[41];
      char group_curr[41];

      _meta_user_z *curr_user   = NULL;
      _info_group_z *curr_group = NULL;
      _meta_local_z *curr_file  = NULL;

      if (meta_user != NULL)
      {
	fprintf(stderr, "**********************\n");

	curr_user = meta_user;
	while (curr_user != NULL)
	{
	  fprintf(stderr, "1\n");
	  sprintf(user_curr, "%s", curr_user->user_id);
	  user_curr[40] = '\0';
	  
	  fprintf(stderr, "[U] = %s\n", user_curr);

	  if (curr_user->groups != NULL)
	  {
	    curr_group = curr_user->groups;
	    while (curr_group != NULL)
	    {
	      sprintf(group_curr, "%s", curr_group->group_id);
	      group_curr[40] = '\0';
	      
	      fprintf(stderr, "  [G] = %s\n", group_curr);
	      
	      curr_file = curr_group->files;
		  
	      while (curr_file != NULL)
	      {
		fprintf(stderr, "    [F] = %s\n", curr_file->file_name);
		curr_file = curr_file->next;
	      }
	    
	      if (curr_group->next != NULL)
		curr_group = curr_group->next;
	      else
		break;
	    }
	  }
	  
	  curr_user = curr_user->next;
	}

	fprintf(stderr, "**********************\n");
      }
    }
    // *** GROUPS ***
    else if( 0 == strncasecmp( argv[1], "group", 5 ) )
    {
        printf("\n");

 	// Make sure there is something to print.
	if( NULL == deviceInfo->inMemGroups )
	{
	  printf( " There are currently no groups.\n" );
	}


	// Print the specified groups.
	else if( 2 < argc )
	{
	    int g;
	    INMEM_GROUP* group = NULL;

	    for( g=2; g < argc; ++g )
	    {
	        if( NULL != (group = group_find_group( argv[g] )) )
		{
		  group_print( stdout, group );
		}
		else
		{
		  printf( " There is no such group \"%s\"\n", argv[2] );
		}
	    }
	}

	// Iterate through the groups.
	else
	{
	  printf("   Total Groups = %d\n", deviceInfo->numGroups);

	  INMEM_GROUP* group = (INMEM_GROUP*)deviceInfo->inMemGroups;
	  
	  while( NULL != group )
	  {
	    printf("   + %s (%d bytes)\n", group->name, group_total_size(group));
	    group = group->next;
	  }
	}

    }

// *** USER ***
    else if( 0 == strncasecmp( argv[1], "user", 4 ) )
    {
        if( NULL != deviceInfo )
	{
	    print_device_info( stdout, deviceInfo );
	}
    }

// *** INODES ***
    else if( 0 == strncasecmp( argv[1], "inodes", 4 ) )
    {
      inode_print_all();
    }


// *** ALL ***
    else if( 0 == strncasecmp( argv[1], "all", 4 ) )
    {
      INMEM_GROUP* group = (INMEM_GROUP*)deviceInfo->inMemGroups;

      while( NULL != group )
      {
	group_print( stdout, group );
	group = group->next;
      }
    }

    // WTF?
    else
    {
      // printf( " Don't know how to show \"%s\"\n", argv[1] );
      exitCode = 1;
    }


    pthread_mutex_unlock(&mutex_cmdline);
    return( exitCode );
}


//------------------------------------------------------------------------------


//------------------------------------------------------------------------------

	static int
cmd_Groups(
    int argc,
    char* argv[]
){
    int exitCode = 0, group = -1, g = 0;

    if( argc > 1 && 0 == strcasecmp( argv[1], "help" ) )
    {
        printf( "\n\
--[ List ]-------------------------------------------------------------------\n\
\n\
    List all group(s) information.\n\
\n\
    Usage\n\
        ls \n\
\n\
    Examples\n\
        ls \n\
            ... list all groups and their files ...\n\
\n\
-----------------------------------------------------------------------------\n\
\n\
" );
    }
    else if (argc > 1)
    {
      // Make sure there is something to print.
      if( NULL == deviceInfo->inMemGroups )
	  printf( "\n There are currently no groups.\n" );
      // Print the specified groups.
      else if( 0 == strncasecmp( argv[1], "all", 3 ) )
      {
	INMEM_GROUP* group = (INMEM_GROUP*)deviceInfo->inMemGroups;
	
	while( NULL != group )
	{
	  group_print( stdout, group );
	  group = group->next;
	}
      }
      else if (0 == strncasecmp(argv[1], "group", 5))
      {
	int g;
	INMEM_GROUP* group = NULL;
	
	for( g=2; g < argc; ++g )
	{
	  if( NULL != (group = group_find_group( argv[2] )) )
	  {
	    group_print( stdout, group );
	  }
	  else
	  {
	    printf( " There is no such group \"%s\"\n", argv[1] );
	  }
	}
      }
      // Iterate through the groups.
      else
      {
	printf("\n   Total Groups = %d\n", deviceInfo->numGroups);
	
	INMEM_GROUP* group = (INMEM_GROUP*)deviceInfo->inMemGroups;
	
	while( NULL != group )
	{
	  printf("   + %s (%d bytes)\n", group->name, group_total_size(group));
	  group = group->next;
	}
      }
    }
    else
    {
      if (deviceInfo != NULL)
      {
	printf("\n   Total Groups = %d\n", deviceInfo->numGroups);
	
	  INMEM_GROUP* group = (INMEM_GROUP*)deviceInfo->inMemGroups;
	  
	  while( NULL != group )
	  {
	    printf("   + %s (%d bytes)\n", group->name, group_total_size(group));
	    group = group->next;
	  }
      }
    }

    pthread_mutex_unlock(&mutex_cmdline);
    return( exitCode );
}

//------------------------------------------------------------------------------

	static int
cmd_User(
    int argc,
    char* argv[]
){
    int exitCode = 0, group = -1, g = 0;

    if( argc > 1 && 0 == strcasecmp( argv[1], "help" ) )
    {
        printf( "\n\
--[ User ]-------------------------------------------------------------------\n\
\n\
    Display user information.\n\
\n\
    Usage\n\
        user\n\
\n\
    Examples\n\
        user\n\
            ... shows all user information ...\n\
\n\
-----------------------------------------------------------------------------\n\
\n\
" );
    }
    else
    {
      if( NULL != deviceInfo )
	print_device_info( stdout, deviceInfo );

      exitCode = 1;
    }

    pthread_mutex_unlock(&mutex_cmdline);
    return( exitCode );
}


//------------------------------------------------------------------------------

	static int
cmd_SignOn(
    int argc,
    char* argv[]
){
    int exitCode = 0;

    if( (1 < argc) && (0 == strcasecmp( argv[1], "help" )) )
    {
        printf( "\n\
--[ Signin/Signoff ]---------------------------------------------------------\n\
\n\
    Connect, or disconnect, from the quartz network.\n\
\n\
    Usage\n\
        signin\n\
	signoff\n\
\n\
-----------------------------------------------------------------------------\n\
\n\
" );
    }

    else if( (0 == strncasecmp( argv[0], "signin", 6))
          || (0 == strncasecmp( argv[0], "login", 5 ))
          || (0 == strncasecmp( argv[0], "conn", 4 )) )
    {
      //exitCode = quartz_signon();
    }
    else if( (0 == strncasecmp( argv[0], "signoff", 7))
          || (0 == strncasecmp( argv[0], "logout", 6 ))
          || (0 == strncasecmp( argv[0], "disconn", 7 )) )
    {
      //exitCode = quartz_signoff();
    }

    pthread_mutex_unlock(&mutex_cmdline);
    return( exitCode );
}


//------------------------------------------------------------------------------

	static int
cmd_Commit(
    int argc,
    char* argv[]
){
    int exitCode = 0;

    // Not currently allowing syncing all files.

    if( (1 == argc) || (0 == strcasecmp( argv[1], "help" )) )
    {
        printf( "\n\
--[ Commit ]-----------------------------------------------------------------\n\
\n\
    Push a newer version of the file out into storage in the Quartz system.\n\
\n\
    Usage\n\
        commit [[group] [groupname ...]]\n\
        commit file groupname filename [...]\n\
\n\
    Examples\n\
        commit bash\n\
            ... synchronize all files in group \"bash\" ...\n\
\n\
        commit file bash .bashrc .bash_profile\n\
            ... synchronize the files \".bashrc\" and \".bash_profile\"\n\
			    in group \"bash\" ...\n\
\n\
-----------------------------------------------------------------------------\n\
\n\
" );
    }
    else if( 1 == argc )
    {
//        exitCode = sync_commitGroup( NULL );
    }

    else if( 0 == strncasecmp( argv[1], "file", 4 ) )
    {
        if( 4 > argc )
	{
	    char* cmd[2] = { "sync", "help" };
	    cmd_Commit( 2, cmd );
	}
	else
	{
	    int file = 3;
	    for( ; (file < argc) && (0 == exitCode) ; ++file )
	    {
		exitCode = sync_commitFile( argv[2], argv[file] );
	    }
	}
    }

    else
    {
        int g=1;
	for( ; (g < argc) && (0 == exitCode); ++g )
	{
	  exitCode = sync_commitGroup( argv[g], FORCE_COMMIT );
	}
    }

    fprintf(stderr, "finished commit group\n");
    pthread_mutex_unlock(&mutex_cmdline);
    return( exitCode );
}


//------------------------------------------------------------------------------

	static int
cmd_Hash(
    int argc,
    char* argv[]
){
    int exitCode = 0;
    FILE*  fileToHash = NULL;

    if( (1 == argc) || (0 == strcasecmp( argv[1], "help" )) )
    {
        printf( "\n\
--[ Hash ]-------------------------------------------------------------------\n\
\n\
    Create an MD5 hash of an object.\n\
\n\
    Usage\n\
        hash string [string...]\n\
        hash file filename [filename...]\n\
\n\
    Note: Quartz does not currently interpret quotes correctly!\n\
\n\
-----------------------------------------------------------------------------\n\
\n\
" );
    }

    else if( (1 < argc) && (0 == strncasecmp( argv[1], "str", 3 )) )
    {
        HASH  strHash = { '\0' };

	// Specified on the command line?
        if( 2 < argc )
	{
	    int s;

	    for( s=2; s < argc; ++s )
	    {
	        if( NULL != sha1_string( argv[s], &strHash ) )
		{
		    printf( "  %s <-- \"%s\"\n",
			    sha1_hashToString( &strHash ), argv[s] );
		}
		else
		{
		    printf( "  \"%s\": Couldn't create hash\n", argv[s] );
		}
	    }

	    fflush( stdout );
	}

	// Go get it.
	else
	{
	    char* input = readline( "  String: " );
	    if( '\0' != *input )
	    {
	        if( NULL != sha1_string( input, &strHash ) )
		{
		    printf( "  %s <-- \"%s\"\n",
			    sha1_hashToString( &strHash ), input );
		}
		else
		{
		    printf( "  \"%s\": Couldn't create hash\n", input );
		}
	    }
	}
    }

    else if( (1 < argc) && (0 == strncasecmp( argv[1], "fil", 3 )) )
    {
        if( 2 == argc )
	{
	    exitCode = 1;
	    printf( "Please specify file to hash!\n" );
	}
	else
	{
	    register int f;
	    HASH fileHash = { '\0' };

	    for( f=2; f < argc; ++f )
	    {
	        if( NULL != sha1_fileByName( argv[f], &fileHash ) )
		{
		    printf( "  %s <-- \"%s\"\n",
			    sha1_hashToString( &fileHash ), argv[f] );
		}
	    }
	}
    }

    else
    {
        printf( "Don't know how to hash a \"%s\"\n", argv[1] );
    }

    pthread_mutex_unlock(&mutex_cmdline);
    return( exitCode );
}


//------------------------------------------------------------------------------

	static int
cmd_Update(
    int argc,
    char* argv[]
){
    int exitCode = 0;
    FILE*  fileToHash = NULL;


    if( (1 < argc) && (0 == strcasecmp( argv[1], "help" )) )
    {
        printf( "\n\
--[ Check ]------------------------------------------------------------------\n\
\n\
    Check and/or display the status.\n\
\n\
    Usage\n\
        check                  - Display general status.\n\
	check group [file ...] - Update the current state of the files.\n\
\n\
-----------------------------------------------------------------------------\n\
\n\
" );
    }

    else if( 1 == argc )
    {
        if( dev_isConnected() )
	{
	    printf( "You are connected as node %s\n", dev_getNodeId() );
	}
	else
	{
	    printf( "You are not currently connected (as %s)\n",
			    dev_getNodeId() );
	}

	if( sync_isSynchronizing() )
	{
	    printf( "Currently synchronizing files.\n" );
	}

	if( NULL != deviceInfo )
	{
	    INMEM_GROUP* group = (INMEM_GROUP*)deviceInfo->inMemGroups;
	    int numFiles = 0;
	    int numGroups = 0;
	    int totSize = 0;

	    while( NULL != group )
	    {
	        numGroups++;
		numFiles += group->numFiles;
		totSize += group->totalSize;

		group = group->next;
	    }

	    printf( "\n There are %d files in %d groups, ", numFiles, numGroups );
	    printf( "\n using a total of %s kilobytes.\n", sizeKb( totSize ) );
	}
    }

    else if( 1 < argc )
    {
        char* group = argv[1];
	INMEM_FILE* file = malloc(sizeof(INMEM_FILE));

	if( 0 == strcmp( group, "*" ) )
	    group = NULL;


	if( 2 == argc)
	{
	  exitCode = sync_updateGroup( group );


	  return ( 0 );
	}
	if( 2 < argc )
	{
	    int f;

	    for( f=2; f < argc; ++f )
	    {
	      sha1_stringToHash(argv[f], &(file->id));

	      file->version = 1;

		sync_updateFile( group, file );
	    }
				//   *************************
	    return( 0 );	//  ***  DOES NOT CONTINUE  ***
				//   *************************
	}

	exitCode = sync_updateFile( group, NULL );
    }

    pthread_mutex_unlock(&mutex_cmdline);
    return( exitCode );
}


//------------------------------------------------------------------------------

	static int
cmd_Add
(
    int argc,
    char* argv[]
){
    int exitCode = 0;

    if( (1 == argc) || (0 == strcasecmp( argv[1], "help" )) )
    {
        printf( "\n\
--[ Add ]--------------------------------------------------------------------\n\
\n\
    Add a file or group to be synchronized.\n\
\n\
    Usage\n\
        add file \n\
        add group [groupname ...]\n\
\n\
    Examples\n\
        add group bash csh\n\
            ... create the groups \"bash\" and \"csh\" ...\n\
\n\
        add group\n\
            ... prompts the user for the group name ...\n\
\n\
        add file\n\
            ... prompts the user for the filename and group ...\n\
\n\
-----------------------------------------------------------------------------\n\
\n\
" );

	pthread_mutex_unlock(&mutex_cmdline);
    }

// *** PROMPTING ***
    else if( 2 == argc )
    {
      if( 0 == strncasecmp( argv[1], "group", 5 ) )
	{
	    char* group = NULL;
	    INMEM_GROUP* inMemGroup = (INMEM_GROUP*)!NULL;
	    printf( "\n  Enter a blank line for the group name to exit.\n" );
	    while( NULL != inMemGroup )
	    {
	        group = readline( "  Group> " );
		if( '\0' == *group )
		{
		    break;
		}

		inMemGroup = group_new_group( trim( group ) );

		if (inMemGroup != NULL)
		  user_incr_groups(1);

		//(INMEM_GROUP*) deviceInfo->inMemGroups = (INMEM_GROUP*) inMemGroup;
	    }

	    pthread_mutex_unlock(&mutex_cmdline);
	}

      else if( 0 == strncasecmp( argv[1], "file", 4 ) )
	{
	    char*  group =NULL;
	    char*  file =NULL;
	    char   lastGroup[ 80 ] = { '\0' };
	    char   groupPrompt[ 80 ] = { '\0' };
	    INMEM_FILE* inMemFile = NULL;
	    INMEM_GROUP* inMemGroup = NULL;
	    struct stat statBuffer;

	    printf( "\n  Enter a blank line for the filename to exit.\n" );
	    while( 0 == exitCode )
	    {
	        file = readline( "\n  File to add> " );
		if( '\0' == *file )
		{
		  pthread_mutex_unlock(&mutex_cmdline);
		  break;
		}

		if( '\0' != *lastGroup )
		{
		    snprintf( groupPrompt, sizeof(groupPrompt),
				    "  Add to group[%s]> ", lastGroup );
		}
		else
		{
		    strncpy( groupPrompt, "  Add to group> ",
				    sizeof(groupPrompt) );
		}
	        group = readline( groupPrompt );
		if( '\0' == *group )
		{
		    if( '\0' == *lastGroup )
		    {
		      pthread_mutex_unlock(&mutex_cmdline);
		      break;
		    }
		    else
		    {
		      group = lastGroup;
		    }
		}
		
		printf("\n");

		if (stat(trim(file), &statBuffer) < 0)
		{
		  printf(" ** File does not exist locally (will not add)\n");
		  pthread_mutex_unlock(&mutex_cmdline);
		}
		else
		{
		  inMemFile = group_addFile( trim( file ), trim( group ) );

		  if (inMemFile != NULL)
		  {
		    // to get the size of the file for quota purposes
		    stat(inMemFile->name, &statBuffer);
		    inMemFile->size    = statBuffer.st_size;
		    inMemFile->version = 0;
		    
		    //user_incr_file_size(statBuffer.st_size, 1);
		    user_incr_files(1);
		    update_config_files();
		    
		    sync_syncFile(group, inMemFile);
		  }
		  else
		    pthread_mutex_unlock(&mutex_cmdline);
		
		  exitCode = (NULL != inMemGroup) ? 0 : 1;
		  
		  if( !exitCode )
		  {
		    strncpy( lastGroup, group, sizeof( lastGroup ) );
		  }
		}
	    }
	}

	else
	{
	    printf( " Don't know how to add \"%s\"\n", argv[1] );
	    pthread_mutex_unlock(&mutex_cmdline);
	}
    }

// *** GROUPS ***
    else if( 0 == strncasecmp( argv[1], "group", 5 ) )
    {
	int group = 2;
	INMEM_GROUP* inMemGroup =NULL;

        for( ; (group < argc) && (0 == exitCode); ++group )
	{
	  if (strlen(argv[group]) <= 32)
	  {
	    inMemGroup = group_new_group( argv[group] );

	    if (inMemGroup != NULL)
	      user_incr_groups(1);

	    exitCode = (NULL != inMemGroup) ? 0 : 1;
	  }
	  else
	  {
	    printf( " ** Group name must be less than 32 characters\n");
	    exitCode = 1;
	  }
	}

	pthread_mutex_unlock(&mutex_cmdline);
    }

// *** FILE ***
    else
    {
        int file = 1;
	int group = argc-1;
	INMEM_FILE* inMemFile = NULL;

        if( 0 == strcasecmp( argv[1], "file" ) )
	{
	    ++file;
	}

	if( file >= group )
	{
	    printf( "You need to specify a group to add the file to.\n" );
	    exitCode = 4;
	}
	else
	{
	    for( ; (file < group) && (0 == exitCode); ++file )
	    {
	        inMemFile = group_addFile( argv[file], argv[group] );

		if (inMemFile != NULL)
		  user_incr_files(1);

		exitCode = (NULL != inMemFile) ? 0 : 1;
	    }
	}

	pthread_mutex_unlock(&mutex_cmdline);
    }

    return( exitCode );
}


//------------------------------------------------------------------------------

	static int
cmd_Remove
(
    int argc,
    char* argv[]
){
    int exitCode = 0;

    if( (1 == argc) || (0 == strcasecmp( argv[1], "help" )) )
    {
        printf( "\n\
--[ Remove ]-----------------------------------------------------------------\n\
\n\
    Remove a file from a group, or an entire group.\n\
\n\
    Usage\n\
        remove [file] group\n\
\n\
    Examples\n\
        remove file .bashrc bash\n\
            ... remove the file \".bashrc\" from the group \"bash\" ...\n\
\n\
        remove group bash\n\
            ... remove the group, and all files in the \"bash\" group ...\n\
\n\
        remove file \n\
            ... prompts the user for the filename and group ...\n\
\n\
        remove group \n\
            ... prompts the user for the group name ...\n\
\n\
-----------------------------------------------------------------------------\n\
\n\
" );
    }

// *** PROMPTING ***
    else if( 2 == argc )
    {
      if( 0 == strncasecmp( argv[1], "group", 5 ) )
	{
	    char* group = NULL;
	    INMEM_GROUP* inMemGroup = (INMEM_GROUP*)!NULL;
	    while( NULL != inMemGroup )
	    {
	        group = readline( "  Group> " );
		if( '\0' == *group )
		{
		    break;
		}

		// remove the group and its files on the p2p side
		sync_removeGroup(trim(group));

		// remove the group and its files on the local side
		inMemGroup = group_remove_group(trim(group));

		if (inMemGroup != NULL)
		  user_incr_groups(0);
	  
		exitCode = (NULL != inMemGroup) ? 0 : 1;

		//(INMEM_GROUP*) deviceInfo->inMemGroups = (INMEM_GROUP*) inMemGroup;
	    }
	}

      else if( 0 == strncasecmp( argv[1], "file", 4 ) )
	{
	    char*  group =NULL;
	    char*  file =NULL;
	    char   lastGroup[ 80 ] = { '\0' };
	    char   groupPrompt[ 80 ] = { '\0' };
	    INMEM_FILE* inMemFile   = NULL;
	    INMEM_GROUP* inMemGroup = NULL;

	    printf( " Enter a blank line for the filename to exit.\n" );
	    while( 0 == exitCode )
	    {
	        file = readline( "\n  File to remove> " );
		if( '\0' == *file )
		{
		    break;
		}

		if( '\0' != *lastGroup )
		{
		    snprintf( groupPrompt, sizeof(groupPrompt),
				    "  Remove from group[%s]> ", lastGroup );
		}
		else
		{
		    strncpy( groupPrompt, "  Remove from group> ",
				    sizeof(groupPrompt) );
		}
	        group = readline( groupPrompt );
		if( '\0' == *group )
		{
		    if( '\0' == *lastGroup )
		    {
			break;
		    }
		    else
		    {
			group = lastGroup;
		    }
		}

		inMemFile = group_removeFile( trim( file ), trim( group ) );
		
		if (inMemFile != NULL)
		  user_incr_files(0);
		
		exitCode = (NULL != inMemGroup) ? 0 : 1;

		if( !exitCode )
		{
		    strncpy( lastGroup, group, sizeof( lastGroup ) );
		}
	    }
	}

	else
	{
	    printf( "Don't know how to remove \"%s\"\n", argv[1] );
	}
    }

// *** GROUPS ***
    else if( 0 == strncasecmp( argv[1], "group", 5 ) )
    {
	int group = 2;
	INMEM_GROUP* inMemGroup =NULL;

        for( ; (group < argc) && (0 == exitCode); ++group )
	{
	  // remove the group and its files on the p2p side
	  sync_removeGroup(argv[group]);

	  // remove the group and its files on the local side
	  inMemGroup = group_remove_group( argv[group] );

	  if (inMemGroup != NULL)
	    user_incr_groups(0);
	  
	  exitCode = (NULL != inMemGroup) ? 0 : 1;
	}
    }

// *** FILE ***
    else
    {
        int file = 1;
	int group = argc-1;
	INMEM_FILE* inMemFile = NULL;

        if( 0 == strcasecmp( argv[1], "file" ) )
	{
	    ++file;
	}

	if( file >= group )
	{
	    printf( "You need to specify a group to add the file to.\n" );
	    exitCode = 4;
	}
	else
	{
	    for( ; (file < group) && (0 == exitCode); ++file )
	    {
	        inMemFile = group_removeFile( argv[file], argv[group] );

		if (inMemFile != NULL)
		  user_incr_files(0);

		exitCode = (NULL != inMemFile) ? 0 : 1;
	    }
	}
    }

    pthread_mutex_unlock(&mutex_cmdline);
    return exitCode;
}


//------------------------------------------------------------------------------

	static int
cmd_Init(
    int argc,
    char* argv[]
){
    int exitCode = 0;

    if( (1 == argc) || (0 == strcasecmp( argv[1], "help" )) )
    {
        printf( "\n\
--[ Initialize ]-------------------------------------------------------------\n\
\n\
    Initialize a quartz component.\n\
\n\
    Usage\n\
        init key       - create a private key.\n\
        init neighbor  - initialize the neihbor set.\n\
        init sync      - start the syncronization thread.\n\
        init user      - initialize the user/node info.\n\
\n\
-----------------------------------------------------------------------------\n\
\n\
" );
    }

    // *** USER ***
    else if( 0 == strncasecmp( argv[1], "user", 4 ) )
    {
        printf( "\n" );
	exitCode = init_user_data();
    }

    // *** KEYS ***
    else if( 0 == strncasecmp( argv[1], "key", 3 ) )
    {
        printf( "\n" );
	char keyFilename[ FILENAME_MAX ];
	get_config_file( "key", keyFilename, FILENAME_MAX );
	//char*  password = misc_getKeyPassword();
	char* password = deviceInfo->password;

	struct stat statBuffer;
	if( -1 == stat( keyFilename, &statBuffer ) )
	{
	    printf( "Creating key in \"%s\"\n", keyFilename );
	    quartz_generate_key( keyFilename, password );
	    exitCode = 0;
	}
    }
    else
    {
        printf( "I do not know how to initialize \"%s\"\n" );
	exitCode = 1;
    }

    pthread_mutex_unlock(&mutex_cmdline);
    return( exitCode );
}

static int cmd_IM( int argc, char* argv[] )
{
  
    pthread_mutex_unlock(&mutex_cmdline);

}

static int cmd_Stats( int argc, char* argv[] )
{
  int exitCode = 0;
  Key key_send;

  if( (1 < argc) && (0 == strcasecmp( argv[1], "help" )) )
  {
    printf( "\n\
--[ Stats ]---------------------------------------------------------\n\
\n\
    Get stats from another node and store in log.\n\
\n\
    Usage\n\
        stats node_IP_address\n\
\n\
-----------------------------------------------------------------------------\n\
\n\
" );
  }
  else if( (0 == strncasecmp( argv[0], "stats", 4)))
  {
    if (argv[1] != NULL)
    {  
      if (deviceInfo != NULL)
      {
	if (strcmp(deviceInfo->userName, "administrator") == 0)
	{
	  if (strlen(argv[1]) < 32)
	  {
	    key_send = (Key)key_makehash( argv[1] );
	  
	    char* s_sendbuf = (char *) malloc(sizeof(Key));
	    memset(s_sendbuf, '\0', sizeof(Key));
	    memcpy(s_sendbuf, &(deviceInfo->nodeKey), sizeof(Key));
	    
	    chimera_send(key_send, CTRL_STATS_REQ, sizeof(Key), s_sendbuf);
	    free(s_sendbuf);
	  }
	}
      }
    }
  }

  pthread_mutex_unlock(&mutex_cmdline);
  return( exitCode );
}

//------------------------------------------------------------------------------

	static int
cmd_Sync(
    int argc,
    char* argv[]
){
    int exitCode = 0;

    // Not currently allowing syncing all files.

    if( (1 == argc) || (0 == strcasecmp( argv[1], "help" )) )
    {
        printf( "\n\
--[ Sync ]-----------------------------------------------------------------\n\
\n\
    Sync a file or group of files from the Quartz system.\n\
\n\
    Usage\n\
        sync [[group] [groupname ...]]\n\
        sync file groupname filename [...]\n\
\n\
    Examples\n\
        sync bash\n\
            ... synchronize all files in group \"bash\" ...\n\
\n\
        sync file bash .bashrc .bash_profile\n\
            ... synchronize the files \".bashrc\" and \".bash_profile\"\n\
			    in group \"bash\" ...\n\
\n\
-----------------------------------------------------------------------------\n\
\n\
" );
      pthread_mutex_unlock(&mutex_cmdline);

    }
    else if( 1 == argc )
    {
//        exitCode = sync_commitGroup( NULL );
    }

    /*
    else if( 0 == strncasecmp( argv[1], "file", 4 ) )
    {
        if( 4 > argc )
	{
	    char* cmd[2] = { "sync", "help" };
	    cmd_Commit( 2, cmd );
	}
	else
	{
	    int file = 3;
	    for( ; (file < argc) && (0 == exitCode) ; ++file )
	    {
		exitCode = sync_syncFile( argv[2], argv[file] );
	    }
	}
    }
    */

    // allow only syncing of groups
    else
    {
        int g=1;
	for( ; (g < argc) && (0 == exitCode); ++g )
	{
	    exitCode = sync_syncGroup( argv[g] );
	    fflush(stdout);
	}
    }

    return( exitCode );
}


//------------------------------------------------------------------------------

	static int
cmd_Setup(
    int argc,
    char* argv[]
){
    int exitCode = 0;

    // setup user information
    init_user_data();
    
    // create symmetric key based on the password
    char keyFilename[ FILENAME_MAX ];
    get_config_file( "key", keyFilename, FILENAME_MAX );
    char* password = deviceInfo->password;
    
    struct stat statBuffer;
    if( -1 == stat( keyFilename, &statBuffer ) )
    {
      printf( "Creating key in \"%s\"\n", keyFilename );
      quartz_generate_key( keyFilename, password );
      exitCode = 0;
    }
    else
    {
      printf( "Recreating key in \"%s\"\n", keyFilename );
      quartz_generate_key( keyFilename, password );
      exitCode = 0;
    }

    pthread_mutex_unlock(&mutex_cmdline);
    return( exitCode );
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
