/********************************
 * Author: Matt Sun
 * Date:   February 23, 2005
 * 
 * Communication between nodes
 ********************************/

#include <stdio.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <string.h>
#include <pthread.h>
#include "../libsha1.h"
#include "qfs/user_data.h"
#include <chimera.h>

#define IP_MAIN  "128.111.40.219"
#define NOT_FOUND 1234567890
#define FORCE_COMMIT   1

/* all the different types of control messages */
#define CTRL_SERVICE             10 // maintenance and performance gathering
#define CTRL_HEARTBEAT           11
#define CTRL_LEAVE_REQ           12
#define CTRL_VERSION_REQ         13
#define CTRL_CFILE_REQ           14 // client request to retreive file/log
#define CTRL_CFILE_REQ_ACK_YES   15 // successful result for file/log retreival
#define CTRL_CFILE_REQ_ACK_NO    16 // failed result for file/log retreival
#define CTRL_CFILE_STORE         17 // client request to store file/log
#define CTRL_CFILE_STORE_ACK_YES 18 // successful result for file/log store
#define CTRL_CFILE_STORE_ACK_NO  19 // failed result for file/log store
#define CTRL_SFILE_REQ           20
#define CTRL_SFILE_RES           21
#define CTRL_SFILE_STORE         22
#define CTRL_SYNCH_REQ           23
#define CTRL_META_FILE_REQ       24
#define CTRL_META_REQ_YES        25
#define CTRL_META_REQ_NO         26
#define CTRL_META_STORE          27
#define CTRL_META_STORE_ACK_YES  28
#define CTRL_META_STORE_ACK_NO   29
#define CTRL_META_COPY           30 // replicating metadata
#define CTRL_UPDATE              31
#define CTRL_UPDATE_META         32
#define CTRL_UPDATE_STORE        33
#define CTRL_META_GROUP_REQ      34
#define CTRL_STATS_REQ           35 // request for usage statistics
#define CTRL_STATS_RES           36 // response from node for stats
#define CTRL_DEL_META            37 // delete the meta data for a file
#define CTRL_DEL_META_COPY       38 // delete copy nodes for meta
#define CTRL_DEL_FILE            39 // delete the actual file contents
#define CTRL_DEL_FILE_COPY       40 // delete copy nodes for the file
#define CTRL_DEL_META_GROUP      41 // delete the entire meta group
#define CTRL_DEL_META_GROUP_COPY 42 // delete the entire meta group
#define CTRL_DEL_FILE_GROUP      43 // delete the entire file group
#define CTRL_DEL_FILE_GROUP_COPY 44 // delete the entire file group

/* the states each node can be in for a particular set */
#define STATE_NULL             0
#define STATE_NORMAL           1
#define STATE_NEED_TO_CHANGE   2
#define STATE_READY_TO_UPDATE  3
#define STATE_UPDATED          4
#define STATE_READY_FOR_ACK    5

/* defining the ports used to send and receive */
#define QUARTZ_PORT 15235
#define QUARTZ_CHAT 101

/* maximum sizes used */
#define MAX_MSG_SIZE     500
#define MAX_NEIGHBOR_SET 5
#define MAX_FILE_SIZE    52256

/* other global pound defines */
#define NEW_TABLE 99
#define NUM_METAS  9
#define NUM_FILES  5
#define NUM_LOOK   7

/* message structure */
typedef struct _struct_message
{
  int type;
  char data[MAX_MSG_SIZE];
} _message_z;

typedef struct _struct_msg_key
{
  Key send_key; // for sending back
} _msg_key_z;

/* message for persistent data storage */
typedef struct _struct_msg_data
{
  uint32_t version;     // for appending to the end of a file name
  Key node_id;          // node that created this msg
  Key send_key;         // key for sending and route_lookups
  HASH user_id;  	// for the name of the directory
  HASH group_id;        // name of the group the file belongs to
  HASH file_name; 	// for specifying the name of the file
  uint32_t size;  	// size of data 
  uint8_t data[MAX_FILE_SIZE]; // log data
} _msg_data_z;

/* message for persistent data storage */
typedef struct _struct_msg_file
{
  uint32_t version;     // for appending to the end of a file name
  Key node_id;          // node that created this msg
  Key send_key;         // key representing the user_id
  char user_id[40];  	// for the name of the directory
  char group_id[40];    // name of the group the file belongs to
  char file_id[40]; 	// for specifying the name of the file
  uint32_t size;  	// size of data
  //  uint8_t* data;  	// log data
} _msg_file_z;


/* message for meta data storage */
typedef struct _struct_msg_meta_data
{
  uint32_t version;     // the most recent version number
  Key node_id;          // node that created this msg
  Key send_key;         // key representing the user_name
  Key group_key;        // key_makehash(group)
  char user_id[41];  	// for the name of the directory
  char file_name[96]; 	// for specifying the name of the file
  char group_id[41];    // name of the group the file belongs to
  uint32_t size;  	// size of TOTAl files in this group
} _msg_meta_z;

/* message for notifying an update */
typedef struct _struct_msg_update
{
  Key key;          // key pertaining to the node
  uint8_t joined;   // whether this node joined or left
  uint8_t total;    // total filled up (in case < 8)
  Key neighbors[9]; // keys returned from route_lookup
} _msg_update;

/* message for updating new META nodes */
typedef struct _struct_msg_meta_update
{
  Key key;                  // key that caused the detection
  Key node_joined;          // new node that joined the routing table
  Key node_left;            // node that just left the network
  Key neighbors[NUM_METAS]; // all hosts returned from route_lookup
} _msg_meta_update;

/* message for updating new FILE nodes */
typedef struct _struct_msg_file_update
{
  Key key;                  // key that caused the detection
  Key node_joined;          // new node that joined the routing table
  Key node_left;            // node that just left the network
  Key neighbors[NUM_FILES]; // all hosts returned from route_lookup
} _msg_file_update;

/* to hold incoming stats res */
typedef struct user_usage_z
{
  char username[64];
  Key devices[32];
  uint8_t totalDevices;
  _user_data_z data[32];
} user_usage;

/* heartbeat table list format */
typedef struct _struct_heartbeat_table
{
  Key user_id;                         /* hash of ID of originator */
  uint16_t pos;                        /* position inside table    */
  uint16_t state[MAX_NEIGHBOR_SET];    /* used to obtain majoriy   */
  Key neighbor[MAX_NEIGHBOR_SET];      /* known neighbor           */
  struct _struct_heartbeat_table *next;
} _hb_table_z;

// Global
extern _hb_table_z*  my_heartbeat;
extern Key myKey;
extern pthread_mutex_t mutex_table;
extern pthread_mutex_t mutex_oper;
extern pthread_mutex_t mutex_stats;
extern pthread_mutex_t mutex_chimera;
extern pthread_mutex_t mutex_update;
extern pthread_mutex_t mutex_upcall;

extern pthread_mutex_t mutex_cfile_s;
extern pthread_mutex_t mutex_cfile_r;
extern pthread_mutex_t mutex_sfile;
extern pthread_mutex_t mutex_meta_s;
extern pthread_mutex_t mutex_meta_r;
extern pthread_mutex_t mutex_metac;

extern int program_exited;
extern pthread_t pids[];
extern int num_proc;

extern int update_last_msg;
extern Key update_last_key;
extern Key update_left_key;
extern ChimeraHost update_left_host;

/* heartbeat message format */
typedef struct _struct_heartbeat_msg
{
  Key user_id;                        /* hash of ID of originator */
  uint16_t pos;                        /* position inside table    */
  uint16_t state[MAX_NEIGHBOR_SET];    /* used to obtain majoriy   */
  //  HASH neighbor[MAX_NEIGHBOR_SET];     /* known neighbor           */
  Key neighbor[MAX_NEIGHBOR_SET];      /* Chimera needs keys       */

  /* 
   * Heartbeat message format (Heartbeat Table):
   * [         USER_ID        ]
   * [ STATE | KNOWN_NEIGHBOR ] 1
   * [ STATE | KNOWN_NEIGHBOR ] 2
   * [ STATE | KNOWN_NEIGHBOR ] 3
   * [ STATE | KNOWN_NEIGHBOR ] 4
   * [ STATE | KNOWN_NEIGHBOR ] 5
   *
   */

} _hb_msg_z;

/* server file send message format */
typedef struct _struct_sfile_msg
{
  char content[MAX_FILE_SIZE];
  int size;
  char *file_name;

} _sfile_msg_z;


/* control message type function handlers */
void handler_hb_normal(_hb_msg_z new_heartbeat);
int  handler_stats_req(Key mykey);
int  handler_stats_res(_user_data_z my_data);
void handler_cfile_res(Message *message);
void handler_sfile_req(Message *message);
void handler_sfile_res(Message *message);

/* chimera specific functions */
void quartz_fwd(Key **kp, Message **mp, ChimeraHost **hp);
void quartz_del(Key *k, Message *m);
void quartz_update(Key *k, ChimeraHost *h, int joined);

/* quartz processes */
void chimera_setup( int join_another, char* _c_host );

/* other necessary function protocols */
int hashes_match(HASH first_hash, HASH second_hash);
int log_stats(_user_data_z data );

