/**********************************************
 * Version 1
 * Date: January 30, 2005
 *
 * Code for each device to run on the network
 *  - responsible for 1+ user(s)
 *  - will store all directory meta-data which
 *    is separated per userID or identity hash
 **********************************************/

#ifndef _DEVICE_H_
#define _DEVICE_H_

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <time.h>

#include "libsha1.h"
#include "chimera.h"
#include "../group.h"

/*  Defines  */

#define  USER_INFO      "user_info.qtz"
#define  USER_INFO_BAK  "user_info-bak.qtz"

#define  USERNAME_LEN	64
#define  USEREMAIL_LEN	64
#define  NODENAME_LEN	64
#define  NODEADDR_LEN	32
#define  FILENAME_LEN  128

#define LOG_SIGNON      "SIGNON"
#define LOG_SIGNOFF     "SIGNOFF"
#define LOG_SEND        "SEND"
#define LOG_RECV        "RECV"
#define LOG_FWD         "FWD"
#define LOG_UPDATE      "UPDATE"
#define LOG_STORE       "STORE"
#define LOG_RM_META     "RM_META"
#define LOG_RM_FILE     "RM_FILE"
#define LOG_GET         "GET"
#define LOG_ERROR       "ERROR"
#define LOG_INIT        "INIT"
#define LOG_JOINED      "JOINED"
#define LOG_LEFT        "LEFT"
#define LOG_GMETA       "GMETA"
#define LOG_META_YES    "META_YES"
#define LOG_FILE_YES    "FILE_YES"

#define LOG_CFILE       "CFILE"
#define LOG_SFILE       "SFILE"
#define LOG_CMETA       "CMETA"
#define LOG_SMETA       "SMETA"
#define LOG_NA          "N/A"


#define MAX_FILE_LEN   64
#define MAX_QUOTA    150000000

#define fixNull(x)       ((NULL != (x)) ? (x) : "")


/*  Variables  */

extern int  verboseLevel;
extern int  stable; // if this is set to 1, then it can proceed with
                    // the update upcall, otherwise it needs to wait
                    // until its own meta and file storage are in
                    // sync with that of the rest of the network
extern int  total_receiving; // total REQ_YES expected (for mutex_cmdline)
extern int  total_received;  // total REQ_YES received thus far

/* Struct definitions */

typedef unsigned char Hash[20];

typedef struct _struct_user_z
{
  char *user_ID;
  void *public_key;
  void *private_key;
  char *identity_hash;
  int reputation_upload;
  int reputation_download;
} _user_z;

typedef struct _struct_directory_z
{
  char type[3];         /* 'd' for directory, 'f' for file, or even
                         more specifically, a different char for
                         each file type (used by handlers) */
  char* srcName;
  char srcHash[20];     /* directory hash does not change */
  char* dstName;
  int mrVersion;        /* most recent version */
  int size;             /* to limit users to 100 MB of outward storage */
  char status;          /* whether file has been modified or not */
  size_t offset;        /* for in-memory data storage */
  struct _struct_directory_z *next; /* next directory mapping if not NULL */
  Hash nextEntry;
} _directory_z;

typedef struct _struct_log_z
{
  char type[3];         /* hopefully the same as directory's type */

  /*
  union
  {
    struct _struct_directory_z dir;
    struct _struct_file_z file;
  }
  */

  int size;
  struct _struct_log_z *nextLogEntry;
  void *content;      /* changes */
} _log_z;

typedef struct _struct_file_z
{
  _log_z *nextLogEntry;
  void *content;
  char *file_name;
  struct _struct_file_z *next; /* next file entry if not NULL */
} _file_z;

typedef struct _struct_table_entry_z
{
  char *group;
  _directory_z *directories;
  _file_z *files;
  struct _struct_table_entry_z *next;
} _table_entry_z;

typedef struct _struct_char_list_z
{
  char content[32];
  struct _struct_char_list_z *next;
} _char_list_z;


#define log_dir  _u._dir
#define log_file _u._file


// Device structure
typedef struct _struct_device_z
{
  char  userName[ USERNAME_LEN ];
  char  userEmail[ USEREMAIL_LEN ];
  char  password[ 16 ];
  HASH  userHash;
  Key   userKey;
  char  nodeName[ NODENAME_LEN ];
  char  nodeAddr[ NODEADDR_LEN ];
  HASH  nodeHash;
  Key   nodeKey;
  int numGroups;
  void*  inMemGroups;
  uint8_t  isValid;
} _device_z;

// Config file
typedef struct _struct_cfg_user_z
{
  char userName[USERNAME_LEN];
  char userEmail[USERNAME_LEN];
  char userHash[64];
  char nodeName[NODENAME_LEN];
  uint16_t groups; // number of groups user is syncing

} _cfg_user_z;

typedef struct _struct_cfg_group_z
{
  char name[USERNAME_LEN];
  uint16_t files;

} _cfg_group_z;

typedef struct _struct_cfg_file_z
{
  char name[FILENAME_LEN];
  HASH id;
  time_t lastChanged;
  uint32_t version;
  uint32_t size;

} _cfg_file_z;


// Device info
extern _device_z*  deviceInfo;

/* Meta-data types */

// directory links on local machine
//  -- stores the groups and files the user's device knows about
typedef struct _struct_meta_local_z
{
  char type;                    /* file type: (a) ascii, (b)binary */
  char file_name[97];           /* name of the file it pertains to */
  uint32_t version;             /* most recent version             */
  uint32_t size;                /* size of the file                */
  struct _struct_meta_local_z *next;
} _meta_local_z;

typedef struct _struct_info_group_z
{
  char group_id[41];            /* hashed name of the group        */
  _meta_local_z *files;         /* list of files within this group */
  struct _struct_info_group_z *next;
} _info_group_z;

typedef struct _struct_info_file_z
{
  char filename[97];            /* name of the file it pertains to */
  uint32_t version;             /* most recent version             */
  uint32_t size;                /* size of the file                */
  char *content;                /* content of the file in memory   */
  struct _struct_info_file_z *next;
} _info_file_z;

typedef struct _struct_store_group_z
{
  char groupname[MAX_FILE_LEN];
  Key send_key;
  _info_file_z *files;
} _store_group_z;

// groups information for the user replicated on user ID key nodes
//  -- for the user to query any changes to groups it knows about
typedef struct _struct_meta_user_z
{
  char user_id[41];                 /* user ID key                 */
  Key send_key;
  uint8_t pos;                      /* position for the user       */
  _info_group_z *groups;            /* list of groups for the user */
  struct _struct_meta_user_z *next; /* next user                   */
} _meta_user_z;

// groups information for files stored for users replicated on nodes
//  -- for nodes to know what they are storing and for whom
typedef struct _struct_meta_store_z
{
  char user_id[41];               /* user ID key                  */
  _store_group_z *groups;         /* list of groups for a user ID */
  struct _struct_meta_store_z *next;
} _meta_store_z;


/* external meta data variables */
extern _meta_user_z  *meta_user;
extern _meta_store_z *meta_store;
extern pthread_mutex_t mutex_mylog;
extern pthread_mutex_t mutex_cmdline;

/* Function prototypes */

_log_z get_most_recent_log(char *file_name);

/* main functions of the device */
//int sign_on(_table_entry_z *my_table, _device_z* devInfo );
int synchronize_file(_table_entry_z *my_table, char *file_name, char *group);
int synchronize_my_files(_table_entry_z *my_table);
int receive_incoming();
int update_logs();
int update_tables();
int update_config_files();
int update_all_files(char *name, char* hash, char *email, char *node,
		     char *ip_address, int _i_total_groups);
int send_modifications(char *user, char *identity_hash, char *file_name);
int shut_down(char *name, HASH hash, char *email, char *node,
	      char *ip_address, int _i_total_groups);

// log executions
extern void log_exec(char* log_type, Key key, char* packet_type, char* other);


//  Shutdown.
    int
shut_down0(
   void
);


//  Update all the files.
    int
update_all_files0(
   void
);

/*
// Connect to the network.
    int
quartz_signon(
  void
);
*/

// Synchronize the files.
    int
synchronize_files(
  void
);

// Get the location based on the QUARTZ_CONFIG/QUARTZ_HOME environment variable.
	extern char*
get_config_file(
    char *filename,
    char* buffer,
    int buffer_size
);

// Get the location based on the QUARTZ_HOME environment variable.
	extern char*
get_quartz_home(
    void
);


// Open (and possible create) a file.
    extern FILE*
open_and_create(
  char* filename,
  char* dirname,
  char* mode
);


// Open (and possible create) a file.
    extern int
device_main(
  int argc,
  char** argv
);


// Create a device structure.

   extern _device_z*
createDevice(
  char* userName,
  char* userEmail,
  char* userHash,
  char* nodeName,
  char* nodeAddr,
  int numGroups
  );

/*
    extern _device_z*
createDevice(
  char* userName,
  char* userEmail,
  char* userHash,
  char* nodeName,
  char* nodeAddr,
  int numGroups,
  _char_list_z  *all_groups
);
*/

// Display the device info. (Always returns true.)
    extern int
print_device_info(
  FILE* file,			// NULL --> stdout.
  _device_z* device_info
);


// Print the contents of a group.
    extern int
print_group(
  FILE* file,			// NULL --> stdout.
  char* group
);

// Get the current node identification (key || hash).
    extern char*
dev_getNodeId(
  void
);

// Is this node connected?
    extern char
dev_isConnected(
  void
);

#endif

/* vim:set shiftwidth=2: */
