//~~ group.h ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
/*
 *  In memory version of a synchronization group.
 *
 * @author:    Andy Pippin   <pippin@cs.ucsb.edu>
 * @version:   $Date: 2005/10/15 03:06:26 $   $Revision: 1.9 $
 * @requires:   
 */

//~~[ Identity ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#ifndef  __GROUP_H__
#define  __GROUP_H__

#ifdef __GROUP_C__
static char* hdrIdentity =
    "$Id: group.h,v 1.9 2005/10/15 03:06:26 mattsun Exp $";
#endif

#ifdef __cplusplus
extern "C" {
#endif

//~~[ Includes ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#include  <time.h>
#include  <stdint.h>

#include  "libsha1.h"


//~~[ Typedefs/Structures/Enums ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

typedef struct _in_memory_file_z
{
    char*    name;
    HASH     id;
    time_t   lastChanged;
    uint32_t version;
    uint8_t  type;
    uint8_t  status;
    uint32_t size;

    struct _in_memory_file_z*  next;
}
  INMEM_FILE;


typedef struct _in_memory_group_z
{
    char*      name;
    uint16_t   numFiles;
    uint32_t   totalSize; 

    INMEM_FILE*  fileList;
    struct _in_memory_group_z*  next;
}
  INMEM_GROUP;


//~~[ Constants ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#define  FILETYPE_Unknown  '?'
#define  FILETYPE_Binary   'b'
#define  FILETYPE_Text     't'

#define  FILE_STATUS_Okay	  0
#define  FILE_STATUS_Commit	  (1 << 1)
#define  FILE_STATUS_CommitError  (1 << 2)
#define  FILE_STATUS_Request	  (1 << 3)
#define  FILE_STATUS_RequestError (1 << 4)



//~~[ Variables ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//~~[ Prototypes ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//------------------------------------------------------------------------------
/*
 *  Create a new group.
 *
 * @param   groupname  The name of the group to add.
 * @return  Success (0), or failure (!0).
 */
        extern INMEM_GROUP*
group_new_group(
    char* groupname
);

//------------------------------------------------------------------------------
/*
 *  Remove a group.
 *
 * @param   groupname  The name of the group to remove.
 * @return  Success (0), or failure (!0).
 */
        extern INMEM_GROUP*
group_remove_group(
    char* groupname
);


//------------------------------------------------------------------------------
/*
 *  Get the in-memory group.
 *
 * @param   groupname  The group to get.
 * @return  The group, or NULL if not found.
 */
	extern INMEM_GROUP*
group_find_group(
    char* groupname
);


//------------------------------------------------------------------------------
/*
 *  Add a file to a group.
 *
 * @param   group    The group to add the file.
 * @param   filename The file.
 * @return  Success (0), or failure (!0)
 */
        extern int
group_add_file(
    INMEM_GROUP* group,
    char* filename
);

//------------------------------------------------------------------------------
/*
 *  Remove a file remove a group.
 *
 * @param   group    The group to remove the file.
 * @param   filename The file.
 * @return  Success (0), or failure (!0)
 */
        extern int
group_remove_file(
    INMEM_GROUP* group,
    char* filename
);


//------------------------------------------------------------------------------
/*
 *  Find a file based on the hash.
 *
 * @param   fileId  The object id.
 * @return  The file, or NULL if not found.
 */
        extern INMEM_FILE*
group_find_file(
    HASH*  fileId
);


//------------------------------------------------------------------------------
/*
 *  Find a file in a group.
 *
 * @param   group    The group to add the file.
 * @param   filename The name file.
 * @return  Success (0), or failure (!0)
 */
        extern INMEM_FILE*
group_get_file(
    INMEM_GROUP* group,
    char* filename
);


//------------------------------------------------------------------------------
/*
 *  Print a group.
 *
 * @param   file   The FILE* to display the group (NULL --> stdout)
 * @param   group  The group to print.
 * @return  True (!0).
 */
        extern int
group_print(
    FILE*  file,
    INMEM_GROUP* group
);


//------------------------------------------------------------------------------
/*
 *  Load all the groups.  Requires the deviceInfo struct.
 *
 * @param   devInfo  The device/user/group info structure.
 * @return  Number of groups loaded, or an error (<0).
 */
        extern int
group_load_all_groups(
    void
);


//------------------------------------------------------------------------------
/*
 *  Helper function.  Add the specified to the specified group.
 *
 * @param   filename   The name of the file to add.
 * @param   groupname  The name of the group to add the file.
 * @return  Success (0), or failure (!0).
 */
        extern INMEM_FILE*
group_addFile(
    char* filename,
    char* groupname
);


//------------------------------------------------------------------------------
/*
 *  Print the total memory this group takes up
 *
 * @param   INMEM_GROUP
 * @return  total size in bytes
 */
        extern int
group_total_size(
    INMEM_GROUP *group
);

//------------------------------------------------------------------------------
/*
 *  Helper function.  Add the specified to the specified group.
 *
 * @param   filename   The name of the file to add.
 * @param   groupname  The name of the group to add the file.
 * @return  Success (0), or failure (!0).
 */
	 void
file_print(
    FILE* out,
    INMEM_FILE* file
);


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#ifdef __cplusplus
}
#endif
#endif // __GROUP_H__
