//~~ log.c ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
/*
 *  description
 *
 * @author:    Andy Pippin   <pippin@cs.ucsb.edu>
 * @version:   $Date: 2005/10/25 19:40:54 $   $Revision: 1.34 $
 * @requires:   
 */

//~~[ Identity ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#define  __LOG_C__
static char* srcIdentity = 
    "$Id: log.c,v 1.34 2005/10/25 19:40:54 mattsun Exp $";


//~~[ Includes ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#include  <errno.h>
#include  <quartz_rsa.h>
#include  <stdlib.h>
#include  <stdint.h>
#include  <string.h>
#include  <sys/stat.h>
#include  <sys/types.h>
#include  <unistd.h>

#include  "diff.h"
#include  "device/device.h"
#include  "device/communication.h"
#include  "error.h"
#include  "group.h"
#include  "libsha1.h"
#include  "log.h"
#include  "misc.h"

#ifndef fixNull
#    define fixNull(x)  ((NULL!=(x))?(x):"<NULL>")
#endif


//~~[ Typedefs/Structures/Enums ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//~~[ Constants ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//~~[ Variables ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//~~[ Prototypes ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    //----------------------------------------------------------------------
    /*
     *  Display a LOG file entry.  Does not check to make sure that it 
     *  actually is a file entry - it assumed that the programmer is smart
     *  enough not to pass in a file log.
     *
     * @param   fileEntry  File LOG entry.
     * @param   outStream  Output location (NULL --> stdout).
     */
	    static void
    file_printEntry(
	LOG* fileEntry,
	FILE* outStream
    );

    //----------------------------------------------------------------------
    /*
     *  Marshal a file LOG entry into a stream of bytes.
     *
     * @param   fileEntry    The file LOG entry to marshal.
     * @param   bufferSize  Number of bytes produced.  // Output
     * @return  The marshalled entry.
     */
	    static uint8_t*
    file_encodeEntry(
	LOG*  fileEntry,
	uint32_t*  bytes
    );

    //----------------------------------------------------------------------
    /*
     *  Convert a marshalled log entry into a LOG file structure.
     *
     * @param   buffer      Buffer containing the marshalled log entry.
     * @param   bufferSize  Size of marshalled log entry.
     * @param   consumed    Number od bytes consumed.  // Output
     * @return  New log entry and bytes consumed.
     */
	    static LOG*
    file_decodeEntry(
	uint8_t*  buffer,
	uint32_t  bufferSize,
	uint32_t*  consumed  // Output
    );


    //----------------------------------------------------------------------
    /*
     *  Display a LOG directory entry.  Does not check to make sure that it 
     *  actually is a directory entry - it assumed that the programmer is
     *  smart enough not to pass in a file log.
     *
     * @param   dirEntry   Directory LOG entry.
     * @param   outStream  Output location (NULL --> stdout).
     */
	    static void
    dir_printEntry(
	LOG* dirEntry,
	FILE* outStream
    );

    //----------------------------------------------------------------------
    /*
     *  Marshal a directory LOG entry into a stream of bytes.
     *
     * @param   dirEntry    The directory LOG entry to marshal.
     * @param   bufferSize  Number of bytes produced.  // Output
     * @return  The marshalled entry.
     */
	    static uint8_t*
    dir_encodeEntry(
	LOG*  dirEntry,
	uint32_t*  bufferSize
    );

    //----------------------------------------------------------------------
    /*
     *  Convert a marshalled log entry into a LOG directory structure.
     *
     * @param   buffer      Buffer containing the marshalled log entry.
     * @param   bufferSize  Size of marshalled log entry.
     * @param   consumed    Number od bytes consumed.  // Output
     * @return  New log entry and bytes consumed.
     */
	    static LOG*
    dir_decodeEntry(
	uint8_t*  buffer,
	uint32_t  bufferSize,
	uint32_t*  consumed  // Output
    );

    //----------------------------------------------------------------------
    /*
     *  Convert raw bytes to a string and dumps it to a file.
     *
     * @param   out     Output file
     * @param   buffer  Buffer to dump.
     * @param   size    Number of bytes to dump.
     * @return  The byte-stream string. Must be freed when done.
     */
	    static char *
    dumpRawBytes(
	FILE* out,
	uint8_t*  buffer,
	uint32_t  size
    );

	//----------------------------------------------------------------------
    /*
     *  Read contents of a file into a string.  Then converts it to raw bytes.
     *
     * @param   in      Input file
     * @param   size    Number of bytes read from file.
     * @return  The raw bytes from the file.
     */
	    static uint8_t*
    readRawBytes(
	FILE* in,
	uint32_t  *size
    );


    //----------------------------------------------------------------------
    /*
     *  Allocate a new log entry, and copy in the parentId's hash.
     *
     * @param   parentId  The hash of the parent.
     * @return  A new log entry, or NULL if not enough memory.
     */
            static LOG*
    createLog(
        HASH  parentId
    );


    //----------------------------------------------------------------------
    /*
     * These string/hex functions taken from v2_utl.c in libcrypto
     * Reproduced to allow our program to customize if neccessary.
     *
     * @param str  Pointer to sequence of bytes.
     * @param len  Number of bytes.
     * returns  Sequence of unsigned chars.
     */
	    static char*
    my_hex_to_string(
	unsigned char *buffer,
	uint32_t len
    );


    //----------------------------------------------------------------------
    /*
     * Give a string of hex digits convert to a buffer.
     *
     * @param str  pointer to string to convert to bytes.
     * @param len  Size of decoded string.
     * returns  Sequence of unsigned chars.
     */
	    static unsigned char*
    my_string_to_hex(
	char *str,
	uint32_t* len
    );


//~~[ Implementation ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	void
log_print(
    LOG* logEntry,
    FILE* outStream
){
    FILE* out = (NULL != outStream) ? outStream : stdout;
    LOG* printEntry = logEntry;

    if( NULL != printEntry )
    {
        fprintf( out, "\n%s%s\n\n",
	    "====[ Log Entries ]=====================", 
	    "========================================" );

	int entry = 0;
        do
	{
	    fprintf( out, "%2d:", ++entry );
	    fflush( out );
	    switch( printEntry->logType )
	    {
	        case  LOG_File:
		    file_printEntry( printEntry, out );
		    break;

	        case  LOG_Dir:
		    dir_printEntry( printEntry, out );
		    break;
	    }

	    printEntry = printEntry->nextInner;

	    fputc( '\n', out );
	}
	while( NULL != printEntry );

        fprintf( out, "%s%s\n\n",
	    "========================================", 
	    "========================================" );
    }

    return;
}


//------------------------------------------------------------------------------

	void
log_delete(
    LOG* logEntry
){
    if( NULL != logEntry )
    {
        if( NULL != logEntry->objName )
	{
	    free( logEntry->objName );
	}

	if( LOG_File == logEntry->logType )
	{
	}
	else if( LOG_Dir == logEntry->logType )
	{
	    if( NULL != logEntry->u_dirEntry.name )
	    {
		free( logEntry->u_dirEntry.name );
	    }
	}

	free( logEntry );
    }

    return;
}


//------------------------------------------------------------------------------

	extern LOG*
log_createFullFile(
    HASH  parentId,
    char*  fileName,
    LOG*  prevInner
){
    LOG*  fileEntry =NULL;
    char  nameBuffer[ FILENAME_MAX ];
    struct stat statBuffer = { '\0' };
    FILE* in =NULL;

    if( NULL == fileName )
    {}
    else if( NULL == (fileEntry = createLog( parentId )) )
    {}
    else if( NULL == misc_getFileName( fileName, nameBuffer, FILENAME_MAX ) )
    {}
    else if( -1 == stat( nameBuffer, &statBuffer ) )
    {}
    else
    {
	fileEntry->logType = LOG_File;
	fileEntry->objName = strdup( fileName );
	fileEntry->nextInner = NULL;

	fileEntry->u_fileEntry.begin = (uint32_t)0;
	fileEntry->u_fileEntry.end = (uint32_t)UINT32_MAX;
	fileEntry->u_fileEntry.size = (uint32_t)statBuffer.st_size;
	fileEntry->u_fileEntry.data = (uint8_t*)malloc( statBuffer.st_size );

	if( NULL == fileEntry->u_fileEntry.data )
	{
	    error( __FILE__, __LINE__, "memory wasn't allocated.\n" );
	}
	else if( NULL == (in = fopen( nameBuffer, "r" )) )
	{
	    free( fileEntry->u_fileEntry.data );
	    error( __FILE__, __LINE__, "couldn't open \"%s\"\n", nameBuffer );
	}
	else
	{
	    size_t read = 0;
	    size_t left = statBuffer.st_size;
	    char buffer[ BUFSIZ ];
	    char* ptr = (char*)fileEntry->u_fileEntry.data;

	    while( 0 < (read = fread( buffer, sizeof(char), BUFSIZ, in )) )
	    {
	        memcpy( ptr, buffer, read );
		ptr += read;
	    }
	    fclose( in );
	}
    }

    return( fileEntry );
}


//------------------------------------------------------------------------------

	extern LOG*
log_createFileEntry(
    HASH  parentId,
    char*  fileName,
    LOG*  prevInner,
    uint32_t  begin,
    uint32_t  end,
    uint32_t  size,
    uint8_t*  content
){
    LOG*  fileEntry = NULL;

    if( NULL != (fileEntry = createLog( parentId )) )
    {
	fileEntry->logType = LOG_File;

	if( NULL != fileName )
	{
	    fileEntry->objName = strdup( fileName );
	}
	else
	{
	    fileEntry->objName = NULL;
	}
	fileEntry->nextInner = NULL;

	fileEntry->u_fileEntry.begin = begin;
	fileEntry->u_fileEntry.end = end;
	fileEntry->u_fileEntry.size = size;
	fileEntry->u_fileEntry.data = content;


	if( NULL != prevInner )
	{
	    // Link the two.
	    prevInner->nextInner = fileEntry;

	    // And return the first.
	    fileEntry = prevInner;
	}
    }

    return( fileEntry );
}


//------------------------------------------------------------------------------

	extern LOG*
log_createDirEntry(
    HASH  parentId,
    char*  dirName,
    LOG*  prevInner,
    DIR_ACTION  action,
    char*  objName,
    HASH  objId
){
    LOG*  dirEntry = NULL;

    if( NULL != (dirEntry = createLog( parentId )) )
    {
        dirEntry->logType = LOG_Dir;
	if( NULL != dirName )
	{
	    dirEntry->objName = strdup( dirName );
	}
	else
	{
	    dirEntry->objName = NULL;
	}
	dirEntry->nextInner = NULL;

	dirEntry->u_dirEntry.action = action;
	if( NULL != objName )
	{
	    dirEntry->u_dirEntry.name = strdup( objName );
	}
	else
	{
	    dirEntry->u_dirEntry.name = NULL;
	}
	memcpy( dirEntry->u_dirEntry.hashId, objId, sizeof(HASH) );

	if( NULL != prevInner )
	{
	    // Link the two.
	    prevInner->nextInner = dirEntry;

	    // And return the first.
	    dirEntry = prevInner;
	}
    }

    return( dirEntry );
}


//------------------------------------------------------------------------------

	LOG*
log_createFileDiff(
    HASH  parentId,
    char*  fileName,
    LOG*  prevInner,
    char* beforeFilename,
    char* afterFilename
){
    LOG*  fileEntry = NULL;

    if( (NULL != fileName)
     && (NULL != beforeFilename)
     && (NULL != afterFilename) )
    {
diff_locate( parentId, fileName, NULL, beforeFilename, afterFilename );
    }

    return( fileEntry );
}


//------------------------------------------------------------------------------
	
	uint8_t*
log_encode(
    LOG*  logEntry,
    uint32_t*  size
){
    LOG* encodeEntry = logEntry;
    uint8_t*  encoded = NULL;
    uint32_t  encodedSize = 0;

    if( (NULL != encodeEntry) && (NULL != size) )
    {
	uint8_t*  newData = NULL;
	uint32_t  newSize = 0;
	uint8_t*  reSize = NULL;

        do
	{
	    switch( encodeEntry->logType )
	    {
	        case  LOG_File:
		    newData = file_encodeEntry( encodeEntry, &newSize );
		    break;

	        case  LOG_Dir:
		    newData = dir_encodeEntry( encodeEntry, &newSize );
		    break;
	    }

	    // Got something, so add some space.
	    if( (NULL != newData) && (0 < newSize)
	     && (NULL != (reSize = realloc( encoded, (encodedSize+newSize) ))) )
	    {
	        encoded = reSize;
		memcpy( (encoded+encodedSize), newData, newSize );
		encodedSize += newSize;
	    }

	    encodeEntry = encodeEntry->nextInner;
	}
        while( NULL != encodeEntry );

	*size = encodedSize;
    }

    return( encoded );
}


//------------------------------------------------------------------------------
	
	LOG*
log_decode(
    uint8_t*  buffer,
    uint32_t  bufferSize 
){
    LOG*  newLog = NULL;
    uint8_t*  ptr = buffer;
    uint32_t  bufferLeft = bufferSize;
    uint32_t  consumed;

    if( (NULL != buffer) && (0 < bufferSize) )
    {
        LOG* entry;
	struct _log*  nextInner = (struct _log*)!0;

	do
	{
	    switch( *ptr )
	    {
		case  LOG_File:
		    entry = file_decodeEntry( ptr, bufferLeft, &consumed );
		    break;

		case  LOG_Dir:
		    entry = dir_decodeEntry( ptr, bufferLeft, &consumed );
		    break;

		default:
		    fflush( stdout );
		    fprintf( stderr, "Unknown log entry type: '%c'\n", *ptr );
		    fflush( stderr );
		    //
		    log_delete( newLog );
		    return( NULL );
		    // ***  DOES NOT CONTINUE  ***
	    }

	    nextInner = 0;
	    if( NULL != entry )
	    {
		if( NULL == newLog )
		{
		    newLog = entry;
		}
		else
		{
		    newLog->nextInner = entry;
		}

		// Another one?
	        nextInner = entry->nextInner;
	    }

	    bufferLeft -= consumed;
	    ptr += consumed;
	}
	while( NULL != nextInner );
    }

    return( newLog );
}


//------------------------------------------------------------------------------

	int
log_applyFileEntry(
    char* fileName,
    LOG*  logEntry
){
    int success = !0;
    FILE*  inputFile = NULL;
    FILE*  tmpFile = NULL;
    char*  tmpFileName = NULL;


    if( (NULL != fileName) && (NULL != logEntry) )
    {
	uint8_t  buffer[ BUFSIZ ];

	size_t  in;
	size_t  out;


        // Open the file to change.  Open it with write priviledges to make it
	// fail here if the file is not modifiable.
	if( NULL == (inputFile = fopen( fileName, "r+" )) )
	{
	    error( "Couldn't open \"%s\": %s\n",
			    fileName, strerror( errno ) );
	    success = 1;
	    goto  Exception;
	}

	// Create a temporary working file.
	char tmpFileNameTemplate[ FILENAME_MAX ];
	strcpy( tmpFileNameTemplate, "/tmp/quartz_applyFileEntry-XXXXXX" );
	int  fd;

	if( (0 > (fd = mkstemp( tmpFileNameTemplate )))
	 || (NULL == (tmpFile = fdopen( fd, "w+" ))) )
	{
	    error( "Couldn't open \"%s\": %s\n",
			    tmpFileNameTemplate, strerror( errno ) );
	    success = 2;
	    goto  Exception;
	}
	tmpFileName = tmpFileNameTemplate;


	// Put all the file that is not being changed.
	size_t  bytesToRead = (size_t)logEntry->u_fileEntry.begin;
	do
	{
	    in = fread( buffer, sizeof(uint8_t), bytesToRead, inputFile );
	    if( ferror( inputFile ) )
	    {
		error( "Error reading: \"%s\": %s\n",
				fileName, strerror( errno ) );
		success = 3;
		goto Exception;
	    }

	    fwrite( buffer, sizeof(uint8_t), in, tmpFile );
	    if( ferror( tmpFile ) )
	    {
		error( "Error writing: \"%s\": %s\n",
				tmpFileName, strerror( errno ) );
		success = 4;
		goto Exception;
	    }

	    bytesToRead -= in;
	}
	while( 0 < bytesToRead );


	// Put all the new data.
	size_t bytesToWrite = (size_t)logEntry->u_fileEntry.size;
	uint8_t*  newData = logEntry->u_fileEntry.data;
	while( 0 < bytesToWrite )
	{
	    out = fwrite( newData, sizeof(uint8_t), bytesToWrite, tmpFile );
	    if( ferror( inputFile ) )
	    {
		error( "Error writing: \"%s\": %s\n",
				tmpfile, strerror( errno ) );
		success = 5;
		goto Exception;
	    }

	    bytesToWrite -= out;
	    newData += out;
	}


	// Consume all the replaced data.
	bytesToRead = (size_t)logEntry->u_fileEntry.end
				 - (size_t)logEntry->u_fileEntry.begin;

	// If the start is 0 and the end is UINT32_MAX, we're replacing the
	// whole file.
	if( (size_t)UINT32_MAX != bytesToRead )
	{
	    do
	    {
		in = fread( buffer, sizeof(uint8_t), bytesToRead, inputFile );
		if( ferror( inputFile ) )
		{
		    error( "Error reading: \"%s\": %s\n",
				    fileName, strerror( errno ) );
		    success = 6;
		    goto Exception;
		}

		bytesToRead -= in;
	    }
	    while( 0 < bytesToRead );


	    // Add the rest of the file.
	    do
	    {
		in = fread( buffer, sizeof(uint8_t), BUFSIZ, inputFile );
		if( ferror( inputFile ) )
		{
		    error( "Error reading: \"%s\": %s\n",
				    fileName, strerror( errno ) );
		    success = 7;
		    goto Exception;
		}

		fwrite( buffer, sizeof(uint8_t), in, tmpFile );
		if( ferror( tmpFile ) )
		{
		    error( "Error writing: \"%s\": %s\n",
				    tmpFileName, strerror( errno ) );
		    success = 8;
		    goto Exception;
		}
	    }
	    while( 0 < in );
	}



	// Put the contents of the tmpFile into the inputFile.  Don't want to
	// copy, since I don't want to mess with file system issues.
	//
	fflush( tmpFile );
	rewind( tmpFile );
	rewind( inputFile );

	// Add the tmp file.
	do
	{
	    in = fread( buffer, sizeof(uint8_t), BUFSIZ, tmpFile );
	    if( ferror( tmpFile ) )
	    {
		error( "Error reading: \"%s\": %s\n",
				fileName, strerror( errno ) );
		success = 9;
		goto Exception;
	    }

	    fwrite( buffer, sizeof(uint8_t), in, inputFile );
	    if( ferror( inputFile ) )
	    {
		error( "Error writing: \"%s\": %s\n",
				tmpFileName, strerror( errno ) );
		success = 10;
		goto Exception;
	    }
	}
	while( 0 < in );

	// Close up shop.
	fclose( inputFile );
	inputFile = NULL;
	fclose( tmpFile );
	tmpFile = NULL;
    }

    // If the tmpFileName is not null, it was probably created.
    if( NULL != tmpFileName)
    {
	if(-1 == unlink( tmpFileName ) )
	{
	    error( "Error deleting: \"%s\": %s\n",
			    tmpFileName, strerror( errno ) );
	    success = 11;
	    goto Exception;
	}
    }

// Because C doesn't have exceptions....
Exception:
    if( NULL != inputFile )
    {
        fclose( inputFile );
	inputFile = NULL;
    }
    if( NULL != tmpFile )
    {
        fclose( tmpFile );
	tmpFile = NULL;
    }

//debug( __FILE__, __LINE__, "Exiting log_applyFileEntry() = %d\n", success );
    return( success );
}


//------------------------------------------------------------------------------

	extern uint8_t*
log_encrypt(
    LOG*  logEntry,
    uint32_t*  size  // Output
){
    uint8_t* result = NULL;

    if( (NULL != logEntry) && (NULL != size) )
    {
	// Encode it.
	char* rawLog = NULL;
	char* keyFilename = NULL;
	char* keyPassword = NULL;

	uint32_t encodedLen = 0;
	uint8_t* encodedLog = log_encode( logEntry, &encodedLen );
//debug(  __FILE__, __LINE__, "log_encrypt()'encodedLog =\n", );
//dumpBytes( stdout, encodedLog, 16, "        " );


	if( (NULL != encodedLog)
	 && (0 != encodedLen)
	 && (NULL != (keyFilename = misc_getKeyFilename()))
	 && (NULL != (keyPassword = misc_getKeyPassword())) )
	{
	    uint8_t* encryptedLog = NULL;
	    uint32_t encryptedLen = encodedLen;

	    int success = quartz_encrypt_string( keyFilename, keyPassword,
		    encodedLog, (char**)&encryptedLog, &encryptedLen );

	    if( (NULL != encryptedLog) && (0 != encryptedLen) )
	    {
//debug( __FILE__, __LINE__, "log_encrypt()'encryptedLog =\n" );
//dumpBytes( stdout, encryptedLog, 16, "        " );
	        result = encryptedLog;
		*size = encryptedLen;
	    }
	}

	// Clean up.
	if( NULL != encodedLog )
	    free( encodedLog );
    }

    return( result );
}


//------------------------------------------------------------------------------

	extern LOG*
log_decrypt(
    uint8_t*  encryptedLog,
    uint32_t  encryptedLen 
){
    LOG*  result = NULL;
    char* keyFilename = NULL;
    char* keyPassword = NULL;

    if( (NULL != encryptedLog)
     && (0 < encryptedLen)
     && (NULL != (keyFilename = misc_getKeyFilename()))
     && (NULL != (keyPassword = misc_getKeyPassword())) )
    {
      uint8_t*  encodedLog   = NULL;
      uint32_t  encodedLen   = encryptedLen;

      //debug( __FILE__, __LINE__, "log_decrypt()'encryptedLog =\n" );
      //dumpBytes( stdout, encryptedLog, 32, "        " );
      int success = quartz_decrypt_string( keyFilename, keyPassword,
					   (char*)encryptedLog, (char**)&encodedLog, &encodedLen );
      fprintf(stderr, "DECRYPTED STRING ** %s **\n", encodedLog);
      
      //debug( __FILE__, __LINE__, "log_decrypt()'encodedLog =\n" );
      //dumpBytes( stdout, encodedLog, 32, "        " );

	if( 0 == success )
	{
	    result = log_decode( encodedLog, encodedLen );
	}
    }

    return( result );
}


//------------------------------------------------------------------------------

	void
log_store_response(
    HASH*     fileId,
    uint32_t  version,
    uint32_t  size,
    uint8_t*  data,
    uint8_t   status
){
    INMEM_FILE* file = NULL;
    LOG* log = NULL;

    if( (status)
     && (NULL != (file = group_find_file( fileId ))) )
    {
        if( (file->version < version)
	    && (0 < size)
	    && (NULL != data))
	    //&& (NULL != (log = log_decrypt( data, size ))) )
	{
	  fprintf(stderr, "LOG_STORE_RESPONSE FOUND!\n");

	  if( 0 == log_applyFileEntry( file->name, log ))
	  {
	    file->version = version;
	    file->lastChanged = time(0);
	  }
	}
    }
}

//------------------------------------------------------------------------------

	void
log_store_update(
    HASH*     fileId,
    uint32_t  version,
    uint8_t*  data,
    uint32_t  size
){
    INMEM_FILE* file = NULL;
    LOG* log = NULL;
    struct stat  statBuffer;
    char  nameBuffer[ FILENAME_MAX ];

    if( (NULL != (file = group_find_file( fileId ))) )
    {
      if( (0 < size) && (NULL != data))
      {
	if (size > MAX_FILE_SIZE)
	{
	  printf("  ** File size (%d) exceeds the allowed maximum (%d)\n", size, MAX_FILE_SIZE);
	  printf("  ** Will not update file \"%s\"\n", file->name);
	}
	else
	{
	  if (user_get_file_size() + size <= MAX_QUOTA)
	  {
	    FILE* ffile;
	    ffile = fopen(file->name, "w");
	    
	    if (ffile != NULL)
	    {
	      fwrite(data, 1, size, ffile);
	      fclose(ffile);
	      
	      file->version = version;
	      
	      misc_getFileName( file->name, nameBuffer, FILENAME_MAX );
	      stat(nameBuffer, &statBuffer);
	      
	      memcpy(&file->lastChanged, &statBuffer.st_mtime, sizeof(statBuffer.st_mtime));
	      file->size = statBuffer.st_size;
	      
	      update_config_files();
	      //user_incr_file_size(statBuffer.st_size, 1);
	      
	      printf("Updated \"%s\"\n", file->name);
	    }
	    else
	      printf(" ** File \"%s\" cannot be opened!\n", file->name);
	  }
	  else
	  {
	    printf("  ** Max quota of 150MB breached!\n");
	    printf("  ** Will not update file \"%s\"\n", file->name);
	  }
	}
      }
      else
      {
	if (size == 0)
	  printf(" File is blank.\n");
      }
    }
}


//------------------------------------------------------------------------------

	void
log_store_ack(
    HASH*     fileId,
    uint32_t  version,
    uint8_t   status
){
    INMEM_FILE* file = NULL;
    struct stat  statBuffer;
    if( (status)
     && (NULL != (file = group_find_file( fileId ))) )
    {
	file->status &= ~FILE_STATUS_Commit;
	file->version = version;

	char fileName[ FILENAME_MAX ];
	snprintf( fileName, FILENAME_MAX, "%s/%s",
				    get_quartz_home(), file->name );

	if( 0 == stat( fileName, &statBuffer ) )
	{
	    file->lastChanged = statBuffer.st_mtime;
	}
    }

    return;
}


//------------------------------------------------------------------------------

	int
log_test(
    void
){
    int test = 0;

    test++;
    HASH  parent;
    sha1_stringToHash( "FA26BE19DE6BFF93F70BC2308434E4A440BBAD02", &parent );
    char*  fileName = "filename";
    LOG*  nextInner = NULL;
    uint32_t begin = (uint32_t)10;
    uint32_t end = (uint32_t)13;
    uint32_t size = (uint32_t)21;
    uint8_t*  content = (uint8_t*)"this is 21 characters";
    LOG* fileEntry;
    if( NULL == (fileEntry = log_createFileEntry(
		    parent, fileName, nextInner, begin, end, size, content )) )
    {
	printf( "Error creating file entry:  fileEntry = NULL\n" );
	goto End_LogTest;
    }
printf( "\n" );
log_print( fileEntry, stdout );
printf( "\n" );


    test++;
    char* dirName = "dirname";
    nextInner = fileEntry;
    DIR_ACTION dirAction = DIR_Add;
    char*  objName = "newfile";
    HASH object;
    sha1_stringToHash( "DEDA74F9F76174D7F266343054BD48AB3F8213D5", &object );
    LOG* dirEntry;
    if( NULL == (dirEntry = log_createDirEntry(
		    parent, dirName, nextInner, dirAction, objName, object )) )
    {
	printf( "Error creating dir entry:  dirEntry = NULL\n" );
	goto End_LogTest;
    }

/*
    test++;
    printf( "\n" );
    printf( "\n" );
    log_print( dirEntry, stdout );
*/


    test++;
    uint32_t bytes = 0;
    uint8_t* data = log_encode( dirEntry, &bytes );
    if( NULL == data )
    {
	printf( "Error encoding log entry:  data = NULL\n" );
	goto End_LogTest;
    }
    //
    int expectedDataSize = 123;
    uint8_t expectedData[] =
	{
	    0x66, 0xFA, 0x26, 0xBE, 0x19, 0xDE, 0x6B, 0xFF,
	    0x93, 0xF7, 0x0B, 0xC2, 0x30, 0x84, 0x34, 0xE4,
	    0xA4, 0x40, 0xBB, 0xAD, 0x02, 0x66, 0x69, 0x6C,
	    0x65, 0x6E, 0x61, 0x6D, 0x65, 0x00, 0x00, 0x00,
	    0x00, 0x0A, 0x00, 0x00, 0x00, 0x0D, 0x00, 0x00,
	    0x00, 0x15, 0x74, 0x68, 0x69, 0x73, 0x20, 0x69,
	    0x73, 0x20, 0x32, 0x31, 0x20, 0x63, 0x68, 0x61,
	    0x72, 0x61, 0x63, 0x74, 0x65, 0x72, 0x73, 0x2B,
	    0x64, 0xFA, 0x26, 0xBE, 0x19, 0xDE, 0x6B, 0xFF,
	    0x93, 0xF7, 0x0B, 0xC2, 0x30, 0x84, 0x34, 0xE4,
	    0xA4, 0x40, 0xBB, 0xAD, 0x02, 0x64, 0x69, 0x72,
	    0x6E, 0x61, 0x6D, 0x65, 0x00, 0x61, 0x6E, 0x65,
	    0x77, 0x66, 0x69, 0x6C, 0x65, 0x00, 0xDE, 0xDA,
	    0x74, 0xF9, 0xF7, 0x61, 0x74, 0xD7, 0xF2, 0x66,
	    0x34, 0x30, 0x54, 0xBD, 0x48, 0xAB, 0x3F, 0x82,
	    0x13, 0xD5, 0x00,
	};
    //
    if( (expectedDataSize != bytes)
     || 0 != memcmp( expectedData, data, expectedDataSize ) )
    {
	printf( "Error encoding log entry:  does not match expected\n" );
	goto End_LogTest;
    }

    // encryption test on files (see test below for encrypting bytes)
/*
    test++;
    printf( "\nGenerating plaintext file from byte-stream string\n" );
    FILE *fp = fopen("rsa_test_plaintext", "w");
    char *rawString = dumpRawBytes(fp, data, bytes);
    fclose(fp);
    printf( "Generating key, encrypted and decrypted files\n" );
    char *pw = "mypassword";
    quartz_generate_key("rsa_test_keyfile", pw);
    quartz_encrypt_file("rsa_test_keyfile", pw, "rsa_test_plaintext",
	    "rsa_test_encrypted");
    quartz_decrypt_file("rsa_test_keyfile", pw, "rsa_test_encrypted",
	    "rsa_test_decrypted");
    printf( "Reading data from the decrypted file\n" );
    fp = fopen("rsa_test_decrypted", "r");
    uint32_t newBytes = 0;
    uint8_t* newData = readRawBytes(fp, &newBytes);
    fclose(fp);   
    //
    if( (newBytes != bytes)
     || (0 != memcmp( newData, data, newBytes )) )
    {
	printf( "%s:  %s\n",
	    "Error with log entry encryption test",
	    "does not match original\n" );
	goto End_LogTest;
    }
    // And clean up.
    if (rawString) free(rawString);
    if (newData) free(newData);
*/

    test++;
    LOG* decodedLog = log_decode( data, bytes );
    if( NULL == decodedLog )
    {
	printf( "Error decoding log entry:  decodedLog is NULL.\n" );
	goto End_LogTest;
    }



    test++;
    char* rawLog = my_hex_to_string((char *)expectedData, expectedDataSize);
    if( NULL == rawLog )
    {
        printf( "Couldn't convert log data to encrypted.\n" );
	goto End_LogTest;
    }
    char keyFilename[ FILENAME_MAX ];
    snprintf( keyFilename, FILENAME_MAX, "%s/.quartz/key", getenv( "HOME" ) );
    char *password = misc_getKeyPassword();
    struct stat statBuffer;

    if( (-1 == stat( keyFilename, &statBuffer )) && (ENOENT == errno) )
    {
	printf( "Creating keypair: %s\n", keyFilename );
	quartz_generate_key(keyFilename, password);
    }

    test++;
    //LOG* fileEntry;
    uint32_t  encryptedLen;
    uint8_t* encryptedLog = log_encrypt( fileEntry, &encryptedLen );
    printf( "\nEncrypted log entry:  (%d bytes)\n", encryptedLen );
    printf( "\nEncrypted log = 0x%x\n", encryptedLog );
//    dumpBytes( stdout, encryptedLog, encryptedLen, "        " );

    LOG* sameLog = log_decrypt( encryptedLog, encryptedLen );
    printf( "\nDecrypted log entry:  (sameLog = 0x%x)\n", sameLog );
//    log_print( sameLog, stdout );

    log_delete( sameLog );
    free( encryptedLog );
    sameLog = NULL;
    encryptedLog = NULL;




/*
    printf( "\n" );
    printf( "\n" );
    log_print( decodedLog, stdout );
*/

    test++;
    int apply;
    if( 0 == (apply = log_applyFileEntry( "testapply", fileEntry )) )
    {
	printf( "Error applying log entry.  (%d)\n", apply );
	goto End_LogTest;
    }

/*
    test++;
    HASH  locHash;
    sha1_stringToHash( "DEDA74F9F76174D7F266343054BD48AB3F8213D5", &locHash );
    char* locFilename = "locate";
    LOG*  locLogEntry = NULL;
    locLogEntry = log_createFileDiff( locHash, locFilename, NULL,
					    "diff.before", "diff.after" );
*/

    test++;
    char nameBuffer[ 80 ];
    sha1_stringToHash( "2D7D86A945035949036B74463BF0347A80194E72", &parent );
    LOG* fullFile = log_createFullFile( parent, "fake_bashrc", NULL );
//    log_print( fullFile, stdout );
    uint32_t  ffenLen;
    uint8_t*  ffenLog;
    ffenLog = log_encrypt( fullFile, &ffenLen );
//    printf( "\nEncrypted full file = 0x%x  (%lu bytes)\n", ffenLog, ffenLen);


    test++;
    group_load_all_groups();
    log_store_response( &parent, 2, ffenLen, ffenLog, 1 );


    test = 0;
End_LogTest:

    if( 0 == test )
    {
        printf( "Test \"log\" passed.\n" );
    }
    else
    {
        printf( "Test \"log\" failed on test %d.\n", test );
    }

    return( test );
}


//~~ Static Functions ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	static void
file_printEntry(
    LOG* fileEntry,
    FILE* out
){
    fprintf( out, " File Log Entry\n" );
//    fprintf( out, "  File Entry (0x%08x)\n", fileEntry );
//    fflush( out );

    fprintf( out, "      parentId:  %s\n",
				sha1_hashToString( &(fileEntry->parentId) ) );
    fprintf( out, "      objName:   \"%s\"\n", fixNull( fileEntry->objName ) );
//    fprintf( out, "      nextInner: 0x%08x\n", fileEntry->nextInner );

    fprintf( out, "      begin:     %lu\n", (long)fileEntry->u_fileEntry.begin );
    fprintf( out, "      end:       %lu\n", (long)fileEntry->u_fileEntry.end );
    fprintf( out, "      size:      %lu\n", (long)fileEntry->u_fileEntry.size );
    fprintf( out, "      content:   0x%08x\n", fileEntry->u_fileEntry.data );

    dumpBytes( out, fileEntry->u_fileEntry.data,
	    fileEntry->u_fileEntry.size, "          " );

    return;
}


//------------------------------------------------------------------------------

	static uint8_t*
file_encodeEntry(
    LOG*  fileEntry,
    uint32_t*  bytes
){
    uint8_t* encoded;
    uint32_t  numBytes = 0;
    int objNameLen =0;
    
    // Calculate the byte count.

    // log type.
    numBytes++;

    // parentId.
    numBytes += sizeof(HASH);

    // object name.
    if( NULL != fileEntry->objName )
    {
	numBytes += (objNameLen = strlen( fileEntry->objName ));
    }
    numBytes++; // Just put a null

    // begin, end, and buffer size.
    numBytes += 3*sizeof(uint32_t);

    // content.
    numBytes += (uint32_t)fileEntry->u_fileEntry.size;

    // Next?
    numBytes++;

    // Try and get the buffer.
    if( NULL != (encoded = (uint8_t*)malloc( numBytes )) )
    {
        uint8_t*  ptr = encoded;
	long  net;

	*(ptr++) = (char)LOG_File;

	memcpy( ptr, fileEntry->parentId, sizeof(HASH) );
	ptr += sizeof(HASH);

	if( 0 != objNameLen )
	{
	    strcpy( ptr, fileEntry->objName );
	    ptr += objNameLen;
	}
	*(ptr++) = '\0';

	uint32_t netSafe;
	//
	netSafe = htonl(fileEntry->u_fileEntry.begin);
	memcpy( ptr, &netSafe, sizeof(uint32_t) );
	ptr += sizeof(uint32_t);
	//
	netSafe = htonl(fileEntry->u_fileEntry.end);
	memcpy( ptr, &netSafe, sizeof(uint32_t) );
	ptr += sizeof(uint32_t);
	//
	netSafe = htonl(fileEntry->u_fileEntry.size);
	memcpy( ptr, &netSafe, sizeof(uint32_t) );
	ptr += sizeof(uint32_t);

	memcpy( ptr, fileEntry->u_fileEntry.data, fileEntry->u_fileEntry.size );
	ptr += fileEntry->u_fileEntry.size;

	if( NULL != fileEntry->nextInner )
	{
	    *(ptr++) = '+';
	}
	else
	{
	    *(ptr++) = '\0';
	}

	*bytes = numBytes;

/*
printf( "\nDEBUG [%s:%d] File byte stream:  (%lu bytes)\n", __FILE__, __LINE__, numBytes );
dumpBytes( stdout, encoded, numBytes, "        " );
//
uint32_t  consumed;
(void)file_decodeEntry( encoded, numBytes, &consumed );
*/
    }

    return( encoded );
}


//------------------------------------------------------------------------------

	static LOG*
file_decodeEntry(
    uint8_t*  buffer,
    uint32_t  bufferSize,
    uint32_t*  consumed  // Output
){
    LOG*  newEntry = NULL;
    uint8_t* ptr = (uint8_t*)buffer;
    uint32_t  dataLen;

    if( (NULL != ptr) && ( 0 < bufferSize) && (NULL != consumed)
     && (NULL != (newEntry = createLog( NULL ))) )
    {
        newEntry->logType = LOG_File;
	++ptr;

	memcpy( newEntry->parentId, ptr, sizeof(HASH) );
	ptr += sizeof(HASH);

	int objNameLen = strlen( ptr );
	if(NULL != (newEntry->objName = (char*)malloc( objNameLen + 1 )) )
	{
	    memcpy( newEntry->objName, ptr, (size_t)(objNameLen + 1) );
	}
	ptr += (objNameLen + 1);

	uint32_t netSafe;
	//
	memcpy( &netSafe, ptr, sizeof(uint32_t) );
	newEntry->u_fileEntry.begin = ntohl( netSafe );
	ptr += sizeof(uint32_t);
	//
	memcpy( &netSafe, ptr, sizeof(uint32_t) );
	newEntry->u_fileEntry.end = ntohl( netSafe );
	ptr += sizeof(uint32_t);
	//
	memcpy( &netSafe, ptr, sizeof(uint32_t) );
	dataLen = newEntry->u_fileEntry.size = ntohl( netSafe );
	ptr += sizeof(uint32_t);

	if( NULL != (newEntry->u_fileEntry.data = (uint8_t*)malloc( dataLen )) )
	{
	    memcpy( newEntry->u_fileEntry.data, ptr, dataLen );
	}
	ptr += dataLen;

	newEntry->nextInner = NULL;
	if( '+' == (char)*((char*)ptr) )
	{
	    newEntry->nextInner = (struct _log*)!NULL;
	}
	ptr++;

	*consumed = (void*)ptr - (void*)buffer;
/*
printf( "\nDEBUG [%s:%d] File log entry:\n", __FILE__, __LINE__ );
file_printEntry( newEntry, stdout );
*/
    }

    return( newEntry );
}


//------------------------------------------------------------------------------

	static void
dir_printEntry(
    LOG* dirEntry,
    FILE* out
){
    fprintf( out, " Directory Log Entry\n" );
//    fprintf( out, "  Dir Entry (0x%08x)\n", dirEntry );
//    fflush( out );

    fprintf( out, "      parentId:  %s\n",
				sha1_hashToString( &(dirEntry->parentId) ) );
    fprintf( out, "      objName:   \"%s\"\n", fixNull( dirEntry->objName ) );
//    fprintf( out, "      nextInner: 0x%08x\n", dirEntry->nextInner );

    char* action;
    switch( dirEntry->u_dirEntry.action ) 
    {
        case  DIR_Add:	action = "DIR_Add";	break;
        case  DIR_Del:	action = "DIR_Delete";	break;
        case  DIR_Chg:	action = "DIR_Change";	break;
	default:	action = "DIR_Unknown";	break;
    }

    fprintf( out, "      action:    %s\n", action );
    fprintf( out, "      name:      %s\n",
			fixNull( dirEntry->u_dirEntry.name ) );
    fprintf( out, "      hashId:    %s\n",
			sha1_hashToString( &(dirEntry->u_dirEntry.hashId) ) );

    fflush( out );
    return;
}


//------------------------------------------------------------------------------

	static uint8_t*
dir_encodeEntry(
    LOG*  dirEntry,
    uint32_t*  bytes
){
    uint8_t* encoded = NULL;
    uint32_t numBytes = 0;
    int  objNameLen = 0;
    int  nameLen = 0;

    // Calculate the byte count.

    // log type;
    numBytes++;

    // parent id.
    numBytes += sizeof(HASH);

    // object name.
    if( NULL != dirEntry->objName )
    {
        numBytes += (objNameLen = strlen( dirEntry->objName ));
    }
    numBytes++;  // for null.

    // action code.
    numBytes++;

    // name
    if( NULL != dirEntry->u_dirEntry.name )
    {
        numBytes += (nameLen = strlen( dirEntry->u_dirEntry.name ));
    }
    numBytes++;  // for null.

    // hash
    numBytes += sizeof(HASH);

    // next?
    numBytes++;

    if( NULL != (encoded = (uint8_t*)malloc( numBytes )) )
    {
        uint8_t*  ptr = encoded;

	*(ptr++) = (char)LOG_Dir;

	memcpy( ptr, dirEntry->parentId, sizeof(HASH) );
	ptr += sizeof(HASH);

	if( 0 != objNameLen )
	{
	    strcpy( ptr, dirEntry->objName );
	    ptr += objNameLen;
	}
	*(ptr++) = '\0';

	*(ptr++) = (char)dirEntry->u_dirEntry.action;

	if( 0 != nameLen )
	{
	    strcpy( ptr, dirEntry->u_dirEntry.name );
	    ptr += nameLen;
	}
	*(ptr++) = '\0';

	memcpy( ptr, dirEntry->u_dirEntry.hashId, sizeof(HASH) );
	ptr += sizeof(HASH);

	if( NULL != dirEntry->nextInner )
	{
	    *(ptr++) = '+';
	}
	else
	{
	    *(ptr++) = '\0';
	}

	*bytes = numBytes;


/*
printf( "\nDEBUG [%s:%d] Directory byte stream:  (%lu bytes)\n", __FILE__, __LINE__, numBytes );
dumpBytes( stdout, encoded, numBytes, "        " );
//
uint32_t  consumed;
(void)dir_decodeEntry( encoded, numBytes, &consumed );
*/
    }


    return( encoded );
}


//------------------------------------------------------------------------------

	static LOG*
dir_decodeEntry(
    uint8_t*  buffer,
    uint32_t  bufferSize,
    uint32_t*  consumed  // Output
){
    LOG*  newEntry = NULL;
    uint8_t* ptr = (uint8_t*)buffer;
    uint32_t  dataLen;

    if( (NULL != ptr) && ( 0 < bufferSize) && (NULL != consumed)
     && (NULL != (newEntry = createLog( NULL ))) )
    {
        newEntry->logType = LOG_Dir;
	++ptr;

	memcpy( newEntry->parentId, ptr, sizeof(HASH) );
	ptr += sizeof(HASH);

	int objNameLen = strlen( ptr );
	if(NULL != (newEntry->objName = (char*)malloc( objNameLen + 1 )) )
	{
	    memcpy( newEntry->objName, ptr, (size_t)(objNameLen + 1) );
	}
	ptr += (objNameLen + 1);

	newEntry->u_dirEntry.action = (DIR_ACTION)*(ptr++);

	int nameLen = strlen( ptr );
	if(NULL != (newEntry->u_dirEntry.name = (char*)malloc( nameLen + 1 )) )
	{
	    memcpy( newEntry->u_dirEntry.name, ptr, (size_t)(nameLen + 1) );
	}
	ptr += (nameLen + 1);

	memcpy( newEntry->u_dirEntry.hashId, ptr, sizeof(HASH) );
	ptr += sizeof(HASH);

	newEntry->nextInner = NULL;
	if( '+' == (char)*((char*)ptr) )
	{
	    newEntry->nextInner = (struct _log*)!NULL;
	}
	ptr++;

	*consumed = (void*)ptr - (void*)buffer;
/*
printf( "\nDEBUG [%s:%d] Decoded directory:\n", __FILE__, __LINE__ );
dir_printEntry( newEntry, stdout );
*/
    }
    return( newEntry );
}



// -----------------------------------------------------------------

/* These string/hex functions taken from v2_utl.c in libcrypto
   Reproduced to allow our program to customize if neccessary */

	static char
*my_hex_to_string(
    unsigned char *buffer,
    uint32_t len
){
        char *tmp, *q;
        unsigned char *p;
        int i;
        static char hexdig[] = "0123456789ABCDEF";
        if(!buffer || !len) return NULL;
        if(!(tmp = malloc(len * 3 + 1))) {
	  fprintf(stderr, "malloc error\n");
	  return NULL;
        }
        q = tmp;
        for(i = 0, p = buffer; i < len; i++,p++) {
                *q++ = hexdig[(*p >> 4) & 0xf];
                *q++ = hexdig[*p & 0xf];
                *q++ = ':';
        }
        q[-1] = 0;

        return tmp;
}

/**
 * Give a string of hex digits convert to a buffer.
 *
 * @param str  pointer to string to convert to bytes.
 * @param len  Size of decoded string.
 * returns  Sequence of unsigned chars.
 */
	static unsigned char*
my_string_to_hex(
    char *str,
    uint32_t* len
){
        unsigned char *hexbuf, *q;
        unsigned char ch, cl, *p;
        if(!str) {
	  return NULL;
        }
        if(!(hexbuf = malloc(strlen(str) >> 1))) goto err;
        for(p = (unsigned char *)str, q = hexbuf; *p;) {
                ch = *(p++);

                if(ch == ':') continue;
                cl = *(p++);

                if(!cl) {
                        free(hexbuf);
                        return NULL;
                }
                if(isupper(ch)) ch = tolower(ch);
                if(isupper(cl)) cl = tolower(cl);

                if((ch >= '0') && (ch <= '9')) ch -= '0';
                else if ((ch >= 'a') && (ch <= 'f')) ch -= 'a' - 10;
                else goto badhex;

                if((cl >= '0') && (cl <= '9')) cl -= '0';
                else if ((cl >= 'a') && (cl <= 'f')) cl -= 'a' - 10;
                else goto badhex;
                *q++ = (ch << 4) | cl;
        }
        if(len) *len = q - hexbuf;

        return hexbuf;

        err:
        if(hexbuf) free(hexbuf);
        fprintf(stderr, "error\n");
        return NULL;

        badhex:
        free(hexbuf);
        fprintf(stderr, "error\n");
        return NULL;

}

//------------------------------------------------------------------------------

	static char *
dumpRawBytes(
    FILE* out,
    uint8_t*  buffer,
    uint32_t  size
){
    register char*  ptr = (char*)buffer;
    register uint32_t  count = size;
    register uint32_t  ch;

    char *rawString = my_hex_to_string((char *)buffer, size);
    
	fprintf(out, "%s", rawString);

    return rawString;
}

//------------------------------------------------------------------------------

        static uint8_t*
readRawBytes(
    FILE* in,
    uint32_t  *size
){
	  int i;
	  uint8_t *rawBytes;
	  char tempString[3];
	  uint32_t numBytes = 0;

	  while (fgets(tempString, 4, in) != NULL){
	    numBytes++;
	  }

	  rewind(in);
	  char newString[numBytes*3];
	  fgets(newString, (numBytes*3)+1, in);
	  
	  rawBytes = (uint8_t *)my_string_to_hex(newString, NULL);
	  *size = numBytes;

	  return rawBytes;
	  
}

//------------------------------------------------------------------------------

	void
dumpBytes(
    FILE* out,
    uint8_t*  buffer,
    uint32_t  size,
    char*  prefix
){
    char* p = (NULL != prefix) ? prefix : "";

    char lineBuffer[ 80 ];
    char asciiChars[ 24 ];

    register char*  ptr = (char*)buffer;
    register uint32_t  count = size;
    register uint32_t  ch;

    register char*  hex = lineBuffer;
    register char*  print = asciiChars;
    register int  inLine = 0;


    for( ch=0; ch < count; ++ch, ++ptr )
    {
        sprintf( hex, "%02X ", (*ptr & 0xff) );
		hex += 3;

		if( isprint(*ptr) )
		{
	    	*(print++) = *ptr;
		}
		else
		{
	    	*(print++) = '.';
		}

		if( 16 == ++inLine )
		{
	    	lineBuffer[47] = '\0';
	    	asciiChars[16] = '\0';
	    	fprintf( out, "%s%s   %s\n", p, lineBuffer, asciiChars );

	    	hex = lineBuffer;
	    	print = asciiChars;
	    	inLine = 0;
		}
    }


    // Flush remaining characters.
    if( 0 < inLine )
    {
	while( inLine++ < 16 )
	{
	    sprintf( hex, "   " );
	    hex += 3;

	    *(print++) = ' ';
	}
	lineBuffer[47] = '\0';
	asciiChars[16] = '\0';
	fprintf( out, "%s%s   %s\n", p, lineBuffer, asciiChars );
    }

    return;
}


//------------------------------------------------------------------------------

	static LOG*
createLog(
    HASH  parentId
){
    LOG* newLogEntry = NULL;

    if( NULL != (newLogEntry = (LOG*)malloc( sizeof(LOG) )) )
    {
        memset( newLogEntry, 0, sizeof(LOG) );
        if( NULL != parentId )
	{
	    memcpy( &(newLogEntry->parentId), parentId, sizeof(HASH) );
	}
    }

    return( newLogEntry );
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
