//~~ log.h ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
/*
 *  description
 *
 * @author:    Andy Pippin   <pippin@cs.ucsb.edu>
 * @version:   $Date: 2005/07/17 18:39:28 $   $Revision: 1.18 $
 * @requires:   
 */

//~~[ Identity ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#ifndef  __LOG_H__
#define  __LOG_H__

#ifdef __LOG_C__
static char* hdrIdentity =
    "$Id: log.h,v 1.18 2005/07/17 18:39:28 mattsun Exp $";
#endif

#ifdef __cplusplus
extern "C" {
#endif

//~~[ Includes ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#include  "libsha1.h"


//~~[ Typedefs/Structures/Enums ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// What type of log entries is this.
typedef enum _log_type_enum
{
    LOG_File ='f',
    LOG_Dir ='d',

    LOG_Unknown ='?'
}
  LOG_TYPE;

// What type of directory actions.
typedef enum _dir_action_enum
{
    DIR_Add ='a',
    DIR_Del ='d',
    DIR_Chg ='c',

    DIR_Unknown='?'
}
  DIR_ACTION;


// File entry.
struct _log_file
{
    uint32_t  begin;
    uint32_t  end;
    uint32_t  size;
    uint8_t*   data;
};

// Directory entry.
struct _log_dir
{
    DIR_ACTION  action;
    char*  name;
    HASH  hashId;
};


// *** LOG ENTRY ***
typedef struct _log
{
    LOG_TYPE  logType;
    HASH  parentId;
    char*  objName;
    union {
        struct _log_file  _file_entry;
        struct _log_dir  _dir_entry;
    } _u;
    struct _log*  nextInner;
}
  LOG;

// Helper values.
#define u_fileEntry  _u._file_entry
#define u_dirEntry  _u._dir_entry


//~~[ Constants ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//~~[ Variables ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//~~[ Prototypes ]~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//------------------------------------------------------------------------------
/*
 *  Display a log entry.
 *
 * @param   logEntry   The entry to display.
 * @param   outStream  Where to display the entry (NULL --> stdout).
 */
	extern void
log_print(
    LOG* logEntry,
    FILE* outStream
);


//------------------------------------------------------------------------------
/*
 *  Release the resources consumed for a log entry (a "deep free").
 *
 * @param   logEntry   The log entry to release.
 */
	extern void
log_delete(
    LOG* logEntry
);


//------------------------------------------------------------------------------
/*
 *  Create a log entry of a File with the specified values.
 *
 * @param   parentId    The parent of this entry.
 * @param   fileName    The name of the file to log.
 * @param   prevInner   Previous inner log entry.
 * @return  A new log entry.
 */
	extern LOG*
log_createFullFile(
    HASH  parentId,
    char*  fileName,
    LOG*  prevInner
);


//------------------------------------------------------------------------------
/*
 *  Create a log entry of a File with the specified values.
 *
 * @param   parentId    The parent of this entry.
 * @param   fileName    The name of the file to log.
 * @param   prevInner   Previous inner log entry.
 * @param   begin       Beginning position of change.
 * @param   end         Last position of change.
 * @param   size        Number of bytes in change.
 * @param   content     Change.
 * @return  A new log entry.
 */
	extern LOG*
log_createFileEntry(
    HASH  parentId,
    char*  fileName,
    LOG*  prevInner,
    uint32_t  begin,
    uint32_t  end,
    uint32_t  size,
    uint8_t*  content
);


//------------------------------------------------------------------------------
/*
 *  Create a log entry of file differences.
 *
 * @param   parentId    The parent of this entry.
 * @param   fileName    The name of the file to log.
 * @param   prevInner   Previous inner log entry.
 * @param   beforeFilename   Content buffer.  If NULL, use an internal buffer.
 * @param   afterFilename  Maximum size of the buffer.
 * @return  A series of log entries.
 */ 
	extern LOG*
log_createFileDiff(
    HASH  parentId,
    char*  fileName,
    LOG*  prevInner,
    char* beforeFilename,
    char* afterFilename
);


//------------------------------------------------------------------------------
/*
 *  Create a log entry of file differences.
 *
 * @param   parentId    The parent of this entry.
 * @param   fileName    The name of the file to log.
 * @param   prevInner   Previous inner log entry.
 * @param   action      What is happenning (add, delete, change, etc.)
 * @param   objName     The name of the object being changed.
 * @param   objId       The name of the new object.
 * @return  A series of log entries.
 */ 
	extern LOG*
log_createDirEntry(
    HASH  parentId,
    char*  dirName,
    LOG*  prevInner,
    DIR_ACTION  action,
    char*  objName,
    HASH  objId
);


//------------------------------------------------------------------------------
/*
 *  Marshal a log entry into a stream of bytes.
 *
 * @param   logEntry  The log entry to convert.
 * @param   size      How many bytes were produced.
 * @return  Encoded bytes stream, and size.
 */
	extern uint8_t*
log_encode(
    LOG*  logEntry,
    uint32_t*  size  // Output
);


//------------------------------------------------------------------------------
/*
 *  Unmarshal data into a sequence of log entries.
 *
 * @param   buffer      A stream of bytes.
 * @param   bufferSize  Number of bytes.
 * @return  returns
 */
	extern LOG*
log_decode(
    uint8_t*  buffer,
    uint32_t  bufferSize 
);


//------------------------------------------------------------------------------
/*
 *  Marshal a log entry into a stream of bytes, and then encrypt it.
 *
 * @param   logEntry  The log entry to convert.
 * @param   size      How many bytes were produced.
 * @return  Encoded & encrypted bytes stream, and size.
 */
	extern uint8_t*
log_encrypt(
    LOG*  logEntry,
    uint32_t*  size  // Output
);


//------------------------------------------------------------------------------
/*
 *  Decrypt a log entry, and unmarshal a log entry into a sequence of logs.
 *
 * @param   buffer      A stream of bytes.
 * @param   bufferSize  Number of bytes.
 * @return  returns
 */
	extern LOG*
log_decrypt(
    uint8_t*  buffer,
    uint32_t  bufferSize 
);



//------------------------------------------------------------------------------
/*
 *  Apply a log entry to a file.
 *
 * @param   fileName  The name of the file to apply the log to.
 * @param   logEntry  The log to apply.
 * @return  Success (0), or failure (!0).
 */
	extern int
log_applyFileEntry(
    char* fileName,
    LOG*  logEntry
);


//------------------------------------------------------------------------------
/*
 *  Display the bytes similar to "od".
 *
 * @param   out     Output location (NULL --> stdout).
 * @param   buffer  Buffer to display.
 * @param   size    Number of bytes to display.
 * @param   prefix  Prepend to each line of output.
 */
	extern void
dumpBytes(
    FILE* out,
    uint8_t*  buffer,
    uint32_t  size,
    char*  prefix
);


//------------------------------------------------------------------------------
/*
 *  Process a response from a storage request.
 *
 * @param   fileId  The hash id of the file.
 * @param   version The version of this file.
 * @param   size    Number of bytes in response.
 * @param   data    The response data.
 * @param   status  Success (!0), or failure (0).
 */
	extern void
log_store_response(
    HASH*     fileId,
    uint32_t  version,
    uint32_t  size,
    uint8_t*  data,
    uint8_t   status
);


//------------------------------------------------------------------------------
/*
 *  Process an acknowledgement from a storage request.
 *
 * @param   fileId  The hash id of the file.
 * @param   version The version of this file.
 * @param   status  Successfully stored (!0), or an error (0).
 */
	extern void
log_store_ack(
    HASH*     fileId,
    uint32_t  version,
    uint8_t   status
);

//------------------------------------------------------------------------------
/*
 *  Store received updated file back into file on disk
 *
 * @param   fileId  The hash id of the file.
 * @param   version The version of this file.
 * @param   status  Successfully stored (!0), or an error (0).
 */
	extern void
log_store_update(
    HASH*     fileId,
    uint32_t  version,
    uint8_t*  data,
    uint32_t  size
);



//------------------------------------------------------------------------------
/*
 *  Perform a test of the functions in this file.
 *
 * @return  Success (0), or the test it failed on (!0).
 */
	extern int
log_test(
    void
);


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#ifdef __cplusplus
}
#endif
#endif // __LOG_H__
