/*== quartz.c ================================================================*\
|*									      *|
|*  Entry point into quartz.					              *|
|*									      *|
|* @author   $Author: mattsun $				                      *|
|* @version  $Revision: 1.24 $  $Date: 2005/11/11 02:45:14 $		      *|
|*									      *|
\*============================================================================*/
#include "group.h"
#include "device/device.h"

//=[ Identity ]=================================================================

#define __QUARTZ_C__
static char* pszSrcIdent =

    "$Id: quartz.c,v 1.24 2005/11/11 02:45:14 mattsun Exp $";


//=[ Includes ]=================================================================

#include  <errno.h>
#include  <getopt.h>
#include  <libgen.h>
#include  <stdio.h>
#include  <stdlib.h>
#include  <string.h>
#include  <sys/types.h>
#include  <sys/stat.h>
#include  <unistd.h>

#include  "quartz.h"
#include  "error.h"
#include  "libsha1.h"
#include  "store.h"
#include  "chimera.h"
#include  "communication.h"
#include  "qfs/block.h"
#include  "device/device.h"
//#include  "qfs/user_data.h"

pthread_mutex_t mutex_cmdline;
int total_receiving;
int total_received;

//=[ Typedefs/Structures/Enums ]================================================


// global defines
#define MAX_USERNAME   64
#define MAX_BUFFER     64

#define TYPE_REGISTER   1
#define TYPE_LOGIN      2
#define TYPE_OTHER      3

#define RESULT_SUCCESS  1   // successful operation
#define RESULT_EXISTS   2   // user already exists
#define RESULT_INVALID  3   // invalid password
#define RESULT_UNKNOWN  4   // unknown user

// incoming and outgoing message structure
typedef struct message_z
{
  uint8_t type;                // 1:signup, 2:login, 3:other
  char username[MAX_USERNAME]; // username for the given user
  char password[MAX_BUFFER];   // password in SHA-1 crypted form
  char buffer[MAX_BUFFER];     // buffer for miscellaneous text
  uint8_t result;              // 1:success, 2:user exists, 3:invalid pass
} message;


//=[ Constants ]================================================================

#define  DEFAULT_Verbosity	1


//=[ Variables ]================================================================

// Command line options.
//
struct option  optionList[] = 
{
    { "help",		optional_argument,	NULL,	'h' },
    { "quiet",		no_argument,		NULL,	'q' },
    { "usage",		no_argument,		NULL,	 1  },
    { "version",	no_argument,		NULL,	'V' },
    { "verbose",	optional_argument,	NULL,	'v' },

    { "id",		required_argument,	NULL,	'i' },
    { "join",		required_argument,	NULL,	'j' },
    { "server",		optional_argument,	NULL,	's' },

    { "testlog",	no_argument,		NULL,	 2  },

    { NULL,		no_argument,		NULL,	 0 }
};
#define  SHORT_OPTIONS  "hi:j:qsvV"


// Global verbose level.
//
int verboseLevel = DEFAULT_Verbosity;

static char*  g_cmdName = "quartz";


//=[ Prototypes ]===============================================================

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
/**
 * Invokes user-defined code (duh!).
 *
 * @param numArgs  Number of arguments on the command line.
 * @param  args  Arguments from the command line.
 * @param  envs  Environmental variables.
 * @return  Exit code.
 */
	int 
main(
    int		numArgs,
    char*	args[],
    char*	envs[]
);


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
/**
 * Display correct usage (synopsis) of the application.
 *
 * @param  programName  Application name.
 * @param  prefix  What to put before each line.
 * @param  outStream  Stream to display output.
 */
	static void
printUsage(
    char*	programName,
    char*	prefix,
    FILE*	outStream
);


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
/**
 *  Display complete description of the options.
 *
 * @param  helpTopic  Topic of interest.
 * @param  programName  Application name.
 * @param  prefix  What to put before each line.
 * @param  outStream  Stream to display output.
 */
	static void
printHelp(
    char*	programName,
    char*	helpTopic,
    char*	prefix,
    FILE*	outStream
);


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
/**
 * Writes the application name and version to stdout.
 *
 * @param  programName - Application name.
 * @param  prefix -      What to put before each line.
 * @param  outStream -   Stream to display output.
 */
	static void
printVersion(
    char*	programName,
    char*	prefix,
    FILE*	outStream
);


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
/**
 * Initialize Quartz.
 *
 * @param  join  Join the network (1), or be the first (0).
 * @param  host  The name and port of the host to join.
 * @return  Success (0), or failure (<0).
 */
	static int
quartz_init(
    int   join,
    char* joinHost
);


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
/**
 * Create the default node key.
 *
 * @return  Default node key.
 */
	static unsigned int
getDefaultUserKey(
    void
);


//=[ Implementation ]===========================================================

	static char
IsEqual(
    char*	str1,
    char*	str2
){
    char  result = 0;
    int   len1, len2;

    if( (NULL != str1) && (NULL != str2) )
    {
        len1 = strlen( str1 );
        len2 = strlen( str2 );

        result = !strncasecmp( str1, str2, ((len1 < len2) ? len1 : len2) );
    }

    return( result );
}

char *strlowr(char *s)
{
   if (s != NULL)
   {
      char *p;

      for (p = s; *p; ++p)
         *p = tolower(*p);
   }
   return s;
}



//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	static void
printVersion(
    char*	programName,
    char*	prefix,
    FILE*	outStream
){
    // Local variables.
    //
    static char*  rcsVersion = "$Revision: 1.24 $";
    static char*  rcsDate    = "%rcs_date%";
    static char*  rcsAuthor  = "$Author: mattsun $";

    char    versBuffer[ 80 ];			// Result.
    char    buffer[ 256 ]; 			// Working tool.
    char*   pBegin;				// Working tool.
    char*   pEnd;				// Working tool.
    char*   pSrc;				// Working tool.
    char*   pDst;				// Working tool.


    // Fix parameters.
    //
    programName = (NULL != programName) ? programName : g_cmdName;
    prefix = (NULL != prefix) ? prefix : (char*)"";
    outStream = (NULL != outStream) ? outStream : stdout;


    // Get version.
    //
    strncpy( buffer, rcsVersion, sizeof(buffer) );
    if( NULL != (pBegin = strchr( buffer, ':' )) )
    {
        pBegin += 2;
	if( NULL != (pEnd = strchr( pBegin, '$' )) )
	{
	    *(--pEnd) = '\0';
	    strncat( versBuffer, " v", sizeof(versBuffer) );
	    strncat( versBuffer, pBegin, sizeof(versBuffer) );
	}
    }

    // Get date.
    //
    strncpy( buffer, rcsDate, sizeof(buffer) );
    if( NULL != (pSrc = strchr( buffer, ':' )) )
    {
        pSrc += 2;
	pDst = versBuffer + strlen( versBuffer );

	*(pDst++) = ' ';
	*(pDst++) = *(pSrc++);		// C.../../..
	*(pDst++) = *(pSrc++);		// .C../../..
	*(pDst++) = *(pSrc++);		// ..Y./../..
	*(pDst++) = *(pSrc++);		// ...Y/../..
	pSrc++;
	*(pDst++) = *(pSrc++);		// ..../M./..
	*(pDst++) = *(pSrc++);		// ..../.M/..
	pSrc++;
	*(pDst++) = *(pSrc++);		// ..../../D.
	*(pDst++) = *(pSrc++);		// ..../../.D
	*pDst = '\0';
    }

    // Get author.
    //
    strncpy( buffer, rcsAuthor, sizeof(buffer) );
    if( NULL != (pBegin = strchr( buffer, ':' )) )
    {
        pBegin += 2;
	if( NULL != (pEnd = strchr( pBegin, '$' )) )
	{
	    *(--pEnd) = '\0';
	    strncat( versBuffer, "/", sizeof(versBuffer) );
	    strncat( versBuffer, pBegin, sizeof(versBuffer) );
	}
    }

    fprintf( outStream, "%s%s %s\n", prefix, programName, versBuffer );
    fflush( outStream );

    return;
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	static void 
printUsage(
    char*	programName,
    char*	prefix,
    FILE*	outStream
){
    // Fix parameters.
    //
    programName = (NULL != programName) ? programName : g_cmdName;
    prefix = (NULL != prefix) ? prefix : (char*)"";
    outStream = (NULL != outStream) ? outStream : stdout;

    fprintf( outStream, "%sUsage: %s [-hqsVv] [-j node] [-i id]\n", prefix, programName );
    fflush( outStream );

    return;
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	static void
printHelp(
    char*	programName,	
    char*	helpTopic,
    char*	prefix,
    FILE*	outStream
){
    // Local variables.
    //
    char  isNull;
    char  isAll;
    char  pre2[ 80 ];
    char  pre4[ 80 ];

    // Set flags.
    //
    isNull = (NULL == helpTopic) ? 1 : 0;
    if( ! isNull )
	isAll = (0 == strcasecmp( "all", helpTopic )) ? 1 : 0;


    // Fix parameters.
    //
    programName = (NULL != programName) ? programName : g_cmdName;
    helpTopic = (NULL != helpTopic) ? helpTopic : (char*)"";
    prefix = (NULL != prefix) ? prefix : (char*)"";
    outStream = (NULL != outStream) ? outStream : stdout;

    strncpy( pre2, prefix, sizeof(pre2) );
    strncpy( pre2, "  ", sizeof(pre2) );
    strncpy( pre4, prefix, sizeof(pre4) );
    strncpy( pre4, "    ", sizeof(pre4) );


    // *** Now start displaying help. ***
    fprintf(outStream, "%s\n", prefix );

    // Description
    //
    if( isNull || isAll || IsEqual( helpTopic, "desc" ) )
    {
        fprintf( outStream, 
"%s    \"quartz\" is an implementation of a peer-to-peer data storage\n\
%s  application.  It allows the user to keep a small set of files \n\
%s  consistant across several machines.\n\
%s\n", prefix, prefix, prefix, prefix, prefix );
    }

    // Usage
    //
    if( isNull || isAll || IsEqual( helpTopic, "usage" ) )
    {
	printUsage( programName, pre4, outStream );
	fprintf(outStream, "%s\n", prefix );
    }

    // Options
    //
    if( isNull || isAll || IsEqual( helpTopic, "option" ) )
    {
        fprintf( outStream,
"%s  If a long option shows an argument as mandatory, then it is mandatory\n\
%s  for the equivalent short option also.  Similarly for optional arguments.\n\
%s\n", prefix, prefix, prefix );

	// Main operation.
	//
	char addr[ 32 ];
	gethostname( addr, sizeof(addr) );
	Key defId = (Key)key_makehash( addr );

        fprintf( outStream, "%s\n%s** Main Operation Modes **\n\
%s  -j key, --join=key     Join a quartz group including this node.\n\
%s  -i id,  --id=id        Use the specified node id (def = \"%02X\").\n\
%s  -s,     --server       Server mode, don't start the command line.\n\
", prefix, prefix, prefix, prefix, defId, prefix );


	// Informative/Debug options.
	//
        fprintf( outStream, "%s\n%s** Informative Output **\n\
%s  -h, --help[=topic]     Display this help message or details on a topic.\n\
%s      --usage            Display synopsis of command line arguments.\n\
%s  -v, --verbose[=level]  Set verbosity (def = %d)\n\
%s  -q, --quiet            Set verbosity to 0.\n\
%s  -V, --version          Print version to stdout and exit.\n",
		prefix, prefix,
		prefix,
		prefix,
		prefix, DEFAULT_Verbosity,
		prefix,
		prefix );
    }

    // Topics
    //
    if( isAll || IsEqual( helpTopic, "topic" ) )
    {
        fprintf( outStream, "%s\n%s** Help Topics **\n\
%s  all     Display this help message or details on a topic.\n\
%s  desc    Description of this command.\n\
%s  topic   Help topics.\n\
%s  usage   Synopsis of command line arguments.\n\
%s  option  Description of command line arguments.\n\
%s\n",
		prefix, prefix,
		prefix,
		prefix,
		prefix,
		prefix,
		prefix,
		prefix );
    }

    fflush( outStream );
    return;
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	int
main(
    int		numArgs,
    char*	args[],
    char*	envs[]
){
    // Local variables.
    //
    int   exitCode = 0;				// Exit code.
    int   shortOption = 0;			// Short option value.
    int   longOption = 0;			// Long option value.
    char  cmdError = 0;				// Command line parsing error.

    char* joinHost = NULL;			// Host to join
    int   doJoin = 0;				// Do join.
    int   doCmdline = 1;			// Start command line?

    int   testLog = 0;

    // Initialize.
    //
    g_cmdName = (char*)basename( args[ 0 ] );


    // Parse the command line arguments.
    //
    while( EOF != (shortOption = getopt_long( numArgs, args, SHORT_OPTIONS,
			optionList, &longOption )) )
    {
	switch( shortOption )
	{
	  // *** HELP ***
	  //
	  case  'h':
	    printHelp( g_cmdName, optarg, "", stdout );
            return( 0 );
	    // ***                                    ***
	    // *** DOES NOT CONTINUE -- PROGRAM EXITS ***
	    // ***                                    ***


          // *** USAGE ***
          //
          case  1:
	    printUsage( g_cmdName, "", stdout );
            return( 0 );
	    // ***                                    ***
	    // *** DOES NOT CONTINUE -- PROGRAM EXITS ***
	    // ***                                    ***


          // *** MY ID ***
          //
          case  'i':
debug( __FILE__, __LINE__, "Implement set id\n" );
	    break;


          // *** JOIN ***
          //
          case  'j':
	    doJoin = 1;
            if( NULL == optarg )
debug( __FILE__, __LINE__, "Implement joining default\n", optarg );
            else
	        joinHost = optarg;
            break;


          // *** SERVER ***
          //
          case  's':
	    doCmdline = 0;
	    break;


          // *** VERBOSE ***
          //
          case  'v':
            if( NULL == optarg )
                verboseLevel++;
            else
                verboseLevel = atoi( optarg );
            break;


	  // *** VERSION ***
	  //
	  case  'V':
	    printVersion( g_cmdName, "", stdout );
            return( 0 );
	    // ***                                    ***
	    // *** DOES NOT CONTINUE -- PROGRAM EXITS ***
	    // ***                                    ***
	    
	case 'a': // administrator
	  doJoin = 51;


          // *** QUIET ***
          //
          case  'q':
	    verboseLevel = 0;
            break;


	  // *** TESTLOG ***
	  //
	  case  2:
	    testLog = 1;
	    break;


	  // *** ERROR ***
	  //
	  case  '?':
	  default:
	    cmdError = 1;
	    break;
	}
    }


    // Make sure we have the right number of items to work with.
    //
    if( 0 == (numArgs - optind) )
    {
        // cmdError = 1;
    }

    // Print error message.
    //
    if( cmdError )
    {
	printUsage( g_cmdName, "", stderr );
	exitCode = 1;
    }

    // If the initialization happened, go into the command line loop.
    //
    else if( 0 <= quartz_init( doJoin, joinHost ) )
    {
	if( testLog )
	{
	    log_test();
	}
        else if( doCmdline )				// Start command line?
	{
	    stable = 1; // now that node has had enough time to receive necessary
	                // updates, the node is stable and can proceed with its own
	    cmdline( "\nQuartz> " );
	}
	else
	{
	    pause();
	}
    }

    // Exit.
    //
    return( exitCode );
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//~~ Local methods ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

static int
quartz_init(
    int   join,
    char* joinHost
){
  // Get the host id
  char addr[ 32 ];
  char *txt_free[BLOCKSIZE];
  int i;
  gethostname( addr, sizeof(addr) );
  myKey = (Key)key_makehash( addr );
  
  // server
  int port, sock_tcp, n;
  struct sockaddr_in serv_addr;
  struct hostent *server;
  message msg_recv, msg_send;
  int login_succeeded;

  // password hashing
  char pw_hash[ 1024 ]   = { '\0' };

  // Create file.sto and meta.sto files in case they DNE
  FILE *ffile, *fmeta, *fdata, *fstat, *flogs, *fconf;
  //super_block *sblock, *temp_sblock;
  /*
  inode *temp_inode;
  block tempblock;
  int inode_num;
  */

  ffile = fopen("file.sto", "r");
  fmeta = fopen("meta.sto", "r");
  fdata = fopen("system.qtz", "r");
  //fstat = fopen("stats.log", "r");
  flogs = fopen("my.log", "r");
  fconf = fopen("config.qtz", "r");

  if (fconf == NULL) // config.qtz does not exist
  {
    // insert a default NULL cfg_user struct
    _cfg_user_z cfg_user;
    sprintf(cfg_user.userName,  "00NULL00");
    sprintf(cfg_user.userEmail, "00NULL00@00NULL00.COM");
    sprintf(cfg_user.userHash,  "00NULL00");
    sprintf(cfg_user.nodeName,  "00NULL00");
    cfg_user.groups = 0;

    fconf = fopen("config.qtz", "w");
    fwrite(&cfg_user, sizeof(_cfg_user_z), 1, fconf);
    fclose(fconf);
  }

  if (flogs == NULL) // my.log does not exist
  {
    flogs = fopen("my.log", "w");
    fprintf(flogs, "day month date hh:mm:ss year type key packet other\n");
    fclose(flogs);
  }
  else
    fclose(flogs);

  /*
  if (fstat == NULL) // stats.log does not exist
  {
    fstat = fopen("stats.log", "w");
    fprintf(fstat, "signons signoffs groups files size received sent commits updates\n");
    fclose(fstat);
  }
  else
    fclose(fstat);
  */

  if (fdata == NULL) // system.qtz does not exist
  {
    fdata = fopen("system.qtz", "w");
    fclose(fdata);

    if (!user_data_init())
      log_exec(LOG_ERROR, myKey, LOG_INIT, "system.qtz");
  }
  else
    fclose(fdata);

  if (ffile == NULL) // file.sto does not exist
  {
    ffile = fopen("file.sto", "w");
    fclose(ffile);

    FILE_STO_NEW = 1;
  }
  else
    fclose(ffile);

  if (fmeta == NULL) // meta.sto does not exist
  {
    fmeta = fopen("meta.sto", "w");
    fclose(fmeta);
  }
  else
    fclose(fmeta);

  // Load the device & user info.
  //device_init(); // device.c

  update_last_key = 0;
  update_left_key = 0;

  stable = 0; // upon startup, node is not stable just yet

  // Get all the groups
  if (group_load_all_groups() == -2)
  {
    printf( "\n*******************************\n");
    printf( " SETTING UP QUARTZ! v1.0");
    printf( "\n*******************************\n\n");

    // invalid user data -- must enter information again

    printf( " You have not setup user information for this device.\n\n" );
    
    init_user_data();

    // create symmetric key based on the password
    char keyFilename[ FILENAME_MAX ];
    get_config_file( "key", keyFilename, FILENAME_MAX );

    if (deviceInfo != NULL)
    {
      char* password = NULL;
      password = misc_getKeyPassword();
    
      struct stat statBuffer;
      if( -1 == stat( keyFilename, &statBuffer ) )
      {
	printf( " Creating key in \"%s\"\n", keyFilename );
	quartz_generate_key( keyFilename, password );
	
	printf( " Either copy this key over to all your other\n");
	printf( " devices or replace this one with your used key.\n\n");
      }
      else
      {
	printf( " Recreating key in \"%s\"\n", keyFilename );
	quartz_generate_key( keyFilename, password );
	
	printf( " Either copy this key over to all your other\n");
	printf( " devices or replace this one with your used key.\n\n");
      }

      printf( " You are now setup to run Quartz!\n");
      printf( " Please run \"quartz\" again. Thank you.\n\n"); 
    }
    else
      printf( " ** Error setting up Quartz\n" );

    exit(0);
  }

  log_exec(LOG_SIGNON, myKey, LOG_NA, LOG_NA);

  printf("\n***********************************\n");
  printf("     WELCOME TO QUARTZ! v 1.0");
  printf("\n***********************************\n\n");

  // login to Quartz
  port = 5151;

  // cyclops is where the server will be held since...
  // Cyclops is the LEADER of the X-Men!
  server = gethostbyname("cyclops.cs.ucsb.edu");
  if (server == NULL)
    printf("Error: no such host\n");
  bzero((char *) &serv_addr, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_port   = htons(port);
  bcopy((char *) server->h_addr, (char *) &serv_addr.sin_addr.s_addr,
	server->h_length);
  
  memset(&msg_recv, '\0', sizeof(message));
  memset(&msg_send, '\0', sizeof(message));

  char username[MAX_USERNAME] = { '\0' };
  char *password;
  int len, login_tries;

  login_succeeded = 0;
  login_tries     = 0;

  while (!login_succeeded)
  {
    if (login_tries >= 3)
    {
      printf("***********************************\n");
      printf(" Too many authentication failures.\n");
      printf(" Good bye!\n");
      printf("***********************************\n\n");
      exit(0); // quit if failure to login within 3 tries
    }

    sock_tcp = socket(AF_INET, SOCK_STREAM, 0);
    if (sock_tcp < 0)
      printf("Error: opening socket\n");

    printf( " Please login...\n\n");
    
    printf( "   Username: " ); fflush( stdout );
    fgets( username, sizeof(username), stdin );
    
    len = strlen( username );
    if( isspace( username[len-1] ) )
      username[len-1] = '\0';
    
    password = getpass("   Password: ");
    
    len = strlen( password );
    if ( isspace( password[len-1] ) )
      password[len-1] = '\0';
    
    if (strcmp(username, "") == 0 || strcmp(password, "") == 0)
    {
      printf( "\n Do not leave any fields blank.\n\n");
    }
    else if (strcmp(strlowr(username), deviceInfo->userName) != 0)
    {
      printf("\n Username does not match with this configuration\n\n");
    }
    else
    {
      HASH pwHash = { '\0' };
      if (NULL != sha1_string(password, &pwHash))
	sprintf(pw_hash, "%s", sha1_hashToString(&pwHash));
      
      memset(&msg_send, '\0', sizeof(message));
      msg_send.type = TYPE_LOGIN;
      sprintf(msg_send.username, "%s", username);
      sprintf(msg_send.password, "%s", pw_hash); // SHA-1 crypted form
      sprintf(msg_send.buffer,   "%s", deviceInfo->userEmail);
      
      sock_tcp = socket(AF_INET, SOCK_STREAM, 0);
      if (sock_tcp < 0)
	printf("Error: opening socket\n");

      n = connect(sock_tcp, &serv_addr, sizeof(serv_addr));
      if (n < 0)
	printf("Error: connecting with the main server\n");
      else
      {
	n = send(sock_tcp, &msg_send, sizeof(message), 0);
	if (n < 0)
	  printf("Error: sending registration information\n");
	else
	{
	  n = recv(sock_tcp, &msg_recv, sizeof(message), 0);
	  if (n < 0)
	    printf("Error: receiving reply from server\n");
	  else
	  {
	    if (msg_recv.result == RESULT_SUCCESS)
	    {
	      login_succeeded = 1;
	      memset(deviceInfo->password, '\0', 15);
	      strncpy(deviceInfo->password, password, 15);
	    }
	    else
	      printf("\n   Sorry.  The username/password is incorrect.\n\n");
	  }
	}
      }

      close(sock_tcp);

      login_tries++;
    }
  }
    
  // Connect to the network.
  //quartz_signon();

  printf("\n Joining the Quartz network... ");
  fflush(stdout);

  chimera_setup( join, msg_recv.buffer );
  total_receiving = 0;
  total_received  = 0;

  // increase signons
  user_incr_signons();

  return( 0 );
}


//= End of file ================================================================
