/*
 *  sha1file.c
 *
 *  Description:
 *      Implementation the functions that create a SHA-1 hash from the
 *	contents of a file.
 */

#include  <stdio.h>
#include  <errno.h>

#include  "sha1.h"
#include  "sha1file.h"

#define  BUFFER_SIZE  64

/*
 *  SHA1FilePointer
 *
 *  Description:
 *      This function will create a SHA-1 hash from an open file pointer.
 *
 *  Parameters:
 *      context: [in/out]
 *          The context to reset.
 *	inputFile:
 *	    The file content to hash.
 *
 
 *  Returns:
 *      sha Error Code.
 *
 */
int SHA1FilePointer(SHA1Context* context, FILE* inputFile )
{
    int shaSuccess = 0;
    char readBuffer[ BUFFER_SIZE ];
    int bytesRead;
    uint8_t messageDigest[20];


    // Safety dance.
    if( (NULL == context) || (NULL == inputFile) )
    {
        shaSuccess = shaNull;
    }

    // Reset the context, or bail.
    else if( 0 != (shaSuccess = SHA1Reset(context)) )
    {
    }

    // Process the file a chunk at a time.
    else
    {
	while( 0 != ( bytesRead = fread( readBuffer,
				    sizeof(char), BUFFER_SIZE, inputFile )) )
	{
	    shaSuccess = SHA1Input( context, readBuffer, bytesRead );
	    if( 0 != shaSuccess )
	    {
		//fprintf(stderr, "SHA1Input Error %d.\n", shaSuccess );
		break;
	    }
	}

	// Make sure that the loop isn't exiting on an error.
	if( !feof( inputFile ) )
	{
	    fprintf( stderr, "Error reading file: %s\n", strerror(errno) );
	    shaSuccess = shaFileReadError;
	}

	// See if there is a result.
	else
	{
	    shaSuccess = SHA1Result(context, messageDigest);
	}
    }

    return( shaSuccess );
}


/*
 *  SHA1FileName
 *
 *  Description:
 *      This function will create a SHA-1 hash from the contents of a file.
 *
 *  Parameters:
 *      context: [in/out]
 *          The context to reset.
 *	filename:
 *	    The file content to hash.
 *
 *  Returns:
 *      sha Error Code.
 *
 */
int SHA1FileName( SHA1Context* context, const char* filename )
{
    int shaSuccess = 0;
    FILE* inputFile = NULL;

    if( (NULL == context) || (NULL == filename) )
    {
        shaSuccess = shaNull;
    }
    else if( NULL == (inputFile = fopen( filename, "r" )) )
    {
        fprintf( stderr, "Couldn't open file \"%s\": %s\n",
		filename, strerror( errno ) );
	shaSuccess = shaFileOpenError;
    }
    else
    {
    	shaSuccess = SHA1FilePointer( context, inputFile );
	fclose( inputFile );
    }

    return( shaSuccess );
}
