/********************************
 * Author: Matt Sun
 * Date:   Mar 1, 2005
 *
 * Updated: April 23, 2005
 * 
 * Filestore functions and test
 ********************************/

#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <unistd.h>
#include <stdio.h>
#include <malloc.h>
#include <gdbm.h>

#include "device/device.h"
#include "chimera.h"
#include "log.h"
#include "quartz_rsa.h"
#include "device/communication.h"
#include "qfs/block.h"

#define USAGE "test [ -j bootstrap:port ] port key"
#define OPTSTR "j:"
extern char *optarg;
extern int optind;

extern pthread_mutex_t mutex_oper;
int total_receiving;
int total_received;

// global variable for now, not sure how you would integrate this
#ifndef debug
#define debug 1
#endif

/* --------------------------------------------------------------------------- */

static void quartz_store_upcall_fwd_(Key **kp, Message **mp, ChimeraHost **hp);
static void quartz_store_upcall_update(Key *k, ChimeraHost *h, int joined);

/* encode/decode to bytes before transmission */
static uint8_t* msg_data_encode( _msg_data_z* msgData, uint32_t* size);
static _msg_data_z* msg_data_decode( char*  buffer, uint32_t  bufferSize );

static uint8_t* msg_meta_encode( _msg_meta_z* msgMeta, uint32_t* size);
static _msg_meta_z* msg_meta_decode( uint8_t*  buffer, uint32_t  bufferSize );


/* other store functions */
static int create_and_store_file(char *user_id, char *file_name, uint32_t version,
						  uint8_t *data, uint32_t size, datum key, 
						  GDBM_FILE *dbf, int gdbm_flag);
static int send_stored_file(char *user_id, char *file_name, uint32_t version,
					 Key key, int type);
static void send_ack( Key key, HASH *file_id, int type, uint32_t ver );
static _msg_data_z * create_msg_data();
static _msg_meta_z * create_msg_meta();

static LOG* create_test_log();

/* --------------------------------------------------------------------------- */

static LOG* create_test_log()
{
  LOG* fileEntry;
  LOG* dirEntry;

  HASH  parent;
  sha1_stringToHash( "FA26BE19DE6BFF93F70BC2308434E4A440BBAD02", &parent );
  char*  fileName = "filename";
  LOG*  nextInner = NULL;
  uint32_t begin = (uint32_t)10;
  uint32_t end = (uint32_t)13;
  uint32_t size = (uint32_t)21;
  uint8_t*  content = (uint8_t*)"this is 21 characters";
  if( NULL == (fileEntry = log_createFileEntry(parent, fileName, nextInner, 
					       begin, end, size, content )) )
  {
    printf( "Error creating file entry:  fileEntry = NULL\n" );
  }

  char* dirName = "dirname";
  nextInner = fileEntry;
  DIR_ACTION dirAction = DIR_Add;
  char*  objName = "newfile";
  HASH object;
  sha1_stringToHash( "DEDA74F9F76174D7F266343054BD48AB3F8213D5", &object );
  if( NULL == (dirEntry = log_createDirEntry(parent, dirName, nextInner, 
					     dirAction, objName, object )) )
  {
    printf( "Error creating dir entry:  dirEntry = NULL\n" );
  }
  return dirEntry;
}

/* --------------------------------------------------------------------------- */

static int create_and_store_file(char *user_id, char *file_name, uint32_t version,
						  uint8_t *data, uint32_t size, datum key, 
						  GDBM_FILE *dbf, int gdbm_flag)
{
  // create file w/ version number appended to file_name
  char newFilename[ FILENAME_MAX ];
  snprintf( newFilename, FILENAME_MAX, "%s/.quartz/%s/%s%d", 
			getenv( "HOME" ), user_id, file_name, version );
  FILE *fp = NULL;
  if ((fp = fopen( newFilename, "wb" )) == NULL){
	printf("ERROR (create_and_store_file): Could not create file\n");
	return -1;
  }else{
	fwrite(data, size, 1, fp);
	fclose(fp);
  }
  
  // insert file info (currently the version) into db
  datum value;
  value.dptr = NULL;
  char version_buf[20];
  sprintf(version_buf, "%d", version);
  value.dptr = version_buf;
  value.dsize = strlen(value.dptr);
  gdbm_store (*dbf, key, value, gdbm_flag);

  return 0;
}

/* --------------------------------------------------------------------------- */

static int send_stored_file(char *user_id, char *file_name, uint32_t version,
					 Key key, int type)
{
  // determine file w/ appended version number
  char storeFilename[ FILENAME_MAX ];
  snprintf( storeFilename, FILENAME_MAX, "%s/.quartz/%s/%s%d", 
			getenv( "HOME" ), user_id, file_name, version );
  FILE *fp = NULL;

  // get stat of file for size
  struct stat filestat;
  if (stat(storeFilename, &filestat) == -1)
  {
	printf("ERROR (send_stored_file): attempting to stat %s\n", storeFilename);
	return -1;
  }
		
  // open file for reading
  if ((fp = fopen( storeFilename, "rb" )) == NULL){
		  
	printf("ERROR (send_stored_file): Could not open file for reading\n");

	////
	//// maybe a malicious user manually deleted the file off his backing store ???
	////

	return -1;
		
  }else{
		  
	// read in file contents
	int filesize = filestat.st_size;
	char buffer[filesize];
	fread(buffer, filesize, 1, fp);
	fclose(fp);
		  
	// create message with requested version data
	_msg_data_z *msgData = create_msg_data();
	msgData->version = version;
	msgData->node_id = (Key)key_makehash(user_id); 
	HASH user;
	HASH fn;
	sha1_stringToHash( user_id, &user );
	sha1_stringToHash( file_name , &fn );
	memcpy( msgData->user_id, user, sizeof(HASH) );
	memcpy( msgData->file_name, fn, sizeof(HASH) );
	msgData->size = filesize;
	// msgData->data = buffer;

	// encode message
	uint32_t encodedLen;
	uint8_t* encodedData = msg_data_encode(msgData, &encodedLen);
		  
	// send encoded msg
	chimera_send(key, type, encodedLen, encodedData);
		  
	// clean up
	free( msgData );
	free( encodedData );

  }

  return 0;
}

/* --------------------------------------------------------------------------- */

static void send_ack( Key key, HASH *file_id, int type, uint32_t ver )
{

  // create message with requested version data
  _msg_data_z *msgData = create_msg_data();
  msgData->version = ver;
  msgData->node_id = key; //ignored anyways
  HASH user;
  // although user and filename migh be ignorable in this case
  sha1_stringToHash( "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx", &user );
  memcpy( msgData->user_id, user, sizeof(HASH) );
  memcpy( msgData->file_name, file_id, sizeof(HASH) );
  msgData->size = 0;
  //msgData->data = 0;
  
  // encode message
  uint32_t encodedLen;
  uint8_t* encodedData = msg_data_encode(msgData, &encodedLen);
  
  // send encoded msg
  chimera_send(key, type, encodedLen, encodedData);
  
  // clean up
  free( msgData );
  free( encodedData );
  
}

/* --------------------------------------------------------------------------- */

static uint8_t* msg_data_encode( _msg_data_z* msgData, uint32_t* size)
{
  _msg_data_z* encodeData = msgData;
  uint8_t*  encoded = NULL;
  uint8_t*  ptr;
  uint32_t  encodedSize = 0;

  if( (NULL != encodeData) && (NULL != size) )
    {
    uint32_t  numBytes = 0;
    
    // Calculate the byte count.

    // version
    numBytes += sizeof(uint32_t);

    // node_id + send_key
    numBytes += 2*sizeof(Key);

    // user_id group_id and file_name
    numBytes += 3*sizeof(HASH);

    // size
    numBytes += sizeof(uint32_t);

    // data
    numBytes += encodeData->size;

    // AHNold is the terminator
    numBytes++;

    // fprintf(stderr, "numBytes = %d\n", numBytes);

    // Try and get the buffer.
    if( NULL != (encoded = (uint8_t*) malloc(numBytes))) //(uint8_t*)malloc( numBytes )) )
    {
      ptr = encoded;

      // version
      uint32_t netSafe;
      netSafe = htonl(encodeData->version);
      memcpy( ptr, &netSafe, sizeof(uint32_t) );
      ptr += sizeof(uint32_t);

      // node_id
      memcpy( ptr, &(encodeData->node_id), sizeof(Key) );
      ptr += sizeof(Key);

      // send_key
      memcpy( ptr, &(encodeData->send_key), sizeof(Key) );
      ptr += sizeof(Key);
      
      // user_id and file_name
      memcpy( ptr, encodeData->user_id, sizeof(HASH) );
      ptr += sizeof(HASH);
      memcpy( ptr, encodeData->group_id, sizeof(HASH) );
      ptr += sizeof(HASH);
      memcpy( ptr, encodeData->file_name, sizeof(HASH) );
      ptr += sizeof(HASH);
      
      // size
      netSafe = htonl(encodeData->size);
      memcpy( ptr, &netSafe, sizeof(uint32_t) );
      ptr += sizeof(uint32_t);

     
      // data
      if (encodeData->size > 0){
		memcpy( ptr, encodeData->data, encodeData->size );
		ptr += encodeData->size;
      }
      
      *(ptr++) = '\0';
      
      encodedSize = numBytes;      
    }
    *size = encodedSize;
  }

  return( encoded );
}


/* --------------------------------------------------------------------------- */
	
static _msg_data_z* msg_data_decode( char*  buffer, uint32_t  bufferSize )
{
  _msg_data_z*  newData = NULL;
  uint8_t*  ptr = buffer;
  uint32_t  dataLen;

  if( (NULL != buffer) && (0 < bufferSize) )
  {
    if( (NULL != ptr) && ( 0 < bufferSize) && (NULL != (newData = create_msg_data( ))) )
    {
      // version
      uint32_t netSafe;
      memcpy( &netSafe, ptr, sizeof(uint32_t) );
      newData->version = ntohl( netSafe );
      ptr += sizeof(uint32_t);

      // node id
      memcpy( &(newData->node_id), ptr, sizeof(Key) );
      ptr += sizeof(Key);

      // send_key
      memcpy( &(newData->send_key), ptr, sizeof(Key) );
      ptr += sizeof(Key);

      // user_id group_id and file_name
      memcpy( newData->user_id, ptr, sizeof(HASH) );
      ptr += sizeof(HASH);
      memcpy( newData->group_id, ptr, sizeof(HASH) );
      ptr += sizeof(HASH);
      memcpy( newData->file_name, ptr, sizeof(HASH) );
      ptr += sizeof(HASH);

      // size
      memcpy( &netSafe, ptr, sizeof(uint32_t) );
      dataLen = newData->size = ntohl( netSafe );
      ptr += sizeof(uint32_t);

      // data
      if (dataLen > 0){
	memset(newData->data, '\0', MAX_FILE_SIZE); // if (newData->data = (uint8_t*)malloc( dataLen )) != NULL)
	{
	  memcpy( newData->data, ptr, dataLen );
	}
	ptr += dataLen;
      }
    }
  }


  return( newData );
}

/* --------------------------------------------------------------------------- */

static uint8_t* msg_meta_encode( _msg_meta_z* msgMeta, uint32_t* size)
{
  _msg_meta_z* encodeMeta = msgMeta;
  uint8_t*  encoded = NULL;
  uint8_t*  ptr;
  uint32_t  encodedSize = 0;

  if( (NULL != encodeMeta) && (NULL != size) )
    {
    uint32_t  numBytes = 0;
    
    // Calculate the byte count.

    // version
    numBytes += sizeof(uint32_t);

    // node_id
    numBytes += sizeof(Key);

    // send_key
    numBytes += sizeof(Key);

    // user_id and group_id
    numBytes += sizeof(HASH);

    // file_name
    numBytes += sizeof(char) * 64;

    numBytes += sizeof(HASH);

    // size
    numBytes += sizeof(uint32_t);

    // AHNold is the terminator
    numBytes++;

    // Try and get the buffer.
    if( NULL != (encoded = (uint8_t*)malloc( numBytes )) )
    {
      ptr = encoded;

      // version
      uint32_t netSafe;
      netSafe = htonl(encodeMeta->version);
      memcpy( ptr, &netSafe, sizeof(uint32_t) );
      ptr += sizeof(uint32_t);

      // node_id
      memcpy( ptr, &(encodeMeta->node_id), sizeof(Key) );
      ptr += sizeof(Key);

      // send_key
      memcpy( ptr, &(encodeMeta->send_key), sizeof(Key) );
      ptr += sizeof(Key);
      
      // user_id and file_name
      memcpy( ptr, encodeMeta->user_id, sizeof(HASH) );
      ptr += sizeof(HASH);

      memcpy( ptr, encodeMeta->file_name, sizeof(char) * 64);
      ptr += sizeof(char) * 64;

      // group_id
      memcpy( ptr, encodeMeta->group_id, sizeof(HASH) );
      ptr += sizeof(HASH);
      
      // size
      netSafe = htonl(encodeMeta->size);
      memcpy( ptr, &netSafe, sizeof(uint32_t) );
      ptr += sizeof(uint32_t);
      
      *(ptr++) = '\0';
      
      encodedSize = numBytes;      
    }
    *size = encodedSize;
  }

  return( encoded );
}


/* --------------------------------------------------------------------------- */
	
static _msg_meta_z* msg_meta_decode( uint8_t* buffer, uint32_t  bufferSize )
{
  _msg_meta_z*  newMeta = NULL;
  uint8_t*  ptr = buffer;
  uint32_t  dataLen;

  if( (NULL != buffer) && (0 < bufferSize) )
  {
    if( (NULL != ptr) && ( 0 < bufferSize) &&
	(NULL != (newMeta = (_msg_meta_z*) create_msg_meta( ))) )
    {
      // version
      uint32_t netSafe;
      memcpy( &netSafe, ptr, sizeof(uint32_t) );
      newMeta->version = ntohl( netSafe );
      ptr += sizeof(uint32_t);
      
      // node id
      memcpy( &(newMeta->node_id), ptr, sizeof(Key) );
      ptr += sizeof(Key);

      // send_key
      memcpy( &(newMeta->send_key), ptr, sizeof(Key) );
      ptr += sizeof(Key);

      // user_id and file_name
      memcpy( newMeta->user_id, ptr, sizeof(HASH) );
      ptr += sizeof(HASH);

      memcpy( newMeta->file_name, ptr, sizeof(char) * 64 );
      ptr += sizeof(char) * 64;

      memcpy( newMeta->group_id, ptr, sizeof(HASH) );
      ptr += sizeof(HASH);
      
      // size
      memcpy( &netSafe, ptr, sizeof(uint32_t) );
      dataLen = newMeta->size = ntohl( netSafe );
      ptr += sizeof(uint32_t);
    }
  }

  return( newMeta );
}

/* --------------------------------------------------------------------------- */


static _msg_data_z * create_msg_data()
{
  _msg_data_z* newMsgData = NULL;

  if( NULL != (newMsgData = (_msg_data_z*)malloc( sizeof(_msg_data_z) )) )
  {
    memset( newMsgData, 0, sizeof(_msg_data_z) );
  }
  
  return( newMsgData );
}

static _msg_meta_z * create_msg_meta()
{
  _msg_meta_z* newMsgMeta = NULL;

  if( NULL != (newMsgMeta = (_msg_meta_z*)malloc( sizeof(_msg_meta_z) )) )
  {
    memset( newMsgMeta, 0, sizeof(_msg_meta_z) );
  }
  
  return( newMsgMeta );
}



/* --------------------------------------------------------------------------- */

static void quartz_store_upcall_fwd(Key **kp, Message **mp, ChimeraHost **hp) 
{

  // N/A for this test

}

/* --------------------------------------------------------------------------- */

extern void quartz_store_upcall_del(Key *k, Message *m) 
{
#ifdef DEBUG
  //printf("Delivered type %d to %x\n", m->type, *k);
#endif

  if (m->type == CTRL_CFILE_STORE) { 

    // ***********************************************************************
    // NODE REQUESTS TO STORE DATA
    // ***********************************************************************
    
    //pthread_mutex_lock(&mutex_cfile_s);

#ifdef DEBUG	
    //printf("Received a CTRL_CFILE_STORE (NODE REQUESTS TO STORE DATA)\n");
#endif
    
    pthread_mutex_lock(&mutex_inode);

    // decode received message
    _msg_data_z *msgData = msg_data_decode(m->payload, m->size);

#ifdef DEBUG
    //	dumpBytes(stderr, m->payload, m->size , " [stor] ");
#endif
    
    int i = 0, exists = 0, inode_num = 0, curr_size = 0, j = 0, block_size = 0;
    int new_file = 0;
    Key user_key;
    inode *temp_inode, *new_inode;
    block *myblock, *temp_block;
    char write_buf[BLOCKSIZE];

    // convert user_id and file_name to char array
    char user_id[41];
    char group_id[41];
    char file_id[41];
    sprintf(user_id, "%s", sha1_hashToString(&(msgData->user_id)));
    sprintf(group_id, "%s", sha1_hashToString(&(msgData->group_id)));
    sprintf(file_id, "%s", sha1_hashToString(&(msgData->file_name)));
    user_id[40]  = '\0';
    group_id[40] = '\0';
    file_id[40]  = '\0';
   
    /*    
    fprintf(stderr, "STORING (MAIN)...\n");
    fprintf(stderr, "-----------------------------\n");
    fprintf(stderr, "USER_ID:  %s\n", user_id);
    fprintf(stderr, "GROUP_ID: %s\n", group_id);
    fprintf(stderr, "FILE_ID:  %s\n", file_id);
    fprintf(stderr, "VERSION:  %d\n", msgData->version);
    fprintf(stderr, "SIZE:     %d\n", msgData->size);
    */
    
    temp_inode = NULL;
    new_inode  = NULL;
    myblock    = NULL;

    // !!! UPDATE QFS !!!
    new_file = 0;

    // check if file already exists in the file system
    inode_num = inode_get(&temp_inode, user_id, group_id, file_id);
    if (inode_num > -1) // file exists
    {
      if (msgData->version > temp_inode->version)
      {
	myblock = (block *)malloc(sizeof(block));
	memset(myblock->content, '\0', BLOCKSIZE);
	
	// clear the blocks list from meta.sto side
	for (i = 0; i < INODE_BLOCKS; i++)
	{
	  // clear the blocks from file.sto side
	  if (temp_inode->blocks[i] != QFS_NULL)
	  {
	    myblock->num = temp_inode->blocks[i];
	    block_free(&myblock);
	    temp_inode->blocks[i] = QFS_NULL;
	  }
	  else
	    break;
	}

	free(myblock);
      }
      else
      {
	//pthread_mutex_unlock(&mutex_cfile_s);

	// if the sent version is less, that sending node must
	// be updated with the most recent version (SFILE_STORE)
      }
    }
    else
    {
      // if not, then add the file to the file system
      inode_init(&temp_inode);
      inode_num = inode_new();
      inode_set(&temp_inode, user_id, group_id, file_id, msgData->version, msgData->size, msgData->send_key);
      
      new_file = 1;
    }

    // either way, both cases will write new blocks out and store it in inode
    // first though, must make sure version received is newer than storage
    if ((msgData->version >= temp_inode->version) || new_file)
    {
      temp_inode->version = msgData->version;
    
      curr_size = 0;
      
      i = 0;
      while (curr_size <= msgData->size)
      {
	//fprintf(stderr, "writing block[%d].. ", i);
	memset(write_buf, '\0', BLOCKSIZE);
	
	// to take care of the last partial block size
	if (msgData->size - curr_size <= BLOCKSIZE)
	{
	  memcpy(write_buf, msgData->data + curr_size, msgData->size - curr_size);
	  block_size = msgData->size - curr_size;
	}
	else
	{
	  memcpy(write_buf, msgData->data + curr_size, BLOCKSIZE);
	  block_size = BLOCKSIZE;
	}
	
	// write block out to file system
	myblock = NULL;
	block_new(&myblock);
	memcpy(myblock->content, write_buf, BLOCKSIZE);
	
	myblock->size = block_size;
	block_write(&myblock);
	
	// save meta data information
	if (i < INODE_BLOCKS)
	{
	  //fprintf(stderr, "inode->blocks[%d] = %d (%d)\n", i, myblock->num, myblock->size);
	  fflush(stderr);
	  temp_inode->blocks[i] = myblock->num;
	}
	else
	{
	  // should not go here
	}

	i++;
	curr_size += BLOCKSIZE;
      }
      
      temp_inode->version = msgData->version;
      inode_write(&temp_inode, inode_num);

      ChimeraHost **neighbors = (ChimeraHost **) route_neighbors(NUM_FILES);
      
      user_incr_file_inodes(1);

      char *s_sendbuf = (char *)malloc(m->size);
      memset(s_sendbuf, '\0', m->size);
      memcpy(s_sendbuf, m->payload, m->size);
      
      for (i = 0; i < NUM_FILES; i++)
      {
	if (neighbors[i] == NULL)
	  break;
	
	if ((neighbors[i])->key != deviceInfo->nodeKey &&
	    (neighbors[i])->key != 0)
	{
	  chimera_send((neighbors[i])->key, CTRL_SFILE_STORE, m->size, s_sendbuf);
	}
      }

      pthread_mutex_unlock(&mutex_inode);


      //printf("FINISHED CTRL_CFILE_STORE\n");
      
      free(msgData);
      free(s_sendbuf);
    }
  }
  else if (m->type == CTRL_META_STORE) { 

    // ***********************************************************************
    // NODE REQUESTS TO STORE META DATA
    // ***********************************************************************

    //pthread_mutex_lock(&mutex_meta_s);

#ifdef DEBUG	
    //printf("Received a CTRL_META_STORE message\n");
#endif

    pthread_mutex_lock(&mutex_table);

    _msg_meta_z msg_meta;
    memset(&msg_meta, '\0', sizeof(_msg_meta_z));
    memcpy(&msg_meta, m->payload, sizeof(_msg_meta_z));


    log_exec(LOG_RECV, msg_meta.node_id, LOG_SMETA, LOG_NA);


    // decode received message
    //_msg_meta_z *msgMeta = msg_meta_decode(m->payload, m->size);
    
#ifdef DEBUG
    //	dumpBytes(stderr, m->payload, m->size , " [stor] ");
#endif
	
    _hb_table_z *temp_hb = my_heartbeat;
    int i = 0, exists = 0;
    Key user_key;

    char user_curr[41];
    char file_curr[97];
    char group_curr[41];

    /*        
    fprintf(stderr, "RECEIVED META_STORE FOR:\n");
    fprintf(stderr, "  user_id:   %s\n", msg_meta.user_id);
    fprintf(stderr, "  file_name: %s\n", msg_meta.file_name);
    fprintf(stderr, "  group_id:  %s\n", msg_meta.group_id);
    fprintf(stderr, "  size:      %d\n", msg_meta.size);
    fprintf(stderr, "  version:   %d\n", msg_meta.version);
    */
    

    _meta_user_z *curr_user   = NULL;
    _meta_user_z *prev_user   = NULL;
    _info_group_z *curr_group = NULL;
    _info_group_z *prev_group = NULL;
    _meta_local_z *curr_file  = NULL;
    _meta_local_z *prev_file  = NULL;

    int exists_user  = 0;
    int exists_file  = 0;
    int exists_group = 0;

    exists_user = 0;
    
    if (meta_user != NULL)
    {
      curr_user = meta_user;
      while (curr_user != NULL)
      {
	sprintf(user_curr, "%s", curr_user->user_id);
	user_curr[40] = '\0';
	
	if (strcmp(user_curr, msg_meta.user_id) == 0) // user already exists
	{
	  exists_user = 1;
	  
	  if (curr_user->groups != NULL)
	  {
	    curr_group = curr_user->groups;
	    exists_group = 0;
	    while (curr_group != NULL)
	    {
	      sprintf(group_curr, "%s", curr_group->group_id);
	      group_curr[40] = '\0';
	      
	      if (strcmp(group_curr, msg_meta.group_id) == 0)
	      {
		exists_group = 1;
		
		curr_file = curr_group->files;
		exists_file = 0;
		while (curr_file != NULL)
		{
		  memset(file_curr, '\0', 95);
		  sprintf(file_curr, "%s", curr_file->file_name);
		  file_curr[95] = '\0';
		  
		  if (strcmp(file_curr, msg_meta.file_name) == 0)
		  {
		    exists_file = 1;
		    
		    // the version must be greater than the already stored
		    if (msg_meta.version > curr_file->version)
		    {
		      curr_file->version = msg_meta.version;
		      curr_file->size    = msg_meta.size;
		    }
		  
		    // if not, then that means user needs to update first
		    // before user can officially commit changes
		    else
		    {
		      /*
		      printf("  Your version is too old (%d < %d).  Must UPDATE first.",
			     msg_meta.version - 1, curr_file->version);
		      */
		      
		      // do not update meta data
		    }
		    
		    break;
		  }
		  
		  curr_file = curr_file->next;
		}
		
		// if file doesn't exist, need to add it to curr_group
		if (!exists_file)
		{
		  if (curr_group->files != NULL)
		  {
		    curr_file = curr_group->files;
		    prev_file = curr_group->files;
		    while (curr_file != NULL)
		    {
		      prev_file = curr_file;
		      curr_file = curr_file->next;
		    }
		    
		    _meta_local_z *new_file  = NULL;
		    
		    new_file = (_meta_local_z *) malloc(sizeof(_meta_local_z));
		    new_file->next = NULL;
		    new_file->type      = 'a';
		    
		    memset(new_file->file_name, '\0', 95);
		    sprintf(new_file->file_name, "%s", msg_meta.file_name);
		    new_file->file_name[95]= '\0';
		    
		    new_file->version = msg_meta.version;
		    new_file->size    = msg_meta.size;

		    prev_file->next = new_file;
		  }
		}
		
		break;
	      }
	      
	      if (curr_group->next != NULL)
		curr_group = curr_group->next;
	      else
		break;
	    }
	  }
	  
	  if (!exists_group)
	  {
	    //pthread_mutex_lock(&mutex_table);
	    
	    if (curr_user->groups != NULL)
	    {
	      curr_group = curr_user->groups;
	      prev_group = curr_user->groups;
	      
	      while (curr_group != NULL)
	      {
		prev_group = curr_group;
		curr_group = curr_group->next;
	      }
	      
	      curr_group = (_info_group_z *) malloc(sizeof(_info_group_z));
	      curr_group->next  = NULL;
	      curr_group->files = NULL;

	      prev_group->next = curr_group;
	    }
	    else
	    {
	      curr_user->groups = (_info_group_z *) malloc(sizeof(_info_group_z));
	      curr_group = curr_user->groups;
	      curr_group->next  = NULL;
	      curr_group->files = NULL;
	    }

	    memset(curr_group->group_id, '\0', 40);
	    sprintf(curr_group->group_id, "%s", msg_meta.group_id);
	    curr_group->group_id[40] = '\0';
	    
	    curr_group->files = (_meta_local_z *) malloc(sizeof(_meta_local_z));
	    curr_group->files->next = NULL;
	    curr_group->files->type = 'a';
	  
	    memset(curr_group->files->file_name, '\0', 95);
	    sprintf(curr_group->files->file_name, "%s", msg_meta.file_name);
	    curr_group->files->file_name[95] = '\0';

	    curr_group->files->version = msg_meta.version;
	    curr_group->files->size    = msg_meta.size;
	  }
	  
	  break;
	}
	
	curr_user = curr_user->next;
      }
    }

    if (!exists_user)
    {
      //pthread_mutex_lock(&mutex_table);

      if (meta_user != NULL)
      {
	curr_user = meta_user;
	prev_user = meta_user;

	while (curr_user != NULL)
	{
	  prev_user = curr_user;
	  curr_user = curr_user->next;
	}

	curr_user = (_meta_user_z *) malloc(sizeof(_meta_user_z));
	curr_user->groups = NULL;
	curr_user->next   = NULL;

	prev_user->next = curr_user;
      }
      else
      {
	meta_user = (_meta_user_z *) malloc(sizeof(_meta_user_z));
	meta_user->groups = NULL;
	meta_user->next   = NULL;

	curr_user = meta_user;
      }

      memset(curr_user->user_id, '\0', 40);
      sprintf(curr_user->user_id, "%s", msg_meta.user_id);
      curr_user->user_id[40] = '\0';

      memcpy( &(curr_user->send_key), &(msg_meta.send_key), sizeof(Key) );

      curr_user->pos = 5;

      if (curr_user->groups == NULL)
      {
	curr_user->groups = (_info_group_z *) malloc(sizeof(_info_group_z));
	curr_user->groups->files = NULL;
	curr_user->groups->next  = NULL;
      }

      curr_group = curr_user->groups;
      //curr_group = (_info_group_z *) malloc(sizeof(_info_group_z));

      memset(curr_group->group_id, '\0', 40);
      sprintf(curr_group->group_id, "%s", msg_meta.group_id);
      curr_group->group_id[40] = '\0';

      curr_group->files = (_meta_local_z *) malloc(sizeof(_meta_local_z));
      curr_group->files->next = NULL;
      curr_group->files->type = 'a';

      memset(curr_group->files->file_name, '\0', 95);
      sprintf(curr_group->files->file_name, "%s", msg_meta.file_name);
      curr_group->files->file_name[95] = '\0';

      curr_group->files->version = msg_meta.version;
      curr_group->files->size    = msg_meta.size;
    }

    pthread_mutex_unlock(&mutex_table);

    // Replicate out to right and left four neighbors
    ChimeraHost **neighbors = (ChimeraHost **) route_neighbors(NUM_METAS);

    char *s_sendbuf = (char *)malloc(sizeof(_msg_meta_z));
    memset(s_sendbuf, '\0', sizeof(_msg_meta_z));
    memcpy(s_sendbuf, &msg_meta, sizeof(_msg_meta_z));
    
    for (i = 0; i < NUM_METAS; i++)
    {
      if (neighbors[i] == NULL)
	break;
      
      if ((neighbors[i])->key != deviceInfo->nodeKey &&
	  (neighbors[i])->key != 0)
      {
	chimera_send((neighbors[i])->key, CTRL_META_COPY, sizeof(_msg_meta_z), s_sendbuf);
      }
    }

    free(s_sendbuf);

    //printf("FINISHED a CTRL_META_STORE message\n");
  }
  else if (m->type == CTRL_DEL_FILE) { 

    // ***********************************************************************
    // NODE REQUESTS TO DELETE DATA
    // ***********************************************************************

    //printf("RECEIVED CTRL_DEL_FILE size(%d)\n", m->size);

    // decode received message
    _msg_data_z *msgData = msg_data_decode(m->payload, m->size);
    
    int i = 0, inode_num = 0;
    inode *temp_inode;
    block *myblock;

    // convert user_id and file_name to char array

    char user_id[41]  = { '\0' };
    char group_id[41] = { '\0' };
    char file_id[41]  = { '\0' };
    sprintf(user_id, "%s", sha1_hashToString(&(msgData->user_id)));
    sprintf(group_id, "%s", sha1_hashToString(&(msgData->group_id)));
    sprintf(file_id, "%s", sha1_hashToString(&(msgData->file_name)));
    user_id[40]  = '\0';
    group_id[40] = '\0';
    file_id[40]  = '\0';
    
    /*
    fprintf(stderr, "DELETING...\n");
    fprintf(stderr, "-----------------------------\n");
    fprintf(stderr, "USER_ID:  %s\n", user_id);
    fprintf(stderr, "GROUP_ID: %s\n", group_id);
    fprintf(stderr, "FILE_ID:  %s\n", file_id);
    */
    
    temp_inode = NULL;
    myblock    = NULL;

    pthread_mutex_lock(&mutex_inode);

    // check if file already exists in the file system
    inode_num = inode_get(&temp_inode, user_id, group_id, file_id);
    if (inode_num > -1) // file exists
    {
      myblock = (block *)malloc(sizeof(block));
      memset(myblock->content, '\0', BLOCKSIZE);

      // clear all the blocks on the file.sto side
      for (i = 0; i < INODE_BLOCKS; i++)
      {
	if (temp_inode->blocks[i] != QFS_NULL)
	{
	  myblock->num = temp_inode->blocks[i];
	  block_free(&myblock);
	  temp_inode->blocks[i] = QFS_NULL;
	}
	else
	  break;
      }

      free(myblock);

      // decrease user statistics
      user_incr_file_inodes(0);
      // user_incr_file_size(temp_inode->size, 0);

      // reset this inode parameters and write back out to system
      inode_clear(&temp_inode);
      inode_write(&temp_inode, inode_num);
    }
    else
    {
      //fprintf(stderr, "file not found\n");
    }

    pthread_mutex_unlock(&mutex_inode);

    // retrieving twice the number of neighbors it would normally
    // need to make sure file is deleted on "every" node in the
    // system in case a node came on outside the original range
    // but ended up back in the original range after other nodes leave
    ChimeraHost **neighbors = (ChimeraHost **) route_neighbors(NUM_FILES * 2);

    char *s_sendbuf = (char *)malloc(m->size);
    memset(s_sendbuf, '\0', m->size);
    memcpy(s_sendbuf, m->payload, m->size);
    
    //fprintf(stderr, "my key[%x]\n", deviceInfo->nodeKey);

    for (i = 0; i < NUM_FILES * 2; i++)
    {
      if (neighbors[i] == NULL)
	break;

      if ((neighbors[i])->key != deviceInfo->nodeKey &&
	  (neighbors[i])->key != 0)
      {
	chimera_send((neighbors[i])->key, CTRL_DEL_FILE_COPY, m->size, s_sendbuf);
      }
    }

    // printf("FINISHED CTRL_DEL_FILE size\n");

    free(msgData);
    free(s_sendbuf);
  }

  else if (m->type == CTRL_DEL_FILE_COPY) { 

    // ***********************************************************************
    // NODE REQUESTS TO DELETE DATA (COPY)
    // ***********************************************************************

    //printf("RECEIVED CTRL_DEL_FILE_COPY size(%d)\n", m->size);


    // decode received message
    _msg_data_z *msgData = msg_data_decode(m->payload, m->size);
    
    int i = 0, inode_num = 0;
    inode *temp_inode;
    block *myblock;

    // convert user_id and file_name to char array
    char user_id[41]  = { '\0' };
    char group_id[41] = { '\0' };
    char file_id[41]  = { '\0' };
    sprintf(user_id, "%s", sha1_hashToString(&(msgData->user_id)));
    sprintf(group_id, "%s", sha1_hashToString(&(msgData->group_id)));
    sprintf(file_id, "%s", sha1_hashToString(&(msgData->file_name)));
    user_id[40]  = '\0';
    group_id[40] = '\0';
    file_id[40]  = '\0';
    
    /*
    fprintf(stderr, "DELETING...\n");
    fprintf(stderr, "-----------------------------\n");
    fprintf(stderr, "USER_ID:  %s\n", user_id);
    fprintf(stderr, "GROUP_ID: %s\n", group_id);
    fprintf(stderr, "FILE_ID:  %s\n", file_id);
    */
    
    temp_inode = NULL;
    myblock    = NULL;

    pthread_mutex_lock(&mutex_inode);

    // check if file already exists in the file system
    inode_num = inode_get(&temp_inode, user_id, group_id, file_id);
    if (inode_num > -1) // file exists
    {
      myblock = (block *)malloc(sizeof(block));
      memset(myblock->content, '\0', BLOCKSIZE);

      // clear all the blocks on the file.sto side
      for (i = 0; i < INODE_BLOCKS; i++)
      {
	if (temp_inode->blocks[i] != QFS_NULL)
	{
	  myblock->num = temp_inode->blocks[i];
	  block_free(&myblock);
	  temp_inode->blocks[i] = QFS_NULL;
	}
	else
	  break;
      }

      free(myblock);

      // decrease user statistics
      user_incr_file_inodes(0);
      //      user_incr_file_size(temp_inode->size, 0);

      // reset this inode parameters and write back out to system
      inode_clear(&temp_inode);
      inode_write(&temp_inode, inode_num);
    }
    else
    {
      //fprintf(stderr, "file not found\n");
    }

    //printf("FINISHED CTRL_DEL_FILE_COPY size\n");

    pthread_mutex_unlock(&mutex_inode);

    free(msgData);
  }

  else if (m->type == CTRL_DEL_FILE_GROUP) { 

    // ***********************************************************************
    // NODE REQUESTS TO DELETE DATA
    // ***********************************************************************

    //printf("RECEIVED CTRL_DEL_FILE_GROUP size(%d)\n", m->size);

    // decode received message
    _msg_data_z *msgData = msg_data_decode(m->payload, m->size);
    
    int i = 0, inode_num = 0;
    inode *temp_inode;
    block *myblock;

    // convert user_id and file_name to char array

    char user_id[41]  = { '\0' };
    char group_id[41] = { '\0' };
    char file_id[41]  = { '\0' };
    sprintf(user_id, "%s", sha1_hashToString(&(msgData->user_id)));
    sprintf(group_id, "%s", sha1_hashToString(&(msgData->group_id)));
    sprintf(file_id, "%s", sha1_hashToString(&(msgData->file_name)));
    user_id[40]  = '\0';
    group_id[40] = '\0';
    file_id[40]  = '\0';
    
    /*
    fprintf(stderr, "DELETING...\n");
    fprintf(stderr, "-----------------------------\n");
    fprintf(stderr, "USER_ID:  %s\n", user_id);
    fprintf(stderr, "GROUP_ID: %s\n", group_id);
    */
    
    temp_inode = NULL;
    myblock    = NULL;

    pthread_mutex_lock(&mutex_inode);

    // check if file already exists in the file system
    inode_num = inode_get_g(&temp_inode, user_id, group_id);

    // must iterate through all possible storage
    // matching both user_id and group_id to
    // effectively get rid of the entire group
    while (inode_num > -1)
    {
      if (inode_num > -1) // file exists
      {
	myblock = (block *)malloc(sizeof(block));
	memset(myblock->content, '\0', BLOCKSIZE);
	
	// clear all the blocks on the file.sto side
	for (i = 0; i < INODE_BLOCKS; i++)
	{
	  if (temp_inode->blocks[i] != QFS_NULL)
	  {
	    myblock->num = temp_inode->blocks[i];
	    block_free(&myblock);
	    temp_inode->blocks[i] = QFS_NULL;
	  }
	  else
	    break;
	}

	free(myblock);
	
	// decrease user statistics
	user_incr_file_inodes(0);
	
	// reset this inode parameters and write back out to system
	inode_clear(&temp_inode);
	inode_write(&temp_inode, inode_num);
      }
      else
      {
	//fprintf(stderr, "file not found\n");
	break;
      }

      // check if there are any more in this group
      inode_num = inode_get_g(&temp_inode, user_id, group_id);
    }

    pthread_mutex_unlock(&mutex_inode);

    // retrieving twice the number of neighbors it would normally
    // need to make sure file is deleted on "every" node in the
    // system in case a node came on outside the original range
    // but ended up back in the original range after other nodes leave
    ChimeraHost **neighbors = (ChimeraHost **) route_neighbors(NUM_FILES * 2);

    char *s_sendbuf = (char *)malloc(m->size);
    memset(s_sendbuf, '\0', m->size);
    memcpy(s_sendbuf, m->payload, m->size);
    
    //fprintf(stderr, "my key[%x]\n", deviceInfo->nodeKey);

    for (i = 0; i < NUM_FILES * 2; i++)
    {
      if (neighbors[i] == NULL)
	break;

      if ((neighbors[i])->key != deviceInfo->nodeKey &&
	  (neighbors[i])->key != 0)
      {
	chimera_send((neighbors[i])->key, CTRL_DEL_FILE_GROUP_COPY, m->size, s_sendbuf);
      }
    }

    // printf("FINISHED CTRL_DEL_FILE size\n");

    free(msgData);
    free(s_sendbuf);
  }
  else if (m->type == CTRL_DEL_FILE_GROUP_COPY) { 

    // ***********************************************************************
    // NODE REQUESTS TO DELETE DATA
    // ***********************************************************************

    //printf("RECEIVED CTRL_DEL_FILE_GROUP_COPY\n");

    // decode received message
    _msg_data_z *msgData = msg_data_decode(m->payload, m->size);
    
    int i = 0, inode_num = 0;
    inode *temp_inode;
    block *myblock;

    // convert user_id and file_name to char array

    char user_id[41]  = { '\0' };
    char group_id[41] = { '\0' };
    char file_id[41]  = { '\0' };
    sprintf(user_id, "%s", sha1_hashToString(&(msgData->user_id)));
    sprintf(group_id, "%s", sha1_hashToString(&(msgData->group_id)));
    sprintf(file_id, "%s", sha1_hashToString(&(msgData->file_name)));
    user_id[40]  = '\0';
    group_id[40] = '\0';
    file_id[40]  = '\0';
        
    temp_inode = NULL;
    myblock    = NULL;

    pthread_mutex_lock(&mutex_inode);

    // check if file already exists in the file system
    inode_num = inode_get_g(&temp_inode, user_id, group_id);

    // must iterate through all possible storage
    // matching both user_id and group_id to
    // effectively get rid of the entire group
    while (inode_num > -1)
    {
      if (inode_num > -1) // file exists
      {
	myblock = (block *)malloc(sizeof(block));
	memset(myblock->content, '\0', BLOCKSIZE);
	
	// clear all the blocks on the file.sto side
	for (i = 0; i < INODE_BLOCKS; i++)
	{
	  if (temp_inode->blocks[i] != QFS_NULL)
	  {
	    myblock->num = temp_inode->blocks[i];
	    block_free(&myblock);
	    temp_inode->blocks[i] = QFS_NULL;
	  }
	  else
	    break;
	}

	free(myblock);
	
	// decrease user statistics
	user_incr_file_inodes(0);
	
	// reset this inode parameters and write back out to system
	inode_clear(&temp_inode);
	inode_write(&temp_inode, inode_num);
      }
      else
      {
	//fprintf(stderr, "file not found\n");
	break;
      }

      // check if there are any more in this group
      inode_num = inode_get_g(&temp_inode, user_id, group_id);
    }

    pthread_mutex_unlock(&mutex_inode);

    // printf("FINISHED CTRL_DEL_FILE_GROUP_COPY\n");

    free(msgData);
  }

  else if (m->type == CTRL_DEL_META) { 

    // ***********************************************************************
    // NODE REQUESTS TO DELETE META DATA
    // ***********************************************************************

    //printf("Received a CTRL_DEL_META message\n");

    _msg_meta_z msg_meta;
    memset(&msg_meta, '\0', sizeof(_msg_meta_z));
    memcpy(&msg_meta, m->payload, sizeof(_msg_meta_z));

    int i = 0, exists = 0;
    Key user_key;

    char user_curr[41]  = { '\0' };
    char file_curr[97]  = { '\0' };
    char group_curr[41] = { '\0' };
    
    /*
    fprintf(stderr, "RECEIVED DEL_META FOR:\n");
    fprintf(stderr, "  user_id:   %s\n", msg_meta.user_id);
    fprintf(stderr, "  file_name: %s\n", msg_meta.file_name);
    fprintf(stderr, "  group_id:  %s\n", msg_meta.group_id);
    */

    _meta_user_z *curr_user   = NULL;
    _meta_user_z *prev_user   = NULL;
    _info_group_z *curr_group = NULL;
    _info_group_z *prev_group = NULL;
    _meta_local_z *curr_file  = NULL;
    _meta_local_z *prev_file  = NULL;

    pthread_mutex_lock(&mutex_table);

    if (meta_user != NULL)
    {
      curr_user = meta_user;
      prev_user = meta_user;

      while (curr_user != NULL)
      {
	sprintf(user_curr, "%s", curr_user->user_id);
	user_curr[40] = '\0';
	
	if (strcmp(user_curr, msg_meta.user_id) == 0) // user already exists
	{
	  if (curr_user->groups != NULL)
	  {
	    curr_group = curr_user->groups;
	    prev_group = curr_user->groups;

	    while (curr_group != NULL)
	    {
	      sprintf(group_curr, "%s", curr_group->group_id);
	      group_curr[40] = '\0';
	      
	      if (strcmp(group_curr, msg_meta.group_id) == 0)
	      {
		curr_file = curr_group->files;
		prev_file = curr_group->files;

		while (curr_file != NULL)
		{
		  memset(file_curr, '\0', 95);
		  sprintf(file_curr, "%s", curr_file->file_name);
		  file_curr[95] = '\0';
		  
		  if (strcmp(file_curr, msg_meta.file_name) == 0)
		  {
		    if (strcmp(file_curr, curr_group->files->file_name) == 0)
		    {
		      if (curr_file->next != NULL)
		      {
			curr_group->files = curr_file->next;
			free(curr_file);
		      }
		      else
		      {
			curr_group->files = NULL;

			if (curr_group->next != NULL)
			{
			  prev_group->next = curr_group->next;
			  free(curr_group);
			}
			else
			  prev_group->next = NULL;

			if (strcmp(group_curr, curr_user->groups->group_id) == 0)
			{
			  free(curr_user->groups);
			  curr_user->groups = NULL;

			  if (curr_user != meta_user)
			  {
			    if (curr_user->next != NULL)
			    {
			      prev_user->next = curr_user->next;
			      free(curr_user);
			    }
			    else
			      prev_user->next = NULL;
			  }
			  else
			  {
			    if (curr_user->next == NULL)
			    {
			      free(curr_user);
			      meta_user = NULL;
			    }
			    else
			      meta_user = curr_user->next;
			  }
			}
		      }
		    }
		    else
		    {
		      if (curr_file->next != NULL)
			prev_file->next = curr_file->next;
		      else
			prev_file->next = NULL;

		      free(curr_file);
		    }

		    break;
		  }
		  
		  if (curr_file != curr_group->files)
		    prev_file = prev_file->next;
		  
		  curr_file = curr_file->next;
		}
		
		break;
	      }

	      if (curr_group != curr_user->groups)
		prev_group = prev_group->next;
	      
	      if (curr_group->next != NULL)
		curr_group = curr_group->next;
	      else
		break;
	    }
	    
	    break;
	  }
	  
	  break;
	}

	if (curr_user != meta_user)
	  prev_user = prev_user->next;
	
	curr_user = curr_user->next;
      }
    }

    // replicate deletion request out to NUM_METAS * 2
    // neighbors to "ensure" that this meta data is
    // removed from the Quartz network
    ChimeraHost **neighbors = (ChimeraHost **) route_neighbors(NUM_METAS * 2);

    char *s_sendbuf = (char *)malloc(sizeof(_msg_meta_z));
    memset(s_sendbuf, '\0', sizeof(_msg_meta_z));
    memcpy(s_sendbuf, &msg_meta, sizeof(_msg_meta_z));
    
    for (i = 0; i < NUM_METAS * 2; i++)
    {
      if (neighbors[i] == NULL)
	break;
      
      if ((neighbors[i])->key != deviceInfo->nodeKey &&
	  (neighbors[i])->key != 0)
      {
	chimera_send((neighbors[i])->key, CTRL_DEL_META_COPY,
		     sizeof(_msg_meta_z), s_sendbuf);
      }
    }

    ChimeraHost **lookup = (ChimeraHost **)
      route_lookup(msg_meta.send_key, NUM_METAS * 2, 1);
    
    for (i = 0; i < NUM_METAS * 2; i++)
    {
      if (lookup[i] == NULL)
	break;
      
      if ((lookup[i])->key != deviceInfo->nodeKey &&
	  (lookup[i])->key != 0)
      {
	chimera_send((lookup[i])->key, CTRL_DEL_META_COPY,
		     sizeof(_msg_meta_z), s_sendbuf);
      }
    }

    if (neighbors != NULL)
      free(neighbors);
    if (lookup != NULL)
      free(lookup);

    pthread_mutex_unlock(&mutex_table);

    fflush(stdout);
    //printf("FINISHED a CTRL_DEL_META message\n");
  }
  else if (m->type == CTRL_DEL_META_COPY) { 

    // ***********************************************************************
    // NODE REQUESTS TO DELETE META DATA (COPY)
    // ***********************************************************************

    // printf("Received a CTRL_DEL_META_COPY message\n");

    _msg_meta_z msg_meta;
    memset(&msg_meta, '\0', sizeof(_msg_meta_z));
    memcpy(&msg_meta, m->payload, sizeof(_msg_meta_z));

    int i = 0, exists = 0;
    Key user_key;

    char user_curr[41]  = { '\0' };
    char file_curr[96]  = { '\0' };
    char group_curr[41] = { '\0' };
    
    /*
    fprintf(stderr, "RECEIVED DEL_META_COPY FOR:\n");
    fprintf(stderr, "  user_id:   %s\n", msg_meta.user_id);
    fprintf(stderr, "  file_name: %s\n", msg_meta.file_name);
    fprintf(stderr, "  group_id:  %s\n", msg_meta.group_id);
    */

    _meta_user_z  *curr_user  = NULL;
    _meta_user_z  *prev_user  = NULL;
    _info_group_z *curr_group = NULL;
    _info_group_z *prev_group = NULL;
    _meta_local_z *curr_file  = NULL;
    _meta_local_z *prev_file  = NULL;

    pthread_mutex_lock(&mutex_table);

    if (meta_user != NULL)
    {
      curr_user = meta_user;
      prev_user = meta_user;

      while (curr_user != NULL)
      {
	sprintf(user_curr, "%s", curr_user->user_id);
	user_curr[40] = '\0';
	
	if (strcmp(user_curr, msg_meta.user_id) == 0) // user already exists
	{
	  if (curr_user->groups != NULL)
	  {
	    curr_group = curr_user->groups;
	    prev_group = curr_user->groups;

	    while (curr_group != NULL)
	    {
	      sprintf(group_curr, "%s", curr_group->group_id);
	      group_curr[40] = '\0';
	      
	      if (strcmp(group_curr, msg_meta.group_id) == 0)
	      {
		curr_file = curr_group->files;
		prev_file = curr_group->files;

		while (curr_file != NULL)
		{
		  memset(file_curr, '\0', 95);
		  sprintf(file_curr, "%s", curr_file->file_name);
		  file_curr[95] = '\0';
		  
		  if (strcmp(file_curr, msg_meta.file_name) == 0)
		  {
		    if (strcmp(file_curr, curr_group->files->file_name) == 0)
		    {
		      if (curr_file->next != NULL)
		      {
			curr_group->files = curr_file->next;
			free(curr_file);
		      }
		      else
		      {
			curr_group->files = NULL;

			if (curr_group->next != NULL)
			{
			  prev_group->next = curr_group->next;
			  free(curr_group);
			}
			else
			  prev_group->next = NULL;

			if (strcmp(group_curr, curr_user->groups->group_id) == 0)
			{
			  free(curr_user->groups);
			  curr_user->groups = NULL;

			  if (curr_user != meta_user)
			  {
			    if (curr_user->next != NULL)
			    {
			      prev_user->next = curr_user->next;
			      free(curr_user);
			    }
			    else
			      prev_user->next = NULL;
			  }
			  else
			  {
			    if (curr_user->next == NULL)
			    {
			      free(curr_user);
			      meta_user = NULL;
			    }
			    else
			      meta_user = curr_user->next;
			  }
			}
		      }
		    }
		    else
		    {
		      if (curr_file->next != NULL)
			prev_file->next = curr_file->next;
		      else
			prev_file->next = NULL;

		      free(curr_file);
		    }

		    break;
		  }
		  
		  if (curr_file != curr_group->files)
		    prev_file = prev_file->next;
		  
		  curr_file = curr_file->next;
		}
		
		break;
	      }

	      if (curr_group != curr_user->groups)
		prev_group = prev_group->next;
	      
	      curr_group = curr_group->next;
	    }
	    
	    break;
	  }
	  
	  break;
	}

	if (curr_user != meta_user)
	  prev_user = prev_user->next;
	
	curr_user = curr_user->next;
      }
    }

    pthread_mutex_unlock(&mutex_table);

    //printf("FINISHED a CTRL_DEL_META_COPY message\n");
  }
  else if (m->type == CTRL_DEL_META_GROUP) { 

    // ***********************************************************************
    // NODE REQUESTS TO DELETE META DATA
    // ***********************************************************************

    //printf("Received a CTRL_DEL_META_GROUP\n");

    _msg_meta_z msg_meta;
    memset(&msg_meta, '\0', sizeof(_msg_meta_z));
    memcpy(&msg_meta, m->payload, sizeof(_msg_meta_z));

    int i = 0, exists = 0;
    Key user_key;

    char user_curr[41]  = { '\0' };
    char file_curr[97]  = { '\0' };
    char group_curr[41] = { '\0' };
    
    /*
    fprintf(stderr, "RECEIVED DEL_META FOR:\n");
    fprintf(stderr, "  user_id:   %s\n", msg_meta.user_id);
    fprintf(stderr, "  group_id:  %s\n", msg_meta.group_id);
    */

    _meta_user_z *curr_user   = NULL;
    _meta_user_z *prev_user   = NULL;
    _info_group_z *curr_group = NULL;
    _info_group_z *prev_group = NULL;
    _meta_local_z *curr_file  = NULL;
    _meta_local_z *prev_file  = NULL;

    pthread_mutex_lock(&mutex_table);

    if (meta_user != NULL)
    {
      curr_user = meta_user;
      prev_user = meta_user;

      while (curr_user != NULL)
      {
	sprintf(user_curr, "%s", curr_user->user_id);
	user_curr[40] = '\0';
	
	if (strcmp(user_curr, msg_meta.user_id) == 0) // user already exists
	{
	  if (curr_user->groups != NULL)
	  {
	    curr_group = curr_user->groups;
	    prev_group = curr_user->groups;

	    while (curr_group != NULL)
	    {
	      sprintf(group_curr, "%s", curr_group->group_id);
	      group_curr[40] = '\0';
	      
	      if (strcmp(group_curr, msg_meta.group_id) == 0)
	      {
		// once group is found, free all
		// files and then free the group

		curr_file = curr_group->files;
		prev_file = curr_group->files;

		// remove each file one by one
		while (curr_file != NULL)
		{
		  prev_file = curr_file;
		  curr_file = curr_file->next;

		  if (prev_file != NULL)
		    free(prev_file);
		}

		if (curr_group->next != NULL)
		{
		  prev_group->next = curr_group->next;
		  free(curr_group);
		}
		else
		  prev_group->next = NULL;

		if (strcmp(group_curr, curr_user->groups->group_id) == 0)
		{
		  free(curr_user->groups);
		  curr_user->groups = NULL;
		  
		  if (curr_user != meta_user)
		  {
		    if (curr_user->next != NULL)
		    {
		      prev_user->next = curr_user->next;
		      free(curr_user);
		    }
		    else
		      prev_user->next = NULL;
		  }
		  else
		  {
		    if (curr_user->next == NULL)
		    {
		      free(curr_user);
		      meta_user = NULL;
		    }
		    else
		      meta_user = curr_user->next;
		  }
		}

		break;
	      }

	      if (curr_group != curr_user->groups)
		prev_group = prev_group->next;
	      
	      if (curr_group->next != NULL)
		curr_group = curr_group->next;
	      else
		break;
	    }
	    
	    break;
	  }
	  
	  break;
	}

	if (curr_user != meta_user)
	  prev_user = prev_user->next;
	
	curr_user = curr_user->next;
      }
    }

    // replicate deletion request out to NUM_METAS * 2
    // neighbors to "ensure" that this meta data is
    // removed from the Quartz network
    ChimeraHost **neighbors = (ChimeraHost **) route_neighbors(NUM_METAS * 2);

    char *s_sendbuf = (char *)malloc(sizeof(_msg_meta_z));
    memset(s_sendbuf, '\0', sizeof(_msg_meta_z));
    memcpy(s_sendbuf, &msg_meta, sizeof(_msg_meta_z));
    
    for (i = 0; i < NUM_METAS * 2; i++)
    {
      if (neighbors[i] == NULL)
	break;

      if ((neighbors[i])->key != deviceInfo->nodeKey &&
	  (neighbors[i])->key != 0)
      {
	chimera_send((neighbors[i])->key, CTRL_DEL_META_GROUP_COPY,
		     sizeof(_msg_meta_z), s_sendbuf);
      }
    }

    if (neighbors != NULL)
      free(neighbors);

    pthread_mutex_unlock(&mutex_table);

    fflush(stdout);
    //printf("FINISHED a CTRL_DEL_META_GROUP message\n");
  }
  else if (m->type == CTRL_DEL_META_GROUP_COPY) { 

    // ***********************************************************************
    // NODE REQUESTS TO DELETE META DATA (COPY)
    // ***********************************************************************

    //printf("Received a CTRL_DEL_META_GROUP_COPY message\n");

    _msg_meta_z msg_meta;
    memset(&msg_meta, '\0', sizeof(_msg_meta_z));
    memcpy(&msg_meta, m->payload, sizeof(_msg_meta_z));

    int i = 0, exists = 0;
    Key user_key;

    char user_curr[41]  = { '\0' };
    char file_curr[96]  = { '\0' };
    char group_curr[41] = { '\0' };
    
    /*
    fprintf(stderr, "RECEIVED DEL_META_COPY FOR:\n");
    fprintf(stderr, "  user_id:   %s\n", msg_meta.user_id);
    fprintf(stderr, "  file_name: %s\n", msg_meta.file_name);
    fprintf(stderr, "  group_id:  %s\n", msg_meta.group_id);
    */

    _meta_user_z  *curr_user  = NULL;
    _meta_user_z  *prev_user  = NULL;
    _info_group_z *curr_group = NULL;
    _info_group_z *prev_group = NULL;
    _meta_local_z *curr_file  = NULL;
    _meta_local_z *prev_file  = NULL;

    pthread_mutex_lock(&mutex_table);

    if (meta_user != NULL)
    {
      curr_user = meta_user;
      prev_user = meta_user;

      while (curr_user != NULL)
      {
	sprintf(user_curr, "%s", curr_user->user_id);
	user_curr[40] = '\0';
	
	if (strcmp(user_curr, msg_meta.user_id) == 0) // user already exists
	{
	  if (curr_user->groups != NULL)
	  {
	    curr_group = curr_user->groups;
	    prev_group = curr_user->groups;

	    while (curr_group != NULL)
	    {
	      sprintf(group_curr, "%s", curr_group->group_id);
	      group_curr[40] = '\0';
	      
	      if (strcmp(group_curr, msg_meta.group_id) == 0)
	      {
		// once group is found, free all
		// files and then free the group

		curr_file = curr_group->files;
		prev_file = curr_group->files;

		// remove each file one by one
		while (curr_file != NULL)
		{
		  prev_file = curr_file;
		  curr_file = curr_file->next;

		  if (prev_file != NULL)
		    free(prev_file);
		}

		if (curr_group->next != NULL)
		{
		  prev_group->next = curr_group->next;
		  free(curr_group);
		}
		else
		  prev_group->next = NULL;

		if (strcmp(group_curr, curr_user->groups->group_id) == 0)
		{
		  free(curr_user->groups);
		  curr_user->groups = NULL;
		  
		  if (curr_user != meta_user)
		  {
		    if (curr_user->next != NULL)
		    {
		      prev_user->next = curr_user->next;
		      free(curr_user);
		    }
		    else
		      prev_user->next = NULL;
		  }
		  else
		  {
		    if (curr_user->next == NULL)
		    {
		      free(curr_user);
		      meta_user = NULL;
		    }
		    else
		      meta_user = curr_user->next;
		  }
		}

		break;
	      }

	      if (curr_group != curr_user->groups)
		prev_group = prev_group->next;
	      
	      if (curr_group->next != NULL)
		curr_group = curr_group->next;
	      else
		break;
	    }
	    
	    break;
	  }
	  
	  break;
	}

	if (curr_user != meta_user)
	  prev_user = prev_user->next;
	
	curr_user = curr_user->next;
      }
    }

    pthread_mutex_unlock(&mutex_table);

    //printf("FINISHED a CTRL_DEL_META_GROUP_COPY message\n");
  }


  else if (m->type == CTRL_META_COPY) { 

    // ***********************************************************************
    // NODE REQUESTS TO STORE COPIED META DATA
    // ***********************************************************************

    //pthread_mutex_lock(&mutex_metac);

#ifdef DEBUG	
    //printf("Received a CTRL_META_COPY message\n");
#endif
    _msg_meta_z msg_meta;
    memset(&msg_meta, '\0', sizeof(_msg_meta_z));
    memcpy(&msg_meta, m->payload, sizeof(_msg_meta_z));
    
#ifdef DEBUG
    //	dumpBytes(stderr, m->payload, m->size , " [stor] ");
#endif
	
    _hb_table_z *temp_hb = my_heartbeat;
    int i = 0, exists = 0;
    Key user_key;

    char user_curr[41];
    char file_curr[97];
    char group_curr[41];

    /*    
    fprintf(stderr, "RECEIVED META_STORE FOR:\n");
    fprintf(stderr, "  user_id:   %s\n", msg_meta.user_id);
    fprintf(stderr, "  file_name: %s\n", msg_meta.file_name);
    fprintf(stderr, "  group_id:  %s\n", msg_meta.group_id);
    fprintf(stderr, "  size:      %d\n", msg_meta.size);
    fprintf(stderr, "  version:   %d\n", msg_meta.version);
    */

    _meta_user_z *curr_user   = NULL;
    _meta_user_z *prev_user   = NULL;
    _info_group_z *curr_group = NULL;
    _info_group_z *prev_group = NULL;
    _meta_local_z *curr_file  = NULL;
    _meta_local_z *prev_file  = NULL;

    int exists_user  = 0;
    int exists_file  = 0;
    int exists_group = 0;

    exists_user = 0;
 
    pthread_mutex_lock(&mutex_table);

    if (meta_user != NULL)
    {
      curr_user = meta_user;
      while (curr_user != NULL)
      {
	sprintf(user_curr, "%s", curr_user->user_id);
	user_curr[40] = '\0';
	
	if (strcmp(user_curr, msg_meta.user_id) == 0) // user already exists
	{
	  exists_user = 1;
	  
	  if (curr_user->groups != NULL)
	  {
	    curr_group = curr_user->groups;
	    exists_group = 0;
	    while (curr_group != NULL)
	    {
	      sprintf(group_curr, "%s", curr_group->group_id);
	      group_curr[40] = '\0';
	      
	      if (strcmp(group_curr, msg_meta.group_id) == 0)
	      {
		exists_group = 1;
		
		curr_file = curr_group->files;
		exists_file = 0;
		while (curr_file != NULL)
		{
		  memset(file_curr, '\0', 95);
		  sprintf(file_curr, "%s", curr_file->file_name);
		  file_curr[95] = '\0';
		  
		  if (strcmp(file_curr, msg_meta.file_name) == 0)
		  {
		    exists_file = 1;
		    
		    // the version must be greater than the already stored
		    if (msg_meta.version > curr_file->version)
		    {
		      curr_file->version = msg_meta.version;
		      curr_file->size    = msg_meta.size;
		    }
		  
		    // if not, then that means user needs to update first
		    // before user can officially commit changes
		    else
		    {
		      // do not update meta data
		    }
		    
		    break;
		  }
		  
		  curr_file = curr_file->next;
		}
		
		// if file doesn't exist, need to add it to curr_group
		if (!exists_file)
		{
		  if (curr_group->files != NULL)
		  {
		    curr_file = curr_group->files;
		    prev_file = curr_group->files;
		    while (curr_file != NULL)
		    {
		      prev_file = curr_file;
		      curr_file = curr_file->next;
		    }
		    
		    _meta_local_z *new_file  = NULL;
		    
		    new_file = (_meta_local_z *) malloc(sizeof(_meta_local_z));
		    new_file->next = NULL;
		    new_file->type      = 'a';
		    
		    memset(new_file->file_name, '\0', 95);
		    sprintf(new_file->file_name, "%s", msg_meta.file_name);
		    new_file->file_name[95]= '\0';

		    new_file->version = msg_meta.version;
		    new_file->size    = msg_meta.size;

		    prev_file->next = new_file;

		  }
		}
		
		break;
	      }
	      
	      if (curr_group->next != NULL)
		curr_group = curr_group->next;
	      else
		break;
	    }
	  }
	  
	  if (!exists_group)
	  {
	    //pthread_mutex_lock(&mutex_table);
	    
	    if (curr_user->groups != NULL)
	    {
	      curr_group = curr_user->groups;
	      prev_group = curr_user->groups;
	      
	      while (curr_group != NULL)
	      {
		prev_group = curr_group;
		curr_group = curr_group->next;
	      }
	      
	      curr_group = (_info_group_z *) malloc(sizeof(_info_group_z));
	      curr_group->next  = NULL;
	      curr_group->files = NULL;

	      prev_group->next = curr_group;
	    }
	    else
	    {
	      curr_user->groups = (_info_group_z *) malloc(sizeof(_info_group_z));
	      curr_group = curr_user->groups;
	      curr_group->next  = NULL;
	      curr_group->files = NULL;
	    }

	    //fprintf(stderr, "creating group %s\n", msg_meta.group_id);

	    memset(curr_group->group_id, '\0', 40);
	    sprintf(curr_group->group_id, "%s", msg_meta.group_id);
	    curr_group->group_id[40] = '\0';
	    
	    curr_group->files = (_meta_local_z *) malloc(sizeof(_meta_local_z));
	    curr_group->files->next = NULL;
	    curr_group->files->type = 'a';
	  
	    memset(curr_group->files->file_name, '\0', 95);
	    sprintf(curr_group->files->file_name, "%s", msg_meta.file_name);
	    curr_group->files->file_name[95] = '\0';

	    curr_group->files->version = msg_meta.version;
	    curr_group->files->size    = msg_meta.size;
	  }
	  
	  break;
	}
	
	curr_user = curr_user->next;
      }
    }

    if (!exists_user)
    {
      if (meta_user != NULL)
      {
	curr_user = meta_user;
	prev_user = meta_user;

	while (curr_user != NULL)
	{
	  prev_user = curr_user;
	  curr_user = curr_user->next;
	}

	curr_user = (_meta_user_z *) malloc(sizeof(_meta_user_z));
	curr_user->groups = NULL;
	curr_user->next   = NULL;

	prev_user->next = curr_user;
      }
      else
      {
	meta_user = (_meta_user_z *) malloc(sizeof(_meta_user_z));
	meta_user->groups = NULL;
	meta_user->next   = NULL;

	curr_user = meta_user;
      }

      memset(curr_user->user_id, '\0', 40);
      sprintf(curr_user->user_id, "%s", msg_meta.user_id);
      curr_user->user_id[40] = '\0';

      memcpy( &(curr_user->send_key), &(msg_meta.send_key), sizeof(Key) );

      curr_user->pos = 5;

      if (curr_user->groups == NULL)
      {
	curr_user->groups = (_info_group_z *) malloc(sizeof(_info_group_z));
	curr_user->groups->files = NULL;
	curr_user->groups->next  = NULL;
      }

      curr_group = curr_user->groups;
      //curr_group = (_info_group_z *) malloc(sizeof(_info_group_z));

      memset(curr_group->group_id, '\0', 40);
      sprintf(curr_group->group_id, "%s", msg_meta.group_id);
      curr_group->group_id[40] = '\0';

      curr_group->files = (_meta_local_z *) malloc(sizeof(_meta_local_z));
      curr_group->files->next = NULL;
      curr_group->files->type = 'a';

      memset(curr_group->files->file_name, '\0', 95);
      sprintf(curr_group->files->file_name, "%s", msg_meta.file_name);
      curr_group->files->file_name[95] = '\0';

      curr_group->files->version = msg_meta.version;
      curr_group->files->size    = msg_meta.size;

      /*
      fprintf(stderr, "  added NEW user entry!\n");
      fprintf(stderr, "  ----------------------------------------\n");
      fprintf(stderr, "   user:  %s\n", curr_user->user_id);
      fprintf(stderr, "   pos:   %d\n", curr_user->pos);
      fprintf(stderr, "   group: %s\n", curr_group->group_id);
      fprintf(stderr, "     file:  %s\n", curr_group->files->file_name);
      fprintf(stderr, "       version: %d\n", curr_group->files->version);
      fprintf(stderr, "       size:    %d\n", curr_group->files->size);
      */
    }

    pthread_mutex_unlock(&mutex_table);

    //printf("FINISHED a CTRL_META_COPY message\n");
  }
  else if (m->type == CTRL_META_FILE_REQ) { 

    // ***********************************************************************
    // NODE REQUESTS TO GET META FILE DATA
    // ***********************************************************************

#ifdef DEBUG	
    // printf("Received a CTRL_META_FILE_REQ\n");
#endif


    pthread_mutex_lock(&mutex_table);

    _msg_meta_z msg_meta;
    memset(&msg_meta, '\0', sizeof(_msg_meta_z));
    memcpy(&msg_meta, m->payload, sizeof(_msg_meta_z));

    // decode received message
    //_msg_meta_z *msgMeta = msg_meta_decode(m->payload, m->size);
    
#ifdef DEBUG
    //	dumpBytes(stderr, m->payload, m->size , " [stor] ");
#endif

    log_exec(LOG_RECV, msg_meta.node_id, LOG_CMETA, LOG_GET);

    _hb_table_z *temp_hb = my_heartbeat;
    int i = 0, exists = 0;
    Key user_key;

    char user_curr[41]  = { '\0' };
    char file_curr[96]  = { '\0' };
    char group_curr[41] = { '\0' };

    _meta_user_z *curr_user   = NULL;
    _info_group_z *curr_group = NULL;
    _meta_local_z *curr_file  = NULL;

    int exists_user  = 0;
    int exists_file  = 0;
    int exists_group = 0;

    if (meta_user != NULL)
    {
      curr_user = meta_user;
      while (curr_user != NULL)
      {
	sprintf(user_curr, "%s", curr_user->user_id);
	user_curr[40] = '\0';
	
	if (strcmp(user_curr, msg_meta.user_id) == 0) // user already exists
	{
	  exists_user = 1;
	  
	  if (curr_user->groups != NULL)
	  {
	    curr_group = curr_user->groups;
	    exists_group = 0;
	    while (curr_group != NULL)
	    {
	      sprintf(group_curr, "%s", curr_group->group_id);
	      group_curr[40] = '\0';
	      
	      if (strcmp(group_curr, msg_meta.group_id) == 0)
	      {
		exists_group = 1;
		
		curr_file = curr_group->files;
		exists_file = 0;
		while (curr_file != NULL)
		{
		  memset(file_curr, '\0', 95);
		  sprintf(file_curr, "%s", curr_file->file_name);
		  file_curr[95] = '\0';
		  
		  if (strcmp(file_curr, msg_meta.file_name) == 0)
		  {
		    exists_file = 1;
		    
		    // store curr_file information back into msg_meta
		    msg_meta.version = curr_file->version;
		    msg_meta.size    = msg_meta.size;

		    // everything else (send_key, file, group)
		    // should be the same that's currently in

		    // send back to requestor (send_key)
		    char *s_sendbuf = (char *)malloc(sizeof(_msg_meta_z));
		    memset(s_sendbuf, '\0', sizeof(_msg_meta_z));
		    memcpy(s_sendbuf, &msg_meta, sizeof(_msg_meta_z));
		    
		    chimera_send(msg_meta.node_id, CTRL_META_REQ_YES, sizeof(_msg_meta_z), s_sendbuf);

		    free(s_sendbuf);
		    
		    break;
		  }
		  
		  curr_file = curr_file->next;
		}
		
		// if file doesn't exist, need to add it to curr_group
		if (!exists_file)
		{
		  //fprintf(stderr, "file does not exist\n");
		}
		
		break;
	      }
	      
	      if (curr_group->next != NULL)
		curr_group = curr_group->next;
	      else
		break;
	    }
	    
	    
	    break;
	  }
	  
	  if (!exists_group)
	  {
	    //fprintf(stderr, "group does not exist\n");
	  }
	  
	  break;
	}
	
	curr_user = curr_user->next;
      }
    }

    if (!exists_file)
    {
      // this entry does not exist -- must store it to Quartz
      char *s_sendbuf = (char *)malloc(sizeof(_msg_meta_z));
      memset(s_sendbuf, '\0', sizeof(_msg_meta_z));
      memcpy(s_sendbuf, &msg_meta, sizeof(_msg_meta_z));

      chimera_send(msg_meta.node_id, CTRL_META_REQ_NO, sizeof(_msg_meta_z), s_sendbuf);
      free(s_sendbuf);
    }

    pthread_mutex_unlock(&mutex_table);

    // printf("FINISHED a CTRL_META_FILE_REQ\n");
  }
  else if (m->type == CTRL_META_GROUP_REQ) { 

    // ***********************************************************************
    // NODE REQUESTS TO GET META GROUP DATA
    // ***********************************************************************

#ifdef DEBUG	
    // printf("Received a CTRL_META_GROUP_REQ\n");
#endif

    pthread_mutex_lock(&mutex_table);

    _msg_meta_z msg_meta;
    memset(&msg_meta, '\0', sizeof(_msg_meta_z));
    memcpy(&msg_meta, m->payload, sizeof(_msg_meta_z));

    // decode received message
    //_msg_meta_z *msgMeta = msg_meta_decode(m->payload, m->size);
    
#ifdef DEBUG
    //	dumpBytes(stderr, m->payload, m->size , " [stor] ");
#endif

    log_exec(LOG_RECV, msg_meta.node_id, LOG_GMETA, msg_meta.user_id);

    int i = 0, exists = 0;
    Key user_key;

    char user_curr[41]  = { '\0' };
    char user_copy[41]  = { '\0' };
    char group_curr[41] = { '\0' };
    
    _meta_user_z *curr_user   = NULL;
    _info_group_z *curr_group = NULL;
    _meta_local_z *curr_file  = NULL;

    int exists_user  = 0;
    int exists_group = 0;
    int total_files  = 0;

    memset(user_copy, '\0', 40);
    sprintf(user_copy, "%s", msg_meta.user_id);

    exists_user = 0;

    // simply count and send the total number of files in this group
    if (meta_user != NULL)
    {
      curr_user = meta_user;
      while (curr_user != NULL)
      {
	sprintf(user_curr, "%s", curr_user->user_id);
	user_curr[40] = '\0';

	if (strcmp(user_curr, msg_meta.user_id) == 0) // user exists
	{
	  if (curr_user->groups != NULL)
	  {
	    curr_group = curr_user->groups;
	    while (curr_group != NULL)
	    {
	      sprintf(group_curr, "%s", curr_group->group_id);
	      group_curr[40] = '\0';
	      
	      if (strcmp(group_curr, msg_meta.group_id) == 0)
	      {
		curr_file = curr_group->files;

		while (curr_file != NULL)
		{
		  if (curr_file->size > 0)
		    total_files++;

		  curr_file = curr_file->next;
		}
		
		break;
	      }
	      
	      if (curr_group->next != NULL)
		curr_group = curr_group->next;
	      else
		break;
	    }
	    
	    break;
	  }
	  
	  break;
	}
	
	curr_user = curr_user->next;
      }
    }

    if (total_files)
    {
      char *s_sendbuf = (char *)malloc(sizeof(_msg_meta_z));
      
      memset(msg_meta.user_id, '\0', 40);
      sprintf(msg_meta.user_id, "META_GROUP_REPLY");
      msg_meta.user_id[40] = '\0';
      msg_meta.size = total_files;
      memset(s_sendbuf, '\0', sizeof(_msg_meta_z));
      memcpy(s_sendbuf, &msg_meta, sizeof(_msg_meta_z));
      chimera_send(msg_meta.node_id, CTRL_META_REQ_YES, sizeof(_msg_meta_z), s_sendbuf);

      free(s_sendbuf);

      memset(msg_meta.user_id, '\0', 40);
      sprintf(msg_meta.user_id, "%s", user_copy);
    }
    
    if (meta_user != NULL)
    {
      curr_user = meta_user;
      while (curr_user != NULL)
      {
	sprintf(user_curr, "%s", curr_user->user_id);
	user_curr[40] = '\0';

	// fprintf(stderr, "U %s == %s\n", user_curr, msg_meta.user_id);

	if (strcmp(user_curr, msg_meta.user_id) == 0) // user exists
	{
	  exists_user = 1;
	  
	  if (curr_user->groups != NULL)
	  {
	    curr_group = curr_user->groups;
	    exists_group = 0;
	    while (curr_group != NULL)
	    {
	      sprintf(group_curr, "%s", curr_group->group_id);
	      group_curr[40] = '\0';
	      
	      // fprintf(stderr, "G %s == %s\n", group_curr, msg_meta.group_id);
	      if (strcmp(group_curr, msg_meta.group_id) == 0)
	      {
		exists_group = 1;
		
		curr_file = curr_group->files;

		char *s_sendbuf = (char *)malloc(sizeof(_msg_meta_z));

		while (curr_file != NULL)
		{
		  // everything else (send_key, file, group)
		  // should be the same that's currently in

		  if (curr_file->size > 0)
		  {
		    msg_meta.user_id[40] = '\0';
		    memset(msg_meta.file_name, '\0', 95);
		    sprintf(msg_meta.file_name, curr_file->file_name);
		    //strcpy(msg_meta.file_name, curr_file->file_name);
		    msg_meta.file_name[95] = '\0';
		    msg_meta.version = curr_file->version;
		    msg_meta.size    = curr_file->size;
		    
		    // send back to requestor (send_key)
		    memset(s_sendbuf, '\0', sizeof(_msg_meta_z));
		    memcpy(s_sendbuf, &msg_meta, sizeof(_msg_meta_z));
		    
		    // fprintf(stderr, "Sending to [%x] YES.\n", msg_meta.node_id);
		    
		    chimera_send(msg_meta.node_id, CTRL_META_REQ_YES, sizeof(_msg_meta_z), s_sendbuf);
		  }
		    
		  curr_file = curr_file->next;
		}

		free(s_sendbuf);
		
		break;
	      }
	      
	      if (curr_group->next != NULL)
		curr_group = curr_group->next;
	      else
		break;
	    }
	    
	    break;
	  }
	  
	  if (!exists_group)
	  {
	    //fprintf(stderr, "group does not exist\n");
	  }
	  
	  break;
	}
	
	curr_user = curr_user->next;
      }
    }

    //pthread_mutex_unlock(&mutex_meta_r);

    if (!exists_user || !exists_group)
    {
      // user or group does not exist
      // user needs to re-sync the group of files
      msg_meta.size = NOT_FOUND;

      char *s_sendbuf = (char *)malloc(sizeof(_msg_meta_z));
      memset(s_sendbuf, '\0', sizeof(_msg_meta_z));
      memcpy(s_sendbuf, &msg_meta, sizeof(_msg_meta_z));

      chimera_send(msg_meta.node_id, CTRL_META_REQ_NO, sizeof(_msg_meta_z), s_sendbuf);
      
      free(s_sendbuf);
    }

    pthread_mutex_unlock(&mutex_table);

    //printf("FINISHED a CTRL_META_GROUP_REQ\n");
  }
  else if (m->type == CTRL_META_REQ_YES) { 

    // ***********************************************************************
    // NODE REQUESTS TO GET META DATA SUCCEEDS
    // ***********************************************************************

#ifdef DEBUG	
    // printf("Received a CTRL_META_REQ_YES message\n");
#endif

    //pthread_mutex_unlock(&mutex_cmdline);
    //pthread_mutex_lock(&mutex_cmdline);
    pthread_mutex_lock(&mutex_meta_r);
    
    _msg_meta_z msg_meta;
    memset(&msg_meta, '\0', sizeof(_msg_meta_z));
    memcpy(&msg_meta, m->payload, sizeof(_msg_meta_z));

    log_exec(LOG_RECV, msg_meta.node_id, LOG_META_YES, msg_meta.user_id);

    if (strcmp(msg_meta.user_id, "META_GROUP_REPLY") == 0 &&
	msg_meta.size > 0 && msg_meta.size < 1000)
    {
      total_receiving = msg_meta.size;
      total_received  = 0;
    }
    else
    {
      int file_exists = 0;
      int len = 0;
      
      char group_id[41];
      
      HASH fid  = { '\0' };
      HASH gid  = { '\0' };
      HASH g_id = { '\0' };
      
      sha1_string(msg_meta.file_name, &fid);
      sha1_stringToHash( msg_meta.group_id, &gid );
      
      INMEM_GROUP* imGroup = NULL;
      INMEM_FILE*  imFile  = NULL;
      
      if( NULL != deviceInfo )
      {
	INMEM_GROUP* scout = (INMEM_GROUP*)deviceInfo->inMemGroups;
	
	while( NULL != scout )
	{
	  if( NULL != scout->name )
	  {
	    sha1_string(scout->name, &g_id);
	    memcpy(group_id, sha1_hashToString(&g_id), 41);
	    
	    if (0 == strcmp( group_id, msg_meta.group_id ))
	    {
	      imGroup = scout;
	      break;
	    }
	  }
	  
	  scout = scout->next;
	}
      }
      
      if (imGroup != NULL) // group found
      {
	if (imGroup->fileList != NULL)
	{
	  INMEM_FILE* imFile = imGroup->fileList;
	  
	  while (imFile != NULL)
	  {
	    if (strcmp(imFile->name, msg_meta.file_name) == 0)
	    {
	      file_exists = 1;
	      
	      // if the version received is older than my version
	      // meaning my version is newer, COMMIT it to Quartz
	      if (msg_meta.version < imFile->version)
	      {
		printf("\n Your version (%d) is newer than the current version (%d)\n",
		       imFile->version, msg_meta.version); fflush(stdout);
		printf(" for file: %s, group: %s\n\n", imFile->name, imGroup->name);
		fflush(stdout);
		
		char* input = readline(" Save this file to Quartz (y/n)?  "); fflush(stdout);
		
		while (*input == '\0')
		{
		  input = readline(" Save this file to Quartz (y/n)?  "); fflush(stdout);
		}

		if (strcmp(input, "y")   == 0 || strcmp(input, "Y")   == 0 ||
		    strcmp(input, "yes") == 0 || strcmp(input, "YES") == 0)
		{
		  sync_commitFile(imGroup->name, imFile, 0);
		  //pthread_mutex_unlock(&mutex_cmdline);
		}
		else if (strcmp(input, "n")  == 0 || strcmp(input, "N")  == 0 || 
			 strcmp(input, "no") == 0 || strcmp(input, "NO") == 0)
		{
		  //printf("\n");
		  input = readline(" Retrieve an update from Quartz instead (y/n)?  "); fflush(stdout);
		  while (*input == '\0')
		  {
		    input = readline(" Retrieve an update from Quartz instead (y/n)?  "); fflush(stdout);
		  }

		  if (strcmp(input, "y")   == 0 || strcmp(input, "Y")   == 0 ||
		      strcmp(input, "yes") == 0 || strcmp(input, "YES") == 0)
		  {
		    printf("\n Warning: This will overwrite your current version!\n"); fflush(stdout);
		    input = readline(" Would you still like to continue (y/n)?  "); fflush(stdout);
		    
		    if (strcmp(input, "y")   == 0 || strcmp(input, "Y")   == 0 ||
			strcmp(input, "yes") == 0 || strcmp(input, "YES") == 0)
		    {
		      //printf("\n Updating this file from Quartz.\n"); fflush(stdout);
		      quartz_store_get_entry(deviceInfo->nodeKey, deviceInfo->userHash,
					     msg_meta.group_key, gid, fid, msg_meta.version);
		    }
		  }
		  else
		  {
		    total_received++;

		    // processed all the expected number of
		    // files for that last group request
		    if (total_received == total_receiving)
		      pthread_mutex_unlock(&mutex_cmdline);
		  }
		}
	      }
	      
	      // if the version received is newer than my version
	      // meaning my version is older, UPDATE the file
	      else if (msg_meta.version > imFile->version)
	      {
		printf("\n Your version (%d) is older than the current version (%d)\n",
		       imFile->version, msg_meta.version); fflush(stdout);
		printf(" for file: %s, group: %s\n\n", imFile->name, imGroup->name); fflush(stdout);
		
		char* input = readline(" Retrieve an update from Quartz (y/n)?  "); fflush(stdout);
		while (*input == '\0')
	        {
		  input = readline(" Retrieve an update from Quartz (y/n)?  "); fflush(stdout);
		}
		
		if (strcmp(input, "y")   == 0 || strcmp(input, "Y")   == 0 ||
		    strcmp(input, "yes") == 0 || strcmp(input, "YES") == 0)
		{
		  //printf("\n Updating this file from Quartz.\n"); fflush(stdout);
		  quartz_store_get_entry(deviceInfo->nodeKey, deviceInfo->userHash,
					 msg_meta.group_key, gid, fid, msg_meta.version);
		}
		else if (strcmp(input, "n")  == 0 || strcmp(input, "N")  == 0 || 
			 strcmp(input, "no") == 0 || strcmp(input, "NO") == 0)
		{
		  input = readline(" Save this version to Quartz instead (y/n)?  "); fflush(stdout);
		  
		  while (*input == '\0')
		  {
		    input = readline(" Save this version to Quartz instead (y/n)?  "); fflush(stdout);
		  }
		  
		  if (strcmp(input, "y")   == 0 || strcmp(input, "Y")   == 0 ||
		      strcmp(input, "yes") == 0 || strcmp(input, "YES") == 0)
		  {
		    sync_commitFile(imGroup->name, imFile, 0);
		    // pthread_mutex_unlock(&mutex_cmdline);
		  }
		  else
		  {
		    total_received++;

		    // processed all the expected number of
		    // files for that last group request
		    if (total_received == total_receiving)
		      pthread_mutex_unlock(&mutex_cmdline);

		    // pthread_mutex_unlock(&mutex_cmdline);
		  }
		}
		else
		{
		  // unknown option -- ignore
		  // pthread_mutex_unlock(&mutex_cmdline);
		}
	      }
	      // if the versions are the same, do nothing but ask first if user still wants to commit
	      else
	      {
		char *buff = NULL;
		
		printf("\n You already have the current version (%d) for\n", msg_meta.version);
		printf(" file: %s, group: %s\n", imFile->name, imGroup->name); fflush(stdout);
		
		// if you have the newest version and lastChanged has indeed
		// been modified since the last update, commit this file
		sync_commitFile(imGroup->name, imFile, 0);
		// pthread_mutex_unlock(&mutex_cmdline);
	      }
	      
	      break;
	    }

	    imFile = imFile->next;
	  }
	}
	
	if (!file_exists)
	{
	  fflush(stdout);
	  printf("\n New file detected from Quartz!\n"); fflush(stdout);
	  printf(" Adding: file \"%s\" to group \"%s\"\n",  msg_meta.file_name, imGroup->name);
	  fflush(stdout);
	  
	  group_addFile(msg_meta.file_name, imGroup->name);
	  quartz_store_get_entry(deviceInfo->nodeKey, deviceInfo->userHash,
				 msg_meta.group_key, gid, fid, msg_meta.version);
	}
      }
      else
      {
	printf(" ** Error ** Group \"%s\" does not exist!\n", group_id); fflush(stdout);
	pthread_mutex_unlock(&mutex_cmdline);
      }
      
    }

    pthread_mutex_unlock(&mutex_meta_r);
  }

  else if (m->type == CTRL_META_REQ_NO) { 

    // ***********************************************************************
    // NODE REQUESTS TO GET META DATA FAILS
    // ***********************************************************************

#ifdef DEBUG	
    // printf("Received a CTRL_META_REQ_NO message\n");
#endif

    /*
    pthread_mutex_unlock(&mutex_cmdline);
    pthread_mutex_lock(&mutex_cmdline);
    */

    pthread_mutex_lock(&mutex_meta_r);

    _msg_meta_z msg_meta;
    memset(&msg_meta, '\0', sizeof(_msg_meta_z));
    memcpy(&msg_meta, m->payload, sizeof(_msg_meta_z));

    /*	
    fprintf(stderr, "RECEIVED META REQ NO FOR:\n");
    fprintf(stderr, "  user_id:   %s\n", msg_meta.user_id);
    fprintf(stderr, "  file_name: %s\n", msg_meta.file_name);
    fprintf(stderr, "  group_id:  %s\n", msg_meta.group_id);
    fprintf(stderr, "  version:   %d\n", msg_meta.version);
    fprintf(stderr, "  size:      %d\n", msg_meta.size);
    */
    
    log_exec(LOG_RECV, msg_meta.node_id, LOG_META_YES, msg_meta.user_id);

    // because it failed, we must commit the group

    char group_id[41];

    HASH fid  = { '\0' };
    HASH gid  = { '\0' };
    HASH g_id = { '\0' };

    sha1_string(msg_meta.file_name, &fid);
    sha1_stringToHash( msg_meta.group_id, &gid );

    INMEM_GROUP* imGroup = NULL;
    INMEM_FILE*  imFile  = NULL;

    if( NULL != deviceInfo )
    {
      INMEM_GROUP* scout = (INMEM_GROUP*)deviceInfo->inMemGroups;

      while( NULL != scout )
      {
	if( NULL != scout->name )
	{
	  sha1_string(scout->name, &g_id);
	  memcpy(group_id, sha1_hashToString(&g_id), 41);
	  
	  if (0 == strcmp( group_id, msg_meta.group_id ))
	  {
	    imGroup = scout;
	    break;
	  }
	}

	scout = scout->next;
      }
    }

    if (imGroup != NULL) // group found
    {
      // if file_name is empty, then the entire group is new
      // and must be created and committed within Quartz
      if (strcmp(msg_meta.file_name, "") == 0)
	sync_commitGroup(imGroup->name, FORCE_COMMIT);
      else
      {
	INMEM_FILE* file = (INMEM_FILE*)imGroup->fileList;

	while(NULL != file)
	{
	  if (strcmp(msg_meta.file_name, file->name) == 0)
	    break;
	  
	  file = file->next;
	}
	
	if (file != NULL)
	  sync_commitFile(imGroup->name, file, FORCE_COMMIT);
      }
    }
    else
      printf(" ** Error ** Group not found\n");

    pthread_mutex_unlock(&mutex_meta_r);
    pthread_mutex_unlock(&mutex_cmdline);
  }

  else if (m->type == CTRL_UPDATE) { 

    pthread_mutex_lock(&mutex_update);
    //fprintf(stderr, "received a CTRL_UPDATE\n");

    char* s_sendbuf  = NULL;
    char* s_read_buf = NULL;
    _msg_update msg_update;
    memset(&msg_update, '\0', sizeof(_msg_update));
    memcpy(&msg_update, m->payload, sizeof(_msg_update));
    
    // first maintain stored nodes since these are the most important
    FILE *f_meta;
    int inode_num = 1;
    int nread = 0;
    int numbytes = 0;
    int j = 0;
    
    pthread_mutex_lock(&mutex_inode);

    f_meta = fopen("meta.sto", "r");

    if (f_meta != NULL)
    {
      int num_blocks, i, total;
      Key send_key;
      char temp[41] = { '\0' };
      block *temp_block = NULL;

      // inode_list *thelist = (inode_list *) malloc(sizeof(inode_list));
      inode_list *thelist = NULL;

      inode_list *curr_list = thelist;
      inode temp_inode;
    
      while (!feof(f_meta))
      {
	memset(&temp_inode, '\0', sizeof(inode));
	nread = fread(&temp_inode, sizeof(inode), 1, f_meta);

	if (temp_inode.version != QFS_NULL &&
	    temp_inode.size    != QFS_NULL && temp_inode.key != 0)
	{
	  if (strcmp(temp_inode.user_id, "") != 0)
	  {
	    if (thelist == NULL)
	    {
	      thelist = (inode_list *) malloc(sizeof(inode_list));
	      memset(&(thelist->the_inode), '\0', sizeof(inode));
	      memcpy(&(thelist->the_inode), &temp_inode, sizeof(inode));
	      thelist->next = NULL;
	      
	      curr_list = thelist;
	    }
	    else
	    {
	      curr_list->next = (inode_list *) malloc(sizeof(inode_list));
	      curr_list = curr_list->next;
	      
	      memset(&(curr_list->the_inode), '\0', sizeof(inode));
	      memcpy(&(curr_list->the_inode), &temp_inode, sizeof(inode));
	      curr_list->next = NULL;
	    }
	  }
	}
      }

      fclose(f_meta);

      inode_list *temp_list = thelist;

      while (temp_list != NULL)
      {
	total = temp_list->the_inode.size;

	s_read_buf = (char *)malloc(total);
	memset(s_read_buf, '\0', total);

	if (total % BLOCKSIZE == 0 && total >= BLOCKSIZE)
	  num_blocks = total / BLOCKSIZE;
	else
	  num_blocks = total / BLOCKSIZE + 1;
	
	temp_block = (block *)malloc(sizeof(block));
	
	if (num_blocks < 7)
	{
	  for (i = 0; i < num_blocks; i++)
	  {
	    temp_block->num = temp_list->the_inode.blocks[i];
	    block_read(&temp_block);
	    
	    if (total < BLOCKSIZE)
	    {
	      memcpy(s_read_buf + (i * BLOCKSIZE), temp_block->content,
		     total);
	    }
	    else
	    {
	      memcpy(s_read_buf + (i * BLOCKSIZE), temp_block->content,
		     BLOCKSIZE);
	    }
	    
	    total -= BLOCKSIZE;
	  }
	}
	
	HASH user_id, group_id, file_id;

	//fprintf(stderr, "+user_id:  %s\n", temp_list->the_inode.user_id);
	//fprintf(stderr, "+group_id: %s\n", temp_list->the_inode.group_id);
	//fprintf(stderr, "+file_id:  %s\n", temp_list->the_inode.file_id);

	sha1_stringToHash(temp_list->the_inode.user_id, &user_id);
	sha1_stringToHash(temp_list->the_inode.group_id, &group_id);
	sha1_stringToHash(temp_list->the_inode.file_id, &file_id);

	// create msg w/ given data
	_msg_data_z *msgData = create_msg_data();
	msgData->version = temp_list->the_inode.version;
	memcpy( &(msgData->node_id), &(temp_list->the_inode.key), sizeof(Key) );
	memcpy( &(msgData->send_key), &(temp_list->the_inode.key), sizeof(Key) );
	memcpy( msgData->user_id, user_id, sizeof(HASH) );
	memcpy( msgData->group_id, group_id, sizeof(HASH) );
	memcpy( msgData->file_name, file_id, sizeof(HASH) );
	msgData->size = temp_list->the_inode.size;
	memset(msgData->data, '\0', temp_list->the_inode.size);
	memcpy(msgData->data, s_read_buf, temp_list->the_inode.size);

	int encodedLen;
	uint8_t* encodedData = msg_data_encode(msgData, &encodedLen);
	
	ChimeraHost **result_set = (ChimeraHost **)route_lookup(temp_list->the_inode.key, NUM_FILES, 1);
	
	for (i = 0 ; i < NUM_FILES; i++)
	{
	  if (result_set[i] != NULL)
	  {
	    // new key is part of this neighbor set for file storage
	    if ((result_set[i])->key == msg_update.key &&
		deviceInfo->nodeKey  != msg_update.key)
	    {
	      //fprintf(stderr, "+ %x -> %s\n", msg_update.key, sha1_hashToString(&(msgData->file_name)));
	      chimera_send(msg_update.key, CTRL_SFILE_STORE, encodedLen, encodedData);
	      break;
	    }
	  }
	  else // break out on first NULL
	    break;
	}

	free(s_read_buf);
	free(encodedData);
	free(result_set);
      
	temp_list = temp_list->next;
      }
      
      free(thelist);
    }

    pthread_mutex_unlock(&mutex_inode);
  
    pthread_mutex_lock(&mutex_table);

    // next maintain the meta nodes among the NUM_METAS replicas
    _msg_meta_z msg_meta;
    int i = 0;

    if (meta_user != NULL)
    {
      _info_group_z* curr_group;
      _meta_local_z* curr_file;
      s_sendbuf = (char *)malloc(sizeof(_msg_meta_z));
      _meta_user_z *curr_user = meta_user;
      while (curr_user != NULL)
      {
	if (curr_user->groups != NULL)
	{
	  curr_group = curr_user->groups;
	  while (curr_group != NULL)
	  {
	    memcpy( &(msg_meta.send_key), &(curr_user->send_key), sizeof(Key) );

	    // msg_meta.send_key = curr_user->send_key;

	    sprintf(msg_meta.user_id, "%s", curr_user->user_id);
	    msg_meta.user_id[40]  = '\0';

	    if (curr_group->files != NULL)
	    {
	      sprintf(msg_meta.group_id, "%s", curr_group->group_id);
	      msg_meta.group_id[40]  = '\0';

	      curr_file = curr_group->files;
	      while (curr_file != NULL)
	      {
		msg_meta.version = curr_file->version;
		msg_meta.size    = curr_file->size;

		memset(msg_meta.file_name, '\0', 95);
		sprintf(msg_meta.file_name, "%s", curr_file->file_name);

		// reset send buffer before sending out
		memset(s_sendbuf, '\0', sizeof(_msg_meta_z));
		memcpy(s_sendbuf, &msg_meta, sizeof(_msg_meta_z));
		
		ChimeraHost **result_set = (ChimeraHost **)route_lookup(curr_user->send_key, NUM_METAS, 1);

		for (i = 0; i < NUM_METAS; i++)
		{
		  if (result_set[i] != NULL)
		  {
		    if ((result_set[i])->key == msg_update.key &&
			deviceInfo->nodeKey  != msg_update.key)
		    {
		      //fprintf(stderr, "+meta+ %x -> %s\n", msg_update.key, msg_meta.file_name);
		      chimera_send(msg_update.key, CTRL_META_COPY, sizeof(_msg_meta_z), s_sendbuf);
		      break;
		    }
		  }
		  else // rest of the list will be NULL as well
		    break;
		}

		curr_file = curr_file->next;
	      }
	    }

	    curr_group = curr_group->next;
	  }
	}

	curr_user = curr_user->next;

	memset(&msg_meta, '\0', sizeof(msg_meta));
      }

      free(s_sendbuf);
    }

    pthread_mutex_unlock(&mutex_table);

    pthread_mutex_unlock(&mutex_update);
  }
  else if (m->type == CTRL_UPDATE_META) { 
    
  }

  else if (m->type == CTRL_SFILE_STORE) { 

	// ***********************************************************************
	// NODE REQUESTS TO STORE REPLICATED DATA
	// ***********************************************************************

	// very similar to CTRL_CFILE_STORE but not sending out any
	// more replication data.  MATT: todo: Currently, not sending any 
	// acks, just assuming stores are successful on target node

    // fprintf(stderr, "RECEIVED CTRL_SFILE_STORE\n");

#ifdef DEBUG	
    //printf("Received a CTRL_SFILE_STORE (STORE REPLICATED DATA)\n");
#endif

    pthread_mutex_lock(&mutex_inode);

    // decode received message
    _msg_data_z *msgData = msg_data_decode(m->payload, m->size);

#ifdef DEBUG
    //  dumpBytes(stderr, m->payload, m->size , " [stor] ");
#endif

    int i = 0, exists = 0, inode_num = 0, curr_size = 0, j = 0, block_size = 0;
    int new_file = 0;
    Key user_key;
    inode *temp_inode, *new_inode;
    block *myblock;
    char write_buf[BLOCKSIZE];

    temp_inode = NULL;
    new_inode  = NULL;
    myblock    = NULL;

    // convert user_id and file_name to char array
    char user_id[41];
    char group_id[41];
    char file_id[41];
    sprintf(user_id, "%s", sha1_hashToString(&(msgData->user_id)));
    sprintf(group_id, "%s", sha1_hashToString(&(msgData->group_id)));
    sprintf(file_id, "%s", sha1_hashToString(&(msgData->file_name)));
    //user_id[40]  = '\0';
    //group_id[40] = '\0';
    //file_id[40]  = '\0';

    FILE* ftemp, *mtemp;

    log_exec(LOG_RECV, msgData->node_id, LOG_SFILE, user_id);

    // !!! UPDATE QFS !!!
    new_file = 0;

    // check if file already exists in the file system
    inode_num = inode_get(&temp_inode, user_id, group_id, file_id);
    if (inode_num > -1) // file exists
    {
      //fprintf(stderr, "exists as ver:%d\n", temp_inode->version);
      if (temp_inode->version < msgData->version)
      {
	//fprintf(stderr, "version is less.. ");

	myblock = (block *)malloc(sizeof(block));
	memset(myblock->content, '\0', BLOCKSIZE);
      
	// clear the blocks list from meta.sto side
	for (i = 0; i < INODE_BLOCKS; i++)
	{
	  // clear the blocks from file.sto side
	  if (temp_inode->blocks[i] != QFS_NULL)
	  {
	    myblock->num = temp_inode->blocks[i];
	    block_free(&myblock);
	    temp_inode->blocks[i] = QFS_NULL;
	  }
	  else
	    break;
	}

	free(myblock);
      }
      else
      {
	// do nothing
      }
    }
    else
    {
      //fprintf(stderr, "adding the NEW file to the system\n");
      
      // if not, then add the file to the file system
      inode_init(&temp_inode);
      inode_num = inode_new();
      inode_set(&temp_inode, user_id, group_id, file_id, msgData->version, msgData->size, msgData->send_key);
      
      new_file = 1;
    }

    /*
    fprintf(stderr, "inode: U: %s\n", temp_inode->user_id);
    fprintf(stderr, "inode: G: %s\n", temp_inode->group_id);

    for (i = 0; i < INODE_BLOCKS; i++)
    {
      if (temp_inode->blocks[i] != QFS_NULL)
	fprintf(stderr, "inode: [%d] = %d\n", i, temp_inode->blocks[i]);
    }
    */

    // either way, both cases will write new blocks out and store it in inode
    // first though, must make sure version received is newer than storage
    if (new_file || temp_inode->version < msgData->version)
    {
      temp_inode->version = msgData->version;

      curr_size = 0;

      i = 0;
      while (curr_size <= msgData->size)
      {
	//fprintf(stderr, "writing block[%d].. ", i);
	memset(write_buf, '\0', BLOCKSIZE);
	  
	// to take care of the last partial block size
	if (msgData->size - curr_size <= BLOCKSIZE)
	{
	  memcpy(write_buf, msgData->data + curr_size, msgData->size - curr_size);
	  block_size = msgData->size - curr_size;
	}
	else
	{
	  memcpy(write_buf, msgData->data + curr_size, BLOCKSIZE);
	  block_size = BLOCKSIZE;
	}
	  
	// write block out to file system
	myblock = NULL;
	block_new(&myblock);
	memcpy(myblock->content, write_buf, BLOCKSIZE);
	
	myblock->size = block_size;
	block_write(&myblock);
	
	// save meta data information
	if (i < INODE_BLOCKS)
	{
	  //fprintf(stderr, "inode->blocks[%d] = %d\n", i, myblock->num);
	  fflush(stderr);
	  temp_inode->blocks[i] = myblock->num;
	}
	else
	{
	  //fprintf(stderr, "shouldn't go in here\n");
	}

	i++;
	curr_size += BLOCKSIZE;
      }
	
      temp_inode->version = msgData->version;
    }
  
    inode_write(&temp_inode, inode_num);
    user_incr_file_inodes(1);

    if (temp_inode != NULL)
      free(temp_inode);
    if (new_inode != NULL)
      free(new_inode);

    free(msgData);

    pthread_mutex_unlock(&mutex_inode);

  }else if (m->type == CTRL_CFILE_REQ) {

    // ***********************************************************************
    // NODE REQUESTS TO GET DATA
    // ***********************************************************************

    pthread_mutex_lock(&mutex_inode);
    
#ifdef DEBUG
    //printf("Received a CTRL_CFILE_REQ\n");
#endif
    
    // decode received message
    _msg_data_z *msgData = msg_data_decode(m->payload, m->size);
    _msg_data_z msg_data;

    Key send_key = msgData->node_id;

    // convert user_id and file_name to char array
    char user_id[41] = { '\0' }; // = sha1_hashToString(&(msgData->user_id));
    char group_id[41] = { '\0' };
    char file_name[41] = { '\0' };

    //memset(group_id, '\0', 40);
    //memset(file_name, '\0', 40);

    sprintf(user_id, "%s", sha1_hashToString(&(msgData->user_id)));
    sprintf(group_id, "%s", sha1_hashToString(&(msgData->group_id)));
    sprintf(file_name, "%s", sha1_hashToString(&(msgData->file_name)));
    user_id[40] = '\0';
    group_id[40] = '\0';
    file_name[40] = '\0';

    log_exec(LOG_RECV, msgData->node_id, LOG_CFILE, LOG_GET);

    // QFS
    inode *temp_inode = NULL;
    inode_init(&temp_inode);
    block *temp_block = NULL;
    int num_blocks, i, total;
    char *s_read_buf = NULL;

    if (inode_get(&temp_inode, user_id, group_id, file_name) != -1)
    {
      total = temp_inode->size;

      s_read_buf = (char *)malloc(temp_inode->size);
      memset(s_read_buf, '\0', temp_inode->size);

      //fprintf(stderr, "FILE FOUND! -- (%d)\n", temp_inode->size);

      if (temp_inode->size % BLOCKSIZE == 0 && temp_inode->size >= BLOCKSIZE)
	num_blocks = temp_inode->size / BLOCKSIZE;
      else
	num_blocks = temp_inode->size / BLOCKSIZE + 1;

      temp_block = (block *)malloc(sizeof(block));


      if (num_blocks < 7)
      {
	for (i = 0; i < num_blocks; i++)
	{
	  temp_block->num = temp_inode->blocks[i];
	  block_read(&temp_block);
	  
	  if (total < BLOCKSIZE)
	  {
	    memcpy(s_read_buf + (i * BLOCKSIZE), temp_block->content,
		   total);
	    //fprintf(stderr, "(%d)", total);
	  }
	  else
	  {
	    memcpy(s_read_buf + (i * BLOCKSIZE), temp_block->content,
		   BLOCKSIZE);
	  }

	  total -= BLOCKSIZE;
	}
      }
      else
      {
	//printf(" File too large\n");
      }

      fflush(stderr);

      // send back to user
      memcpy( &(msg_data.send_key), &send_key, sizeof(Key) );
      memcpy( msg_data.user_id, msgData->user_id, sizeof(HASH) );

      memcpy( msg_data.group_id, msgData->group_id, sizeof(HASH) );
      memcpy( msg_data.file_name, msgData->file_name, sizeof(HASH) );

      msg_data.version = temp_inode->version;
      msg_data.size    = temp_inode->size;
      
      fflush(stderr);

      memset(msg_data.data, '\0', MAX_FILE_SIZE);
      memcpy(msg_data.data, s_read_buf, temp_inode->size);

      int encodedLen;
      uint8_t* encodedData = msg_data_encode(&msg_data, &encodedLen);
      //uint8_t* encodedData = msg_data_encode(msgData, &encodedLen);

      // fprintf(stderr, "sending [%x] back cfile_req_ack_yes (%d)\n", send_key, encodedLen);
      chimera_send(send_key, CTRL_CFILE_REQ_ACK_YES, encodedLen, encodedData);
      //fprintf(stderr, "DONE - size: (%d)\n", temp_inode->size);
      
    }
    else
    {
      // file does not exist on this system
      // fprintf(stderr, "FILE DOES NOT EXIST --> %s\n", file_name);

    }

    pthread_mutex_unlock(&mutex_inode);
    
    free(msgData);

    // fprintf(stderr, "FINISHED a CTRL_CFILE_REQ\n");

  }else if (m->type == CTRL_CFILE_STORE_ACK_YES){

	// ***********************************************************************
	// NODE RECEIVED A SUCCEEDED STORE ACK
	// ***********************************************************************

#ifdef DEBUG
    //printf("Received a CTRL_CFILE_STORE_ACK_YES (our store/commit succeeded)\n");
#endif

	// decode message and decrypt encrypted data
	_msg_data_z *msgData = msg_data_decode(m->payload, m->size); 

	// Notify user
	log_store_ack(&(msgData->file_name), msgData->version, 1);

	//cleanup
	free( msgData );
	
  }else if (m->type == CTRL_CFILE_STORE_ACK_NO) {

	// ***********************************************************************
	// NODE RECEIVED A FAILED STORE ACK
	// ***********************************************************************

#ifdef DEBUG
    //printf("Received a CTRL_CFILE_STORE_ACK_NO (our store/commit failed)\n");
#endif
	// decode message
	_msg_data_z *msgData = msg_data_decode(m->payload, m->size);
	
	// check version number for response
	if (msgData->version == -1){
	  
	  // the store failed. 
#ifdef DEBUG
	  printf("the store failed.\n");
#endif

	  log_store_ack(&(msgData->file_name), msgData->version, 0);
	  
	}else{

	  // our desired store failed, but we received a newer version
#ifdef DEBUG
	  printf("our desired store failed, but we received a newer version\n");
#endif

	  // call function to store the data
	  //// ignoring for now
#ifdef DEBUG
	  printf("ignoring for now\n");
#endif
	  ////log_store_response(&(msgData->file_name), msgData->version,
	  ////					   msgData->size, msgData->data);

	  log_store_ack(&(msgData->file_name), msgData->version, 1);
	  
	}
	// cleanup
	free(msgData);
	
  }else if (m->type == CTRL_CFILE_REQ_ACK_YES) {
  
	// ***********************************************************************
	// NODE RECEIVED A SUCCEEDED REQUEST ACK
	// ***********************************************************************

#ifdef DEBUG
    // printf("Received a CTRL_CFILE_REQ_ACK_YES (our req/update succeeded)\n");
#endif	

    //pthread_mutex_unlock(&mutex_cmdline);
    //pthread_mutex_lock(&mutex_cmdline);

	// This means the message sent back has the data we requested
  
	// decode message and decrypt encrypted data
	_msg_data_z *msgData = msg_data_decode(m->payload, m->size); 

#ifdef DEBUG
	//dumpBytes(stderr, m->payload, m->size , " [recv] ");
#endif

	log_exec(LOG_RECV, msgData->node_id, LOG_FILE_YES, LOG_GET);

	// todo: check versions. if msg version is newer, there is more than
	// one chunk of data (add a new state flag for sent version)
	// currently not caring if version is newer. just accept it.

	//fprintf(stderr, "received size: (%d)\n", msgData->size);

	// DECRYPT
	int success = 0;
	char* keyFilename = NULL;
	char* keyPassword = NULL;

	if( (NULL != (keyFilename = misc_getKeyFilename())) &&
	    (NULL != (keyPassword = misc_getKeyPassword())) )
	{
	  char* decryptedLog = NULL;
	  uint32_t decryptedLen = msgData->size;
	  
	  printf(" Decrypting... ");
	  
	  success = quartz_decrypt_string( keyFilename, keyPassword,
					   msgData->data,
					   (char**)&decryptedLog,
					   &decryptedLen);
	  
	  printf("done! "); fflush(stdout);
	  
	  // call function to store the data
	  log_store_update(&(msgData->file_name), msgData->version,
			   decryptedLog, decryptedLen);

	  if (total_receiving > 0)
	  {
	    total_received++;
	    
	    // processed all the expected number of
	    // files for that last group request
	    if (total_received == total_receiving)
	      pthread_mutex_unlock(&mutex_cmdline);
	  }
	}
  
	// cleanup
	free(msgData);
  
	//pthread_mutex_unlock(&mutex_cmdline);
  
  }else if (m->type == CTRL_CFILE_REQ_ACK_NO) {
	
	// ***********************************************************************
	// NODE RECEIVED A FAILED REQUEST ACK
	// ***********************************************************************

#ifdef DEBUG
    //printf("Received a CTRL_CFILE_REQ_ACK_NO (our req/update failed)\n");
#endif
	
	// Request could have failed for many reasons. 
	// Must check version numbers to see type of failure
    
	// decode message and decrypt encrypted data
	_msg_data_z *msgData = msg_data_decode(m->payload, m->size);
	
	if (msgData->version == -1){
	  
	  // error occured with filestore
#ifdef DEBUG
	  printf("error occured with filestore\n");
#endif
	  
	  log_store_response(&(msgData->file_name), msgData->version, 
						 0, (uint8_t *)NULL, 0);
		
	}else if (msgData->version == 0){
	  
	  // requested file did not exist
#ifdef DEBUG
	  printf("requested file did not exist\n");
#endif

	  log_store_response(&(msgData->file_name), msgData->version,
						 0, (uint8_t *)NULL, 0);
		
	}else if (msgData->version > 0){
	  
	  // requested a file that was newer than the one on the node
	  // msgData->version holds the latest ver that the node has
#ifdef DEBUG
	  printf("requested a file that was newer than the one on the node\n");
#endif

	  log_store_response(&(msgData->file_name), msgData->version,
						  0, (uint8_t *)NULL, 0);
	  
	}else{
	  
	  // error not accounted for	
#ifdef DEBUG
	  printf("error not accounted for\n");
#endif
	  
	  log_store_response(&(msgData->file_name), msgData->version,
						  0, (uint8_t *)NULL, 0);

	}
	
	// cleanup
	free(msgData);

  }else{

	// ***********************************************************************
	// NODE RECEIVED AN UNKNOWN CTRL MESSAGE
	// ***********************************************************************
	
#ifdef DEBUG
	printf("WARNING: Received unknown type of message\n");
#endif

  }

}

/* --------------------------------------------------------------------------- */

static void quartz_store_upcall_update(Key *k, ChimeraHost *h, int joined) {

  if(joined) {
#ifdef DEBUG
    //printf("Node %x:%s:%d joined neighbor set\n", *k, h->name, h->port);
#endif
  }
  else {
#ifdef DEBUG
    //printf("Node %x:%s:%d leaving neighbor set\n", *k, h->name, h->port);
#endif
  }

}

/* -- external functions ----------------------------------------------------- */

extern int quartz_store_init()
{
  /*
  chimera_register(CTRL_CFILE_STORE);
  chimera_register(CTRL_CFILE_STORE_ACK_YES);
  chimera_register(CTRL_CFILE_STORE_ACK_NO);
  chimera_register(CTRL_CFILE_REQ);
  chimera_register(CTRL_CFILE_REQ_ACK_YES);
  chimera_register(CTRL_CFILE_REQ_ACK_NO);
  chimera_register(CTRL_SFILE_STORE);
  chimera_register(CTRL_META_FILE_REQ);
  chimera_register(CTRL_META_GROUP_REQ);
  chimera_register(CTRL_META_REQ_YES);
  chimera_register(CTRL_META_REQ_NO);
  chimera_register(CTRL_META_STORE);
  chimera_register(CTRL_META_STORE_ACK_YES);
  chimera_register(CTRL_META_STORE_ACK_NO);
  chimera_register(CTRL_META_COPY);
  chimera_register(CTRL_DEL_FILE);
  chimera_register(CTRL_DEL_FILE_COPY);
  chimera_register(CTRL_DEL_META);
  chimera_register(CTRL_DEL_META_COPY);
  */

  return 0;
}

/* -- external functions ----------------------------------------------------- */

extern int quartz_store_send_entry( Key node_id, HASH user_id, char* group,
				    HASH group_id, HASH file_id, uint32_t version,
				    uint32_t size, uint8_t *data )
{
  Key key_send;
  key_send = (Key) key_makehash(group);

  // create msg w/ given data
  _msg_data_z *msgData = create_msg_data();
  msgData->version = version;
  memcpy( &(msgData->node_id), &(node_id), sizeof(Key) );
  memcpy( &(msgData->send_key), &(key_send), sizeof(Key) );
  memcpy( msgData->user_id, user_id, sizeof(HASH) );
  memcpy( msgData->group_id, group_id, sizeof(HASH) );
  memcpy( msgData->file_name, file_id, sizeof(HASH) );
  msgData->size = size;
  memcpy( msgData->data, data, size);

  //  msgData->data = data;


  /*
  char* keyFilename = NULL;
  char* keyPassword = NULL;
  char* encryptedLog = NULL;
  char* decryptedLog = NULL;
  uint32_t numBytes = 0;
  uint32_t encryptedLen = strlen(temp_buf);

  encryptedLog = NULL;
  decryptedLog = NULL;
  encryptedLen = strlen(temp_buf);

  if( (NULL != (keyFilename = misc_getKeyFilename())) &&
      (NULL != (keyPassword = misc_getKeyPassword())) )
  {
    fprintf(stderr, "encrypting... ");
    
    quartz_encrypt_string( keyFilename, keyPassword,
			   temp_buf,
			   (char**)&encryptedLog, &encryptedLen );
    
    fprintf(stderr, "done!\n");
    numBytes = encryptedLen;

    fprintf(stderr, "encrypted size = %d, un = %d\n", encryptedLen, strlen(temp_buf));


    // uint32_t decryptedLen = strlen(temp_buf);
    uint32_t decryptedLen = encryptedLen ;
	      
      fprintf(stderr, "decrypting... ");
	      
      quartz_decrypt_string( keyFilename, keyPassword,
			     (char *) encryptedLog,
			     (char**)&decryptedLog,
			     &decryptedLen);
	      
	      fprintf(stderr, "done!\n");
      
	      fflush(stderr);
      fprintf(stderr, "** %s **\n", decryptedLog);
  }

  */





  // encode the message
  uint32_t encodedLen;
  uint8_t* encodedData = msg_data_encode(msgData, &encodedLen);

#ifdef DEBUG
  //dumpBytes( stderr, encodedData, encodedLen, " [send] ");
#endif

  //fprintf(stderr, "-----store_send_key = [%x]\n", key_send);

  log_exec(LOG_SEND, key_send, LOG_CFILE, LOG_STORE);

  // send msg to user_id
  chimera_send(key_send, CTRL_CFILE_STORE, encodedLen, encodedData);

  user_incr_sent();
  user_incr_commits();

  // cleanup
  free( msgData );
  free( encodedData );

  return 0;

}

extern int quartz_remove_entry( Key node_id, HASH user_id, char* group,
				HASH group_id, HASH file_id, uint32_t version,
				uint32_t size, uint8_t *data )
{
  Key key_send;
  key_send = (Key) key_makehash(group);

  // create msg w/ given data
  _msg_data_z *msgData = create_msg_data();
  msgData->version = 0;
  memcpy( &(msgData->node_id), &(node_id), sizeof(Key) );
  memcpy( &(msgData->send_key), &(key_send), sizeof(Key) );
  memcpy( msgData->user_id, user_id, sizeof(HASH) );
  memcpy( msgData->group_id, group_id, sizeof(HASH) );
  memcpy( msgData->file_name, file_id, sizeof(HASH) );
  msgData->size = size;
  memcpy( msgData->data, "mattsun", 7);

  // encode the message
  uint32_t encodedLen;
  uint8_t* encodedData = msg_data_encode(msgData, &encodedLen);

  log_exec(LOG_SEND, key_send, LOG_CFILE, LOG_STORE);

  // send msg to user_id
  chimera_send(key_send, CTRL_DEL_FILE, encodedLen, encodedData);

  user_incr_sent();

  // cleanup
  free( msgData );
  free( encodedData );

  return 0;
}

extern int quartz_remove_group( Key node_id, HASH user_id, char* group, HASH group_id )
{
  Key key_send;
  key_send = (Key) key_makehash(group);

  // create msg w/ given data
  _msg_data_z *msgData = create_msg_data();
  msgData->version = 0;
  memcpy( &(msgData->node_id), &(node_id), sizeof(Key) );
  memcpy( &(msgData->send_key), &(key_send), sizeof(Key) );
  memcpy( msgData->user_id, user_id, sizeof(HASH) );
  memcpy( msgData->group_id, group_id, sizeof(HASH) );
  msgData->size = 0;
  memcpy( msgData->data, "mattsun", 7);

  // encode the message
  uint32_t encodedLen;
  uint8_t* encodedData = msg_data_encode(msgData, &encodedLen);

  log_exec(LOG_SEND, key_send, LOG_CFILE, LOG_STORE);

  // send msg to user_id
  chimera_send(key_send, CTRL_DEL_FILE_GROUP, encodedLen, encodedData);

  user_incr_sent();

  // cleanup
  free( msgData );
  free( encodedData );

  return 0;
}


/* -- external functions ----------------------------------------------------- */

extern int quartz_store_get_entry( Key node_id, HASH user_id, Key group_key,
				   HASH group_id, HASH file_id, uint32_t version )
{
  Key key_send;
  INMEM_FILE *file = NULL;
  // create msg for request
  _msg_data_z *msgData = create_msg_data();

  // create msg for request
  //_msg_data_z *msgData = create_msg_data();
  msgData->version = version;
  memcpy( &(msgData->node_id), &(deviceInfo->nodeKey), sizeof(Key) );
  memcpy( &(msgData->send_key), &(group_key), sizeof(Key) );
  memcpy( msgData->user_id, deviceInfo->userHash, sizeof(HASH) );
  memcpy( msgData->group_id, group_id, sizeof(HASH) );
  memcpy( msgData->file_name, file_id, sizeof(HASH) );
  msgData->size = 0;
  memset( msgData->data, '\0', MAX_FILE_SIZE);
  // msgData->data = 0;

  // encode msg
  uint32_t encodedLen;
  uint8_t* encodedData = msg_data_encode(msgData, &encodedLen);

  if ( (NULL != (file = group_find_file(file_id))) )
  {
    if (version != file->version)
    {
      log_exec(LOG_SEND, group_key, LOG_CFILE, LOG_GET);
  
      // send msg to user_id
      chimera_send(group_key, CTRL_CFILE_REQ, encodedLen, encodedData);

      user_incr_sent();
      user_incr_updates();
    }
    else
      printf("file \"%s\" - already have most recent version!\n", file->name);
  }
  else
    fprintf(stderr, "file not found locally\n");

  //cleanup
  free( msgData );
  free( encodedData );
  
  return 0;
}

/* -- external functions ----------------------------------------------------- */

extern int quartz_store_send_meta( Key node_id, char* user_id, char* file_name,
				   char *group_name, char* group_id,
				   uint32_t version, uint32_t size)
{
  // create msg w/ given data
  int i;
  _msg_meta_z msg_meta;
  memset(&msg_meta, '\0', sizeof(msg_meta));
  char* s_sendbuf;
  
  msg_meta.version = version;

  memcpy( &(msg_meta.node_id), &(node_id), sizeof(Key) );
  memcpy( &(msg_meta.send_key), &(deviceInfo->userKey), sizeof(Key) );

  sprintf(msg_meta.user_id, "%s", user_id);
  msg_meta.user_id[40]  = '\0';

  sprintf(msg_meta.file_name, "%s", file_name);
  msg_meta.file_name[95] = '\0';


  sprintf(msg_meta.group_id, "%s", group_id);
  msg_meta.group_id[40]  = '\0';

  msg_meta.size = size;

  s_sendbuf = (char *)malloc(sizeof(_msg_meta_z));
  memset(s_sendbuf, '\0', sizeof(_msg_meta_z));
  memcpy(s_sendbuf, &msg_meta, sizeof(_msg_meta_z));

  /*
  fprintf(stderr, "SENDING...[meta]\n");
  fprintf(stderr, "  user_id:   %s\n", msg_meta.user_id);
  fprintf(stderr, "  file_name: %s\n", msg_meta.file_name);
  fprintf(stderr, "  group_id:  %s\n", msg_meta.group_id);
  fprintf(stderr, "  version:   %d\n", msg_meta.version);
  fprintf(stderr, "  size:      %d\n", msg_meta.size);
  */

  log_exec(LOG_SEND, deviceInfo->userKey, LOG_CMETA, LOG_STORE);

  chimera_send(deviceInfo->userKey, CTRL_META_STORE, sizeof(_msg_meta_z), s_sendbuf);

  free( s_sendbuf );
  return 0;
}

/* -- external functions ----------------------------------------------------- */

extern int quartz_store_get_meta( Key node_id, char* user_id, char* file_name,
				  char *group_name, char* group_id)
{
  // create msg w/ given data
  int i;
  _msg_meta_z msg_meta;
  memset(&msg_meta, '\0', sizeof(msg_meta));
  char* s_sendbuf;
  
  msg_meta.version = 0;

  memcpy( &(msg_meta.node_id), &(node_id), sizeof(Key) );
  memcpy( &(msg_meta.send_key), &(deviceInfo->userKey), sizeof(Key) );

  sprintf(msg_meta.user_id, "%s", user_id);
  msg_meta.user_id[40]  = '\0';

  memset(msg_meta.file_name, '\0', 95);
  sprintf(msg_meta.file_name, "%s", file_name);
  msg_meta.file_name[95] = '\0';

  sprintf(msg_meta.group_id, "%s", group_id);
  msg_meta.group_id[40]  = '\0';

  msg_meta.size = 0;

  s_sendbuf = (char *)malloc(sizeof(_msg_meta_z));
  memset(s_sendbuf, '\0', sizeof(_msg_meta_z));
  memcpy(s_sendbuf, &msg_meta, sizeof(_msg_meta_z));

  log_exec(LOG_SEND, deviceInfo->userKey, LOG_CFILE, LOG_GET);

  //fprintf(stderr, "sending ctrl_meta_file_req to %x\n", deviceInfo->userKey);
  chimera_send(deviceInfo->userKey, CTRL_META_FILE_REQ, sizeof(_msg_meta_z), s_sendbuf);
 
  // cleanup
  free( s_sendbuf );

  return 0;


}

/* -- external functions ----------------------------------------------------- */

extern int quartz_store_get_metas( Key node_id, char* user_id,
				   char *group_name, char* group_id)
{
  int i;
  _msg_meta_z msg_meta;
  memset(&msg_meta, '\0', sizeof(msg_meta));
  char* s_sendbuf;
  
  Key key_send = (Key) key_makehash(group_name);

  memcpy( &(msg_meta.node_id), &(node_id), sizeof(Key) );
  memcpy( &(msg_meta.send_key), &(deviceInfo->userKey), sizeof(Key) );
  memcpy( &(msg_meta.group_key), &(key_send), sizeof(Key) );
    
  sprintf(msg_meta.user_id, "%s", user_id);
  msg_meta.user_id[40]  = '\0';

  sprintf(msg_meta.group_id, "%s", group_id);
  msg_meta.group_id[40]  = '\0';
  msg_meta.size = 0;

  s_sendbuf = (char *)malloc(sizeof(_msg_meta_z));
  memset(s_sendbuf, '\0', sizeof(_msg_meta_z));
  memcpy(s_sendbuf, &msg_meta, sizeof(_msg_meta_z));

  log_exec(LOG_SEND, deviceInfo->userKey, LOG_GMETA, LOG_GET);

  chimera_send(deviceInfo->userKey, CTRL_META_GROUP_REQ, sizeof(_msg_meta_z), s_sendbuf);
  fflush(stdout);
  
  // cleanup
  free( s_sendbuf );
  //pthread_mutex_unlock(&mutex_cmdline);

  return 0;
}


/* --------------------------------------------------------------------------- */

int store_main(int argc, char **argv) {

  int opt;
  char *hn;
  int port;
  ChimeraHost *host = NULL;
  char tmp[256];
  unsigned int key, my_key;
  int i;
  char *action;

  while((opt = getopt(argc, argv, OPTSTR)) != EOF) {
    switch((char) opt) {
      case 'j':
        for(i=0; optarg[i] != ':' && i<strlen(optarg); i++);
	optarg[i] = 0;
	hn = optarg;
	sscanf(optarg + (i + 1), "%d", &port);
	break;
      default:
        fprintf(stderr, "invalid option %c\n", (char) opt);
        fprintf(stderr, "usage: %s\n", USAGE);
        exit(1);
    }
  }

  if((argc - optind) != 2) {
    fprintf(stderr, "usage: %s\n", USAGE);
    exit(1);
  }

  port = atoi(argv[optind]);
  sscanf(argv[optind+1], "%x", &i);

  my_key = (Key)i;

  chimera_init(port);
  chimera_setkey(my_key);

  chimera_forward(quartz_store_upcall_fwd);
  chimera_deliver(quartz_store_upcall_del);
  chimera_update(quartz_store_upcall_update);

  // init other chimera parts
  quartz_store_init();

  if(host != NULL) {
    chimera_join(host);
  }

  printf("** send test log to key with command <key> send **\n");
  printf("** get test log from key with command <key> get **\n");

  HASH  user;              // user should be current user
  sha1_stringToHash( "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa", &user );

  while(fgets(tmp, 256, stdin) != NULL) 
  {
    if(strlen(tmp) > 2) {
      // parse input
      for(i = 0; tmp[i] != '\n'; i++);
      tmp[i] = 0;
      for(i = 0; tmp[i] != ' ' && i < strlen(tmp); i++);
      tmp[i] = 0;
      i++;
      sscanf(tmp, "%x", &key);

      action = malloc ((strlen(tmp+i)+1) * sizeof(char));
      strncpy(action, tmp+i, (strlen(tmp+i)+1)); 

      if (strcmp(action, "send")==0){
	
		/*
		// assuming key was generated into file HOME/.quartz/key
		char keyFilename[ FILENAME_MAX ];
		snprintf( keyFilename, FILENAME_MAX, "%s/.quartz/key", getenv( "HOME" ) );
	
		char *password = getKeyPassword();
		quartz_generate_key(keyFilename, password);
		*/

		// encode and encrypt a test log
		LOG *testLog = create_test_log();
		uint32_t  encryptedLen;
		uint8_t* encryptedLog = log_encrypt( testLog, &encryptedLen );

		
		// test version sending
		int test_version = 2;

		printf("hashtostring: %s\n", sha1_hashToString(&user));

		printf("Storing to: %d\n", key_makehash(user));
		printf("OR      to: %d\n", key_makehash(user));

		// store it
		//		quartz_store_send_entry( my_key, user, testLog->parentId, test_version,
		//							encryptedLen, encryptedLog );	

		/// should also do a log delete on the testLog links
		log_delete( testLog );

		// cleanup
		free( encryptedLog );

      }else if (strcmp(action, "get")==0){

		// create test hashes to get

		HASH  file_name;         // actual hash of file we want
		sha1_stringToHash( "FA26BE19DE6BFF93F70BC2308434E4A440BBAD02", &file_name );

		// test version to get
		int test_version = 1;

		printf("hashtostring: %s\n", sha1_hashToString(&user));

		printf("Requesting to: %d\n", key_makehash(user));
		printf("OR         to: %d\n", key_makehash(user));

		// get it
		//quartz_store_get_entry( my_key, user, file_name, test_version );

      }else{
		printf("** command (%s) not yet supported **\n", action);
      }
      
      free ( action );
    }
  }

  return 0;
}
