/*
** $Id: message.c,v 1.18 2005/04/27 04:30:08 msa Exp $
**
** Matthew Allen
** description: 
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <errno.h>
#include <string.h>
#include "chimera.h"
#include "message.h"
#include "job_queue.h"
#include "log.h"
#include "network.h"
#include "jrb.h"
#include "jval.h"

/* message on the wire is encoded in the following way:
** [ type ] [ size ] [ key ] [ data ] */
#define HEADER_SIZE (sizeof(unsigned long) + sizeof(unsigned long) + sizeof(Key))

struct {
  JRB handlers;
  void *jobq;
  pthread_attr_t attr;
  pthread_mutex_t lock;
} MESSAGE_GLOBAL;

Message *message_create(Key dest, int type, int size, char *payload) {

  Message *message;
  message = (Message *) malloc(sizeof(Message));
  message->dest = dest;
  message->type = type;
  message->size = size;
  message->payload = malloc(sizeof(char) * size);
  memcpy(message->payload, payload, size);
  return(message);

}

int message_init(int port) {

  pthread_t tid;

  MESSAGE_GLOBAL.handlers = make_jrb();
  if(!network_init(port)) {
    return(0);
  }

  if(pthread_attr_init(&MESSAGE_GLOBAL.attr) != 0) {
    log_message(LOG_ERROR, "pthread_attr_init: %s", strerror(errno));
    return(0);
  }
  if(pthread_attr_setscope(&MESSAGE_GLOBAL.attr, PTHREAD_SCOPE_SYSTEM) != 0) {
    log_message(LOG_ERROR, "pthread_attr_setscope: %s", strerror(errno));
    return(0);
  }
  if(pthread_attr_setdetachstate(&MESSAGE_GLOBAL.attr, PTHREAD_CREATE_DETACHED) != 0) {
    log_message(LOG_ERROR, "pthread_attr_setdetachstate: %s", strerror(errno));
    return(0);
  }       

  if(pthread_mutex_init(&MESSAGE_GLOBAL.lock, NULL) != 0) {
    log_message(LOG_ERROR, "pthread_mutex_init: %s", strerror(errno));
    return(0);
  }

  if(pthread_create(&tid, &MESSAGE_GLOBAL.attr, network_activate, NULL) != 0) {
    log_message(LOG_ERROR, "pthread_mutex_init: %s", strerror(errno));
    return(0);
  }

  MESSAGE_GLOBAL.jobq = job_queue_init(10);

  return(1);

}

void message_reciever(Message *message) {

  JRB node;
  messagehandler_t func;

  /* find message handler */
  pthread_mutex_lock(&MESSAGE_GLOBAL.lock);
  node = jrb_find_int(MESSAGE_GLOBAL.handlers, message->type);
  if(node == NULL) {
    log_message(LOG_WARN, "received unrecognized message tpye %d\n", message->type);
    pthread_mutex_unlock(&MESSAGE_GLOBAL.lock);
    return;
  }
  func = node->val.v;
  pthread_mutex_unlock(&MESSAGE_GLOBAL.lock);

  /* call the handler */
  func(message);
  free(message->payload);
  free(message);

}

void message_recieved(char *data, int size) {

  unsigned long msgtype;
  unsigned long msgsize;
  unsigned long msgdest;
  Message *message;

  /* decode message and create Message structure */
  memcpy(&msgtype, data, sizeof(unsigned long));
  msgtype = ntohl(msgtype);
  memcpy(&msgsize, data + sizeof(unsigned long), sizeof(unsigned long));
  msgsize = ntohl(msgsize);
  memcpy(&msgdest, data + (2 * sizeof(unsigned long)), sizeof(Key));
  msgdest = ntohl(msgdest);
  message = message_create((Key) msgdest, (int) msgtype, (int) msgsize, data + HEADER_SIZE);

  job_submit(MESSAGE_GLOBAL.jobq, message_reciever, message, 0);

}

void message_handler(int type, messagehandler_t func) {

  JRB node;

  /* add message handler function into the set of all handlers */
  pthread_mutex_lock(&MESSAGE_GLOBAL.lock);
  node = jrb_find_int(MESSAGE_GLOBAL.handlers, type);
  /* don't allow duplicates */
  if(node != NULL) {
    log_message(LOG_WARN, "message handler already registered with %d\n", type);
    pthread_mutex_unlock(&MESSAGE_GLOBAL.lock);
    return;
  }
  jrb_insert_int(MESSAGE_GLOBAL.handlers, type, new_jval_v(func));
  pthread_mutex_unlock(&MESSAGE_GLOBAL.lock);

}

int message_send(ChimeraHost *host, Message *message) {

  char *data;
  unsigned long size;
  unsigned long type;
  unsigned long dest;
  int ret;

  size = HEADER_SIZE + message->size;
  data = (char *) malloc(sizeof(char) * size);

  /* encode the message */
  type = htonl((unsigned long) message->type);
  memcpy(data, &type, sizeof(unsigned long));
  size = htonl((unsigned long) message->size);
  memcpy(data + sizeof(unsigned long), &size, sizeof(unsigned long));
  dest = htonl((unsigned long) message->dest);
  memcpy(data + (2 * sizeof(unsigned long)), &dest, sizeof(Key));
  memcpy(data + HEADER_SIZE, message->payload, message->size);

  /* send the message */
  size = HEADER_SIZE + message->size; /* reset due to htonl */
  ret = network_send(host, data, size);
  free(data);

  return(ret);

}
